Package: CopernicusDataspace
Title: Search Download and Handle Data from the Copernicus Data Space
        Ecosystem
Version: 0.0.4
Authors@R: 
    c(
      person("Pepijn", "de Vries", role = c("aut", "cre"),
             email = "pepijn.devries@outlook.com",
             comment = c(ORCID = "0000-0002-7961-6646")),
      person("Alicia", "Hamer", role = c("rtm"),
             email = "alicia.hamer@wur.nl",
             comment = c(ORCID = "0000-0001-6809-622X")),
      person("LVVN (Ministerie van Landbouw, Visserij, Voedselzekerheid en Natuur)", , , , role = "fnd",
             comment = c(ROR = "03b1hdw57")),
      person("WMR (Wageningen Marine Research)", , , , role = "ctr",
             comment = c(ROR = "013dzwk66"))
    )
Description: 
    The Copernicus Data Space Ecosystem, is an open
    ecosystem that provides free instant access to a wide range
    of data and services from the Copernicus Sentinel missions and
    more on our planet’s land, oceans and atmosphere. This package
    provides entry points to several APIs allowing users to access
    the data directly in R.
Depends: R (>= 4.1.0)
Imports: dplyr, httr2, jsonlite, lubridate, memoise, methods, paws,
        rlang, sf, stringr, tibble, tidyr, xml2
Suggests: clipr, jose, knitr, lwgeom, rmarkdown, rstudioapi, stars,
        testthat (>= 3.0.0)
License: GPL (>= 3)
Encoding: UTF-8
Language: en-GB
Collate: 'CopernicusDataspace-package.R' 'helpers.R' 'account.R'
        'init.R' 'geometry.R' 'tidyverse.R' 'login.R' 's3.R'
        'req_perform.R' 'odata_products.R' 'odata_bursts.R'
        'sentinelhub.R' 'stac.R'
Config/testthat/edition: 3
URL: https://github.com/pepijn-devries/CopernicusDataspace,
        https://pepijn-devries.github.io/CopernicusDataspace/
BugReports: https://github.com/pepijn-devries/CopernicusDataspace/issues
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-20 10:43:50 UTC; vries171
Author: Pepijn de Vries [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7961-6646>),
  Alicia Hamer [rtm] (ORCID: <https://orcid.org/0000-0001-6809-622X>),
  LVVN (Ministerie van Landbouw, Visserij, Voedselzekerheid en Natuur)
    [fnd] (ROR: <https://ror.org/03b1hdw57>),
  WMR (Wageningen Marine Research) [ctr] (ROR:
    <https://ror.org/013dzwk66>)
Maintainer: Pepijn de Vries <pepijn.devries@outlook.com>
Repository: CRAN
Date/Publication: 2026-07-20 12:50:14 UTC

# Generated by roxygen2: do not edit by hand

export(arrange.odata_request)
export(arrange.stac_request)
export(collect.odata_request)
export(collect.sentinel_request)
export(collect.stac_request)
export(compute.odata_request)
export(dse_access_token)
export(dse_get_client_id)
export(dse_get_client_secret)
export(dse_get_password)
export(dse_get_token_details)
export(dse_get_username)
export(dse_has_account)
export(dse_has_client_info)
export(dse_has_s3_secret)
export(dse_odata_attributes)
export(dse_odata_bursts)
export(dse_odata_bursts_request)
export(dse_odata_download)
export(dse_odata_download_path)
export(dse_odata_product_nodes)
export(dse_odata_products)
export(dse_odata_products_request)
export(dse_odata_quicklook)
export(dse_public_access_token)
export(dse_s3)
export(dse_s3_download)
export(dse_s3_get_key)
export(dse_s3_get_secret)
export(dse_s3_set_gdal_options)
export(dse_s3_set_key)
export(dse_s3_set_secret)
export(dse_s3_uri_to_vsi)
export(dse_set_client_id)
export(dse_set_client_secret)
export(dse_set_gdal_token)
export(dse_set_password)
export(dse_set_username)
export(dse_sh_collections)
export(dse_sh_custom_scripts)
export(dse_sh_features)
export(dse_sh_get_custom_script)
export(dse_sh_prepare_input)
export(dse_sh_prepare_output)
export(dse_sh_process)
export(dse_sh_queryables)
export(dse_sh_search_request)
export(dse_sh_use_requests_builder)
export(dse_stac_client)
export(dse_stac_collections)
export(dse_stac_download)
export(dse_stac_get_uri)
export(dse_stac_guess_collection)
export(dse_stac_queryables)
export(dse_stac_search_request)
export(dse_usage)
export(dse_user_statistics)
export(filter.odata_request)
export(filter.sentinel_request)
export(filter.stac_request)
export(req_perform)
export(select.odata_request)
export(select.sentinel_request)
export(select.stac_request)
export(slice_head.odata_request)
export(slice_head.sentinel_request)
export(slice_head.stac_request)
export(st_intersects.odata_request)
export(st_intersects.stac_request)
importFrom(memoise,memoise)
importFrom(methods,setOldClass)
importFrom(rlang,.data)
importFrom(rlang,.env)

# CopernicusDataspace <img src="man/figures/logo.svg" align="right" height="139" alt="logo" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/pepijn-devries/CopernicusDataspace/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/pepijn-devries/CopernicusDataspace/actions/workflows/R-CMD-check.yaml)
![cranlogs](https://cranlogs.r-pkg.org/badges/CopernicusDataspace)
[![version](https://www.r-pkg.org/badges/version/CopernicusDataspace)](https://CRAN.R-project.org/package=CopernicusDataspace)
[![cran
checks](https://badges.cranchecks.info/worst/CopernicusDataspace.svg)](https://cran.r-project.org/web/checks/check_results_CopernicusDataspace.html)
[![CopernicusDataspace status
badge](https://pepijn-devries.r-universe.dev/badges/CopernicusDataspace)](https://pepijn-devries.r-universe.dev/CopernicusDataspace)
[![Codecov test
coverage](https://codecov.io/gh/pepijn-devries/CopernicusDataspace/graph/badge.svg)](https://app.codecov.io/gh/pepijn-devries/CopernicusDataspace)
<!-- badges: end -->

The [Copernicus Data Space Ecosystem](https://dataspace.copernicus.eu/),
is an open ecosystem that provides free instant access to a wide range
of data and services from the Copernicus Sentinel missions and more on
our planet’s land, oceans and atmosphere. This package provides entry
points to several APIs allowing users to access the data directly in R.

## Installation

Install CRAN release:

``` r
install.packages("CopernicusDataspace")
```

Install latest developmental version from R-Universe:

``` r
install.packages("CopernicusDataspace", repos = c('https://pepijn-devries.r-universe.dev', 'https://cloud.r-project.org'))
```

## Introduction

The Copernicus Data Space Ecosystem offers access to its services
through several
[APIs](https://documentation.dataspace.copernicus.eu/APIs.html). This
package offers access via the following APIs. This README shows only
essential methods for downloading data. For more detailed information
consult the respective vignettes:

- [OData](https://documentation.dataspace.copernicus.eu/APIs/OData.html);
  details in `vignette("OData")`,
- [STAC](https://documentation.dataspace.copernicus.eu/APIs/STAC.html);
  details in `vignette("STAC")`, and
- [SentinelHub](https://documentation.dataspace.copernicus.eu/APIs/SentinelHub.html);
  details in `vignette("SentinelHub")`.

## Authentication

When actually downloading files from the Copernicus Data Space
Ecosystem, you usually need some form of authentication. Examples below
use credentials stored as environment variables. For more information
about how to effectively authenticate download requests, please see
`vignette("Authentication")`.

## Downloading Using STAC Catalogue

When you know the STAC identifier (`asset_id`) and the asset name you
wish to download, you can simply use the example shown below. It will
often speed you request up if you include the `collection` id. When you
don’t know these details, you should first explore the catalogue. Please
see `vignette("STAC")` for more details on working with the STAC
catalogue.

``` r
library(CopernicusDataspace)
library(stars) ## For reading and plotting the downloaded file

## Only run this if an S3 secret is specified:
if (dse_has_s3_secret()) {
  
  filename  <- dse_stac_download(
    asset_id   = "Copernicus_DSM_COG_30_S69_00_W062_00_DEM",
    asset      = "data",
    collection = "cop-dem-glo-90-dged-cog",
    tempdir())
  
  tile_stac <- read_stars(filename)
  
  plot(tile_stac, col = hcl.colors(100), axes = TRUE)
  
}
```

<img src="man/figures/README-download-stac-1.png" alt="Example of downloading with STAC API" width="100%" />

This produces a tile in the Southern Ocean near Antarctica.

## Downloading Using OData API

The example below briefly demonstrate how to download data using with
the OData API wrapped by this package. This produces a map of Lake
Upemba in Congo. For more extensive information about the API and
workflow check out `vignette("OData")`.

``` r

## Only run this if client info is available:
if (dse_has_account()) {
  response <-
    dse_odata_download_path(
    product     = "2f497806-0101-5eea-83fa-c8f68bc56b0c",
    node_path   = paste(
      "DEM1_SAR_DTE_90_20101213T034716_20130408T035028_ADS_000000_5033.DEM",
      "Copernicus_DSM_30_S09_00_E026_00", "DEM",
      "Copernicus_DSM_30_S09_00_E026_00_DEM.dt1", sep = "/"),
    destination = tempdir())
  
  tile_odata <- read_stars(response$body)
  plot(tile_odata, col = terrain.colors(100), axes = TRUE)
}
```

<img src="man/figures/README-download-odata-1.png" alt="Example of downloading with OData API" width="100%" />

## Downloading Using SentinelHub

SentinelHub is a service that remotely processes raw Sentinel data and
returns the result. Below a quick demonstration of the process. For more
detailed information check `vignette("SentinelHub")`

``` r
## Only run this if client info is available:
if (dse_has_client_info()) {
  bounds <- c(5.261, 52.680, 5.319, 52.715)

  ## prepare input data:
  input <-
    dse_sh_prepare_input(
      bounds = bounds,
      time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
    )

  ## prepare ouput format:
  output <- dse_sh_prepare_output(bbox = bounds)

  ## retrieve processing script:
  evalscript <- dse_sh_get_custom_script("/sentinel-2/l2a_optimized/")

  ## destination file:
  fl <- tempfile(fileext = ".tiff")
  
  ## send request and download result:
  dse_sh_process(input, output, evalscript, fl)

  ## read and plot result:
  enkhuizen <- read_stars(fl) |> suppressWarnings()
  plot(enkhuizen, rgb = 1:3, main = "Enkhuizen")
}
```

<img src="man/figures/README-download-sentinelhub-1.png" alt="Example of downloading with Sentinel Hub API" width="100%" />

## Function Names

Understanding the naming logic of the functions in this package, will
help you navigating through all features, and effectively use them.

Most function names in this package start with the acronym `dse` (Data
Space Ecosystem). This is done to avoid confusion with other Copernicus
packages (see below) and unintended masking of other functions.

Exceptions are functions that (intentionally) mask functions from other
packages (like `req_perform()`), and functions that implement generic
functions from tidyverse packages (see `?tidy_verbs`).

Functions are also grouped to highlight the specific API they support.
After the `dse` acronym, these function names contain a reference to the
relevant API (`odata`, `stac` and `sh` (SentinelHub)). If a function
name starts with `dse` but is not followed by a reference to a specific
API, it will have a more generic purpose (like for instance
authorisation).

## OpenEO

OpenEO is and will not be supported by the package at hand. It has a
dedicated R package on CRAN:
[openeo](https://doi.org/10.32614/CRAN.package.openeo), with additional
documentation available at
[copernicus.eu](https://documentation.dataspace.copernicus.eu/APIs/openEO/R_Client/R.html).
[OpenEO](https://openeo.org/about.html) allows to process data remotely
and retrieve only processed data.

## More of Copernicus

More R packages for exploring other Copernicus data services:

- [CopernicusClimate](https://github.com/pepijn-devries/CopernicusClimate)
  Dedicated to climate change datasets
- [CopernicusMarine](https://github.com/pepijn-devries/CopernicusMarine)
  Dedicated to marine datasets

## Code of Conduct

Please note that the CopernicusDataspace project is released with a
[Contributor Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

## Acknowledgement

This software has been developed as part of the statutory task programme
“fisheries research” and is subsidised by the [Dutch Ministry of
Agriculture, Nature and Food Quality](https://ror.org/03b1hdw57).

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 8,
  fig.height = 7,
  fig.ext = "png"
)
eval_chunks <- Sys.getenv("COPERNICUS_DATA_VIGNETTE") != ""

## ----setup, message=FALSE-----------------------------------------------------
library(CopernicusDataspace)
library(dplyr)
library(jsonlite)
library(sf)
library(stars)

## ----data-download, eval=eval_chunks------------------------------------------
# if (dse_has_s3_secret()) {
#   ## Define the area of interest:
#   bbox <- st_bbox(c(xmin =  4.66,
#                     ymin = 52.842,
#                     xmax =  7.15,
#                     ymax = 53.62), crs = 4326)
# 
#   ## Assume that we have already located the asset below within our
#   ## area of interest:
#   id <- "S1A_IW_GRDH_1SDV_20241125T055820_20241125T055845_056707_06F55C_12F9_COG"
# 
#   ## And it's from this collection:
#   cllctn <- "sentinel-1-grd"
# 
#   ## Get the URI for the manifest.safe file. This is the preferred entry
#   ## point into the data. It contains all bands and all meta-information
#   ## (like accurate Ground Control Points):
#   uri_safe <- dse_stac_get_uri(id, "safe_manifest", cllctn)
# 
#   ## Translate the URI into a virtual system indicator (VSI), such that
#   ## the stars package can interact with it:
#   vsi_safe <- dse_s3_uri_to_vsi(uri_safe)
# 
#   ## Set up the S3 credentials such that the stars package (and its
#   ## underpinning GDAL library) will recognise them.
#   dse_s3_set_gdal_options()
# 
#   ## Create a stars proxy object from the VSI. No data is downloaded
#   ## yet. But you can set up lazy data manipulations with tidyverse.
#   proxy <- read_stars(vsi_safe, proxy = TRUE)
# }

## ----meta-info, eval=eval_chunks----------------------------------------------
# ## Use `gdal_info` to retrieve the raster information
# json_info <-
#   gdal_utils("info", vsi_safe, options = c("-json"), quiet = TRUE) |>
#   fromJSON()
# 
# ## Get the ground control points for the specific raster.
# ## You will need it for subsetting the proxy.
# gcp <-
#   st_as_sf(json_info$gcps$gcpList, coords= c("x", "y"), crs =
#              st_crs(json_info$gcps$coordinateSystem$wkt))
# 
# ## A plot showing where the raster's GCPs are, and where the bbox lies
# par(mar = c(1, 1, 1, 1))
# plot(gcp["pixel"] |> st_geometry(), main = "Ground Control Points", col = NA)
# plot(gcp["pixel"], add = TRUE)
# plot(bbox |> st_as_sfc() |> st_geometry(), border = "red", add = TRUE)
# 
# ## Determine which raster indices are within the bbox
# ## of our area of interest:
# gcp_crop <-
#   st_crop(gcp,
#           st_transform(bbox |>
#                          st_as_sfc() |>
#                          st_buffer(units::as_units(1, "km")),
#                        st_crs(json_info$gcps$coordinateSystem$wkt)))
# 
# dim_x <- st_get_dimension_values(proxy, "x") |> sort()
# dim_x <- which(dim_x >= min(gcp_crop$pixel) & dim_x <= max(gcp_crop$pixel))
# dim_y <- st_get_dimension_values(proxy, "y") |> sort()
# dim_y <- which(dim_y >= min(gcp_crop$line) & dim_y <= max(gcp_crop$line))

## ----echo=FALSE---------------------------------------------------------------
knitr::include_graphics("../man/figures/read-data-gcp.png")

## ----data-mask, eval=eval_chunks----------------------------------------------
# masked_proxy <-
#   proxy[,dim_x, dim_y,] |>
#   st_apply(c("x", "y"), \(x) {
#     ## From json_info$bands we know that band 1=VH and 2=VV
#     ## We want VH but only when VV >= 30:
#     ifelse(x[2] >= 30, x[1], NA)
#   })
# 
# ## Store the masked file:
# masked_file <- tempfile(fileext = ".tif")
# write_stars(masked_proxy, masked_file, chunk_size = c(1024, 1024),
#             NA_value = -9999)

## ----data-read-masked, eval=eval_chunks---------------------------------------
# masked <- read_stars(masked_file)
# plot(masked, main = "masked")

## ----echo=FALSE---------------------------------------------------------------
knitr::include_graphics("../man/figures/read-data-masked.png")

## ----data-add-gcp, eval=eval_chunks-------------------------------------------
# ## The masked file has lost original GCP information
# ## So create a reference file where we explicitly add the GCP info
# ref_file <- tempfile(fileext = ".tif")
# 
# ## Update GCP info for selected raster indices:
# gcp <-
#   json_info$gcps$gcpList |>
#   mutate(pixel = pixel - min(dim_x) - 1,
#          line = line - min(dim_y) - 1) |>
#   filter(line >= 0 & pixel >= 0)
# opts <-
#   cbind("-gcp", as.matrix(gcp[, c("pixel", "line", "x", "y", "x")])) |>
#   apply(1, c) |> c()
# 
# ## Create reference file with GCP info
# gdal_utils(
#   "translate",
#   source = masked_file,
#   destination = ref_file,
#   options = c(opts,
#               "-a_srs", json_info$gcps$coordinateSystem$wkt,
#               "-a_nodata", "-9999") ## missing values
# )

## ----data-warp, eval=eval_chunks----------------------------------------------
# ## Warp the masked file to a fixed raster with meaningful CRS
# ## (In this case UTM31 with 10x10m cells)
# warped_file <- tempfile(fileext = ".tif")
# 
# gdal_utils(
#   "warp",
#   source = ref_file,
#   destination = warped_file,
#   options = c(
#     "-t_srs", "EPSG:32631",
#     "-tps",
#     "-tr", c(10, 10), "-tap", "-r", "bilinear",
#     "-srcnodata", "-9999", ## missing values
#     "-dstnodata", "-9999", ## missing values
#     "-overwrite"))
# warped <- read_stars(warped_file)
# 
# plot(warped, reset = FALSE, axes = TRUE, main = "warped")

## ----echo=FALSE---------------------------------------------------------------
knitr::include_graphics("../man/figures/read-data-warped.png")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
eval_chunks <-
  CopernicusDataspace::dse_has_client_info() &&
  CopernicusDataspace::dse_has_account() &&
  CopernicusDataspace::dse_has_s3_secret()

## ----stac-collections, eval=eval_chunks---------------------------------------
library(CopernicusDataspace)
dse_stac_collections()

## ----queryables, eval=eval_chunks---------------------------------------------
dse_stac_queryables("sentinel-1-grd") |> summary()

## ----stac-search, eval=eval_chunks--------------------------------------------
library(dplyr)
library(sf)

bbox <-
  sf::st_bbox(
    c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
    crs = 4326)

dse_stac_search_request("sentinel-1-grd") |>
  filter(`sat:orbit_state` == "ascending") |>
  arrange("id") |>
  st_intersects(bbox) |>
  collect()

## ----stac-uri, eval=eval_chunks-----------------------------------------------
dse_stac_get_uri(
  asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
  asset = "B01",
  collection = "sentinel-2-l1c"
)

## ----stac-download, eval=FALSE------------------------------------------------
# dse_stac_download(
#   asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
#   asset = "B01",
#   collection = "sentinel-2-l1c",
#   destination = tempdir()
# )

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
eval_chunks <-
  CopernicusDataspace::dse_has_client_info() &&
  CopernicusDataspace::dse_has_account() &&
  CopernicusDataspace::dse_has_s3_secret()

## ----sh-collections, message=FALSE, eval=eval_chunks--------------------------
library(CopernicusDataspace)
library(stars) ## required below for reading and plotting data
library(dplyr) ## used for data manipulation below
dse_sh_collections()

## ----sh-search, eval=eval_chunks----------------------------------------------
bounds <- c(5.261, 52.680, 5.319, 52.715)
if (dse_has_client_info()) {
  dse_sh_search_request(
    collection = "sentinel-2-l2a",
    bbox       = bounds,
    datetime   = c("2025-01-01 UTC", "2025-01-31 UTC")
  ) |>
    filter(`eo:cloud_cover` <= 10) |>
    collect()
}

## ----prepare-sentinelhub, eval=eval_chunks------------------------------------
## prepare input data:
input <-
  dse_sh_prepare_input(
    bounds = bounds,
    time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
  )

## prepare ouput format:
output <- dse_sh_prepare_output(bbox = bounds)

## retrieve processing script:
evalscript <- dse_sh_get_custom_script("/sentinel-2/l2a_optimized/")

## destination file:
fl <- tempfile(fileext = ".tiff")

## ----download-sentinelhub, eval=FALSE-----------------------------------------
# if (dse_has_client_info()) {
#   ## send request and download result:
#   dse_sh_process(input, output, evalscript, fl)
# }

## ----eval-scripts, eval=eval_chunks-------------------------------------------
dse_sh_custom_scripts()

## ----water-bodies, eval=eval_chunks-------------------------------------------

if (dse_has_client_info()) {
  evalscript <-
    dse_sh_get_custom_script("/sentinel-2/simple_water_bodies_mapping-swbm/")
  
  ## destination file:
  fl_waterbody <- tempfile(fileext = ".tiff")
  
  ## send request and download result:
  dse_sh_process(input, output, evalscript, fl_waterbody)

  ## read and plot result:
  waterbodies <- read_stars(fl_waterbody) |> suppressWarnings()
  plot(waterbodies, rgb = 1:3, main = "Water bodies")
}

## ----land-use, eval=eval_chunks-----------------------------------------------
if (dse_has_client_info()) {
  evalscript <-
    dse_sh_get_custom_script("/sentinel-2/land_use_with_linear_discriminant_analysis/")
  
  ## destination file:
  fl_landuse <- tempfile(fileext = ".tiff")
  
  ## send request and download result:
  dse_sh_process(input, output, evalscript, fl_landuse)

  ## read and plot result:
  landuse <- read_stars(fl_landuse) |> suppressWarnings()
  plot(landuse, rgb = 1:3, main = "Land use")
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
eval_chunks <-
  CopernicusDataspace::dse_has_client_info() &&
  CopernicusDataspace::dse_has_account() &&
  CopernicusDataspace::dse_has_s3_secret()

## ----odata-search, message=FALSE, eval=eval_chunks----------------------------
library(CopernicusDataspace)
library(dplyr) # Tidiverse package for data manipulation
library(sf)    # package for handling simple (spatial) features

## Define a bounding box for our search query:
bbox <-
  st_bbox(
    c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
    crs = 4326) |>
    st_as_sfc()

## Create an API request:
dse_odata_products_request() |>
  ## Filter content on ContentDate/Start`:
  filter(
    `ContentDate/Start` > "2025-01-01") |>
  ## Only data that intersects with our bounding box:
  st_intersects(bbox) |>
  ## Arrange by descending Id:
  arrange(desc(Id)) |>
  ## Only the first 5 hits:
  slice_head(n = 5) |>
  ## Collect the results:
  collect()

## ----odata-bursts, eval=FALSE-------------------------------------------------
# burst_req <-
#   dse_odata_bursts_request(ParentProductId == "879d445c-2c67-5b30-8589-b1f478904269")
# 
# ## Note that these are large files and may take a while to download:
# dse_odata_download(
#   burst_req,
#   tempdir()
# )

## ----odata-nodes, eval=eval_chunks--------------------------------------------
dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4",
                        recursive = TRUE)

## ----odata-quicklook, eval=FALSE----------------------------------------------
# dse_odata_quicklook(
#     "91822f33-b15c-5b60-aa39-6d9f6f5c773b",
#     tempfile(fileext = ".jpg"))

## ----odata-download-path, eval=FALSE------------------------------------------
# if (dse_has_account()) {
#   dse_odata_download_path(
#     product     = "2f497806-0101-5eea-83fa-c8f68bc56b0c",
#     node_path   =
#       paste("DEM1_SAR_DTE_90_20101213T034716_20130408T035028_ADS_000000_5033.DEM",
#             "Copernicus_DSM_30_S09_00_E026_00", "DEM",
#             "Copernicus_DSM_30_S09_00_E026_00_DEM.dt1", sep = "/"),
#     destination = tempdir()
#     )
# }

## ----download-odata, eval=FALSE-----------------------------------------------
# if (dse_has_s3_secret()) {
#   dse_odata_download(
#       dse_odata_products(
#         Name ==
#           "S1A_IW_OCN__2SDH_20250707T210608_20250707T210625_059983_07739B_893E.SAFE"),
#       destination = tempdir())
# }

#' @include geometry.R
NULL

#' Tidy Verbs for OData, SentinelHub and STAC API Requests
#' 
#' Implementation of tidy generics for features supported any of OData,
#' SentinelHub or STAC API requests.
#' They can be called on objects any of the classes: `odata_request`,
#' `sentinel_request` or `stac_request`.
#' The first is produced
#' by [dse_odata_products_request()] and [dse_odata_bursts_request()];
#' the second by [dse_sh_search_request()]; and the last
#' by [dse_stac_search_request()].
#' 
#' These special request class objects use lazy evaluation.
#' This means that functions are only evaluated after calling [dplyr::collect()]
#' on a request.
#' 
#' Note that you should not call the functions exported in this package directly.
#' Instead, call the generics as declared in the `dplyr` package. This is
#' illustrated by the examples.
#' 
#' ## Slice Head
#' In order to manage server traffic, the OData API never returns more than
#' 20 rows. If you want to obtain results beyond the first 20 rows, you need
#' to specify the `skip` argument.
#' 
#' The Sentinel and STAC API limits its results to the first 10 rows. You can expand that limit
#' with [dplyr::slice_head()]. For STAC the number of rows is capped at 10,000 records.
#' For SentinelHub this number is capped at 100.
#' 
#' ## Deviations
#' Due to limitations posed by the OData API, some tidyverse verbs deviate
#' from its tidy standards. Most notably:
#' 
#'  * [dplyr::select()]: Cannot change the order of columns. It will only
#'    affect which columns are selected. Also, tidy selection helpers like
#'    [dplyr::any_of()] and [dplyr::all_of()] are NOT supported
#'  * [dplyr::arrange()]: OData only allows to sort up to 32 columns.
#'    Adding more columns will produce a warning.
#'  * Grouping is not supported
#'  * Only tidy methods listed in the usage section are supported for the
#'    special request class objects. If you want to apply the full spectrum of
#'    tidyverse methods, call `dplyr::collect()` on the request class
#'    object first. That will return a normal `data.frame`, which can be
#'    manipulated further.
#' @param .data,x An object of any of the following classes `odata_request`,
#' `sentinel_request` or `stac_request`.
#' These are produced by
#' [dse_odata_products_request()], [dse_odata_bursts_request()],
#' [dse_sh_search_request()] and [dse_stac_search_request()]
#' @param n Maximum number of rows to return.
#' @param skip Number of rows to skip when collecting results. The APIs
#' return a limited number rows. Specify the number of rows to skip
#' in order to get results beyond the predefined limit.
#' @param by,.by,.by_group,.preserve,prop Arguments inherited from generic
#' `dplyr` functions. Ignored in the current context as either grouping
#' is not allowed for an OData API request or is otherwise not supported.
#' @param ... Data masking expressions, or arguments passed to embedded functions
#' @returns All functions (except `collect()`) return a modified
#' `stac_request`/`sentinel_request`/`odata_request`
#' object, containing the lazy tidy operations. `collect()` will return a
#' `data.frame()` yielding the result of the request.
#' @examples
#' library(dplyr)
#' if (interactive()) {
#'   dse_odata_products_request() |>
#'     filter(contains(Name, "WRR")) |>
#'     select("Id", "Name") |>
#'     arrange(Id, desc(Name)) |>
#'     slice_head(n = 5) |>
#'     collect()
#'
#'   dse_stac_search_request("sentinel-1-grd") |>
#'     filter(`sat:orbit_state` == "ascending") |>
#'     arrange("id") |>
#'     collect()
#'   
#'   if (dse_has_client_info()) {
#'     dse_sh_search_request(
#'       collection = "sentinel-2-l2a",
#'       bbox       = c(5.261, 52.680, 5.319, 52.715),
#'       datetime   = c("2025-01-01 UTC", "2025-01-31 UTC")
#'     ) |>
#'       filter(`eo:cloud_cover` <= 10) |>
#'       collect()
#'   }
#' }
#' @name tidy_verbs
#' @rdname tidy_verbs
NULL

#' @rdname tidy_verbs
#' @name filter
#' @family tidyverse
#' @export filter.odata_request
filter.odata_request <-
  function (.data, ..., .by = NULL, .preserve = FALSE) {
    new_filters <- rlang::enquos(..., .named = TRUE)
    .data$odata$filters <- c(.data$odata$filters, new_filters)
    .data
  }

.update_cql2_filter <- function(.data, ..., format) {
  old_filter <- .data$body$data$filter
  if (length(old_filter) == 1 && is.na(old_filter)) old_filter <- NULL
  new_filter <-
    .translate_filters(rlang::enquos(..., .named = TRUE)[[1]], format)
  if (!is.null(old_filter)) {
    new_filter <- list(
      args = list(old_filter, new_filter),
      op = "and"
    )
  }
  .data$body$data$filter <- new_filter
  .data
}

#' @rdname tidy_verbs
#' @name filter
#' @family tidyverse
#' @export filter.sentinel_request
filter.sentinel_request <-
  function (.data, ..., .by = NULL, .preserve = FALSE) {
    .update_cql2_filter(.data, ..., format = "sh")
  }

#' @rdname tidy_verbs
#' @name filter
#' @export filter.stac_request
filter.stac_request <-
  function (.data, ..., .by = NULL, .preserve = FALSE) {
    .update_cql2_filter(.data, ..., format = "stac")
  }

#' @rdname tidy_verbs
#' @name compute
#' @export compute.odata_request
compute.odata_request <-
  function(x, skip = 0L, ...) {
    x |>
      .parse_filters() |>
      .parse_arrange() |>
      .parse_slice() |>
      .parse_select() |>
      httr2::req_url_query(`$skip` = skip)
  }

#' @rdname tidy_verbs
#' @name collect
#' @family tidyverse
#' @export collect.odata_request
collect.odata_request <-
  function(x, skip = 0L, ...) {
    result <-
      x |>
      dplyr::compute(skip = skip) |>
      req_perform(...) |>
      httr2::resp_body_json()
    at <- result[setdiff(names(result), "value")]
    result <-
      result$value |>
      .simplify() |>
      dplyr::mutate(
        dplyr::across(dplyr::any_of(c("Assets", "Locations", "Attributes")), ~ {
          lapply(., .simplify)
        })
      )
    attributes(result) <- c(attributes(result), at)
    at <- c(at, meta_columns = dplyr::select(result, dplyr::starts_with("@odata")))
    result <- dplyr::select(result, !dplyr::starts_with("@odata"))
    result
  }

#' @rdname tidy_verbs
#' @name collect
#' @family tidyverse
#' @export collect.sentinel_request
collect.sentinel_request <-
  function(x, skip = 0L, ...) {

    if (skip > 0) x$body$data[["next"]] <- skip
    x |>
      httr2::req_perform() |>
      httr2::resp_body_json() |>
      .wrap_sh_features()
  }

#' @rdname tidy_verbs
#' @name collect
#' @family tidyverse
#' @export collect.stac_request
collect.stac_request <-
  function(x, ...) {
    items <- 
      httr2::req_perform(x) |>
      httr2::resp_body_json()
    .stac_items(items)
  }

#' @rdname tidy_verbs
#' @name arrange
#' @export arrange.odata_request
arrange.odata_request <-
  function(.data, ..., .by_group = FALSE) {
    new_arrange <- rlang::enquos(..., .named = TRUE)
    .data$odata$arrange <- c(.data$odata$arrange, new_arrange)
    .data
  }

#' @rdname tidy_verbs
#' @name arrange
#' @family tidyverse
#' @export arrange.stac_request
arrange.stac_request <-
  function(.data, ..., .by_group = FALSE) {
    my_arrange <-
      rlang::enquos(..., .named = TRUE) |>
      .column_select(allow_desc = TRUE)
    dsc <- attr(my_arrange, "is_desc")
    my_arrange <- lapply(seq_len(length(my_arrange)), \(i){
      list(field = my_arrange[[i]], direction = ifelse(dsc[[i]], "desc", "asc"))
    })
    
    if (is.na(.data$body$data$sortby)) .data$body$data$sortby <- NULL
    .data$body$sortby <- c(.data$body$data$sortby, my_arrange)
    .data
  }

.prop_error <- function(class) {
  rlang::abort(c(
    x = sprintf("'prop' argument is not implemented for %s requests", class),
    i = "Use 'n' instead."))
}

.check_slice <- function(sl) {
  if (!is.null(sl) && !is.na(sl))
    rlang::warn("Previously defined slice will be replaced with latest")
}

#' @rdname tidy_verbs
#' @name slice_head
#' @family tidyverse
#' @export slice_head.odata_request
slice_head.odata_request <-
  function(.data, ..., n, prop, by = NULL) {
    .check_slice(.data$odata$slice_head)
    if (!missing(prop)) .prop_error("OData")
    .data$odata$slice_head <- n
    .data
  }

#' @rdname tidy_verbs
#' @name slice_head
#' @export slice_head.stac_request
slice_head.stac_request <-
  function(.data, ..., n, prop, by = NULL) {
    if (!missing(prop)) .prop_error("STAC")
    .check_slice(.data$body$data$limit)
    .data$body$data$limit <- n
    .data
  }

#' @rdname tidy_verbs
#' @name slice_head
#' @family tidyverse
#' @export slice_head.sentinel_request
slice_head.sentinel_request <-
  function(.data, ..., n, prop, by = NULL) {
    if (!missing(prop)) .prop_error("Sentinel")
    .check_slice(.data$body$data$limit)
    .data$body$data$limit <- n
    .data
  }

#' @rdname tidy_verbs
#' @name select
#' @family tidyverse
#' @export select.odata_request
select.odata_request <-
  function(.data, ...) {
    new_select <- rlang::enquos(..., .named = TRUE)
    .data$odata$select <- c(.data$odata$select, new_select)
    .data
  }

.update_fields <- function(.data, ...) {
  new_select <-
    rlang::enquos(..., .named = TRUE) |>
    .column_select()
  if (is.null(.data$body$data$fields) || is.na(.data$body$data$fields))
    .data$body$data$fields <- NULL
  new_select <- c(.data$body$data$fields$include |> unlist(), new_select)
  .data$body$data$fields$include <- as.list(new_select)
  .data
}

#' @rdname tidy_verbs
#' @name select
#' @family tidyverse
#' @export select.stac_request
select.stac_request <-
  function(.data, ...) {
    .update_fields(.data, ...)
  }

#' @rdname tidy_verbs
#' @name select
#' @family tidyverse
#' @export select.sentinel_request
select.sentinel_request <-
  function(.data, ...) {
    .update_fields(.data, ...)
  }

.parse_filters <- function(.data) {
  filters <- .data$odata$filters
  geoms   <- .data$odata$geoms
  if (!is.null(filters)) {
    filters <-
      filters |>
      lapply(.translate_filters) |>
      unlist()
  }
  if (!is.null(geoms)) {
    geoms <-
      geoms |>
      lapply(.translate_geoms) |>
      unlist() |>
      sprintf(fmt = "OData.CSC.Intersects(area=geography'SRID=4326;%s')")
  }
  if (is.null(filters) && is.null(geoms)) return(.data)
  filters <- c(filters, geoms) |>
    paste(collapse = " and ")
  .data |>
    httr2::req_url_query(`$filter` = filters)
}

.parse_arrange <- function(.data) {
  my_arrange <- .data$odata$arrange
  if (is.null(my_arrange)) return(.data)
  my_arrange <-
    .column_select(my_arrange, allow_desc = TRUE)
  dsc <- attr(my_arrange, "is_desc")
  if (length(my_arrange) > 32L) {
    rlang::warn(c(
      "Cannot arrange OData by more than 32 columns",
      "Ignoring all but first 32 elements"),
      .frequency = "regularly",
      .frequency_id = "odata_arrange")
    my_arrange <- my_arrange[1:32]
  }
  my_arrange <-
    cbind(my_arrange, ifelse(dsc, " desc", "")) |>
    apply(1L, paste, collapse = "") |>
    paste(collapse = ",")
  .data |>
    httr2::req_url_query(`$orderby` = my_arrange)
}

.parse_slice <- function(.data) {
  my_slice <- .data$odata$slice_head
  if (is.null(my_slice)) return(.data)
  my_slice <- as.integer(my_slice)
  .data |>
    httr2::req_url_query(
      `$top` = my_slice
    )
}

.column_select <- function(q, allow_desc = FALSE) {

  result <-
    lapply(q, \(xpr) {
      xpr <- rlang::quo_get_expr(xpr)
      env <- if (rlang::is_quosure(xpr))
        rlang::quo_get_env(xpr) else
          environment()
      result <- NULL
      is_desc <- FALSE
      if (allow_desc && rlang::is_call(xpr) &&
          identical(rlang::eval_tidy(xpr[[1]], env = env), dplyr::desc)) {
        is_desc <- TRUE
        xpr <- rlang::as_quosure(xpr[[2]], env)
        env <- if (rlang::is_quosure(xpr))
          rlang::quo_get_env(xpr) else
            environment()
      }
      if (rlang::is_call(xpr)) {
        xpr1 <- if (rlang::is_quosure(xpr))
          rlang::quo_get_expr(xpr) else xpr[[1]]
        
        result <- if (length(xpr) > 2 &&
                      ((identical(rlang::eval_tidy(xpr1, env = env), `[[`)) ||
                      (identical(rlang::eval_tidy(xpr1, env = env), `$`)) &&
                      rlang::as_string(xpr[[2]]) == ".data")) {
          rlang::as_string(xpr[[3]])
        } else {
          if (rlang::is_quosure(xpr))
            rlang::as_string(rlang::quo_get_expr(xpr)) else
              rlang::eval_tidy(xpr)
        }
        if (is.null(result))
          stop(sprintf("Sorry, '%s' is not implemented in this context",
                       rlang::as_string(xpr[[1]])))
        attr(result, "is_desc") <- is_desc
        result
      } else {
        result <-
          if (rlang::is_string(xpr))
            rlang::eval_tidy(xpr) else
              rlang::as_string(xpr)
        attr(result, "is_desc") <- is_desc
        result
      }
    })
  attribs <- lapply(result, attr, "is_desc") |> unname() |> unlist()
  result <- unname(result) |> unlist()
  if (allow_desc) attr(result, "is_desc") <- attribs
  result
}

.parse_select <- function(.data) {
  my_select <- .data$odata$select
  if (is.null(my_select)) return(.data)
  
  sel <- .column_select(.data$odata$select)
  .data |>
    httr2::req_url_query(
      `$select` = paste(sel, collapse = ",")
    )
}

.odata_operators <- dplyr::tibble(
  r_code = list(`==`, `!=`, `>`, `>=`, `<`, `<=`, `&`, `|`, `!`, is.na, dplyr::contains),
  api_code = c("%s eq %s", "%s ne %s", "%s gt %s", "%s ge %s", "%s lt %s", "%s le %s",
               "%s and %s", "%s or %s", "not (%s)", "%s eq null", "contains(%s,%s)")
)

## STAC uses cq2-json schema.
.stac_operators <- dplyr::tibble(
  r_code = list(`==`, `!=`, `>`, `>=`, `<`, `<=`, `&`, `|`, `!`, is.na, `%in%`,
                dplyr::between),
  api_code = c("eq", "neq", "gt", "gte", "lt", "lte", "and", "or", "not",
               "isNull", "in", "between")
)

.sh_operators <- dplyr::tibble(
  r_code = list(`==`, `!=`, `>`, `>=`, `<`, `<=`, `&`),
  api_code = c("=", "<>", ">", ">=", "<", "<=", "and")
)

.match_function <- function(expr, format = "odata") {
  base <- if (format == "odata") .odata_operators else {
    if (format == "stac") .stac_operators else {
      if(format == "sh") .sh_operators
    }
  }
  result <- NA_integer_
  for (i in seq_len(nrow(base))) {
    if (identical(base$r_code[[i]], rlang::eval_tidy(expr))) {
      result <- i
      break
    }
  }
  return (list(idx = result, lookup = base))
}

.translate_filters <- function(quo, format = "odata") {
  expr   <- rlang::quo_get_expr(quo)
  env    <- rlang::quo_get_env(quo)
  idx    <- .match_function(expr[[1]], format)
  lookup <- idx$lookup
  idx    <- idx$idx
  op <- ""
  if (!is.na(idx)) {
    op <- lookup$api_code[idx]
    if (rlang::is_call(expr[[2]])) {
      left <- .translate_filters(rlang::as_quosure(expr[[2]], env), format)
      if (format == "odata") left <- sprintf("(%s)", left)
    } else {
      left <- as.character(expr[[2]])
    }
  }
  if (is.na(idx)) {
    
    ## ==== START: Check operator for special cases:
    
    if (identical(c, rlang::eval_tidy(expr[[1]]))) {
      result <- rlang::eval_tidy(expr, env)
      if (format == "odata") {
        if (is.character(result)) result <- sprintf("'%s'", result)
        return (paste(result, collapse = ","))
      } else if (format %in% c("stac", "sh")) {
        return (result)
      }
    } else if (identical(`%in%`, rlang::eval_tidy(expr[[1]])) &&
               format == "odata") {
      right <- rlang::eval_tidy(expr[[3]])
      if (is.character(right)) right <- sprintf("'%s'", right)
      paste(
        sprintf("%s eq %s", as.character(expr[[2]]), right),
        collapse = " or ") |>
        sprintf(fmt = "(%s)")
    } else if (identical(`$`, rlang::eval_tidy(expr[[1]])) ||
               identical(`[[`, rlang::eval_tidy(expr[[1]]))) {
      if (rlang::as_string(expr[[2]]) == ".data") {
        return(rlang::as_string(expr[[3]]))
      } else if (rlang::as_string(expr[[2]]) == ".env") {
        expr <- rlang::as_quosure(expr[[3]], env)
      } else {
        result <- rlang::eval_tidy(expr, env = env)
        if (format == "odata" && is.character(result))
          result <- sprintf("'%s'", result)
        result
      }
      
    ## ==== END: Check operator for special cases:

    } else {
      result <- rlang::eval_tidy(expr)
      if (format == "odata" && is.character(result))
        result <- sprintf("'%s'", result)
      result
    }
  } else {
    if (length(expr) < 3) {
      if (format %in% c("stac", "sh")) return( list( args = left, op = op ) )
      return(sprintf(op, left))
    } else {
      quote_right <- TRUE
      if (rlang::is_call(expr[[3]])) {
        right <- .translate_filters(
          rlang::as_quosure(expr[[3]], env), format)
        if (rlang::is_call(right)) right <- rlang::eval_tidy(right) else
          quote_right <- FALSE
      } else {
        right <- rlang::eval_tidy(expr[[3]], env = env)
      }
      left_check <- left
      if (is.list(left_check)) left_check <- left_check$args[[1]]$property
      if (is.null(left_check)) left_check <- ""
      if (grepl("date", left_check, ignore.case = TRUE) &&
          !is.list(right) && quote_right) {
        right <- lubridate::as_datetime(right)
        right <- lubridate::format_ISO8601(right, usetz = TRUE)
        if (format %in% c("sh", "stac")) right <- list(timestamp = right)
      }
      if (quote_right && is.character(right) && format == "odata")
        right <- sprintf("'%s'", right)
    }

    if (format == "odata") {
      return(sprintf(op, left, right))
    } else if (format %in% c("stac", "sh")) {
      if (is.character(left)) {
        left <- strsplit(left, "[.]")[[1]]
        domain <- "property"
        if (length(left) > 1) {
          if (left[[1]] != "properties") stop("Unknown domain")
          left <- left[[2]]
        }
        left <- structure(list(left), names = domain)
      }
      list(
        args = list(
          left,
          right
        ),
        op = op
      )
    }
  }
}

#' @include tidyverse.R
NULL

.odata_url <- "https://catalogue.dataspace.copernicus.eu/odata/v1"
.odata_s3_endpoint <- "https://eodata.dataspace.copernicus.eu"
.authent_url <- "https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token"

## Template used by roxygen for documentation:
.roxygen_account <- function() {
  paste(
    "## Creating an Account",
    "First step is creating an account. You can create one by visiting the",
    "[login page](https://identity.dataspace.copernicus.eu/auth/realms/CDSE/account/#/personal-info)",
    "and click \"register\". Fill out the form and follow the instructions.",
    sep = "\n"
  )
}

.dse_public_access_token <- function(username = dse_get_username(),
                                     password = dse_get_password()) {
  .authent_url |>
    httr2::request() |>
    httr2::req_headers(
      `Content-Type` = "application/x-www-form-urlencoded"
    ) |>
    httr2::req_body_form(
      username   = username,
      password   = password,
      grant_type = "password",
      client_id  = "cdse-public"
    ) |>
    httr2::req_method("POST") |>
    httr2::req_perform() |>
    httr2::resp_body_json()
}

.dse_access_token <- function(client_id     = dse_get_client_id(),
                              client_secret = dse_get_client_secret(), ...) {
  access_token <-
    .authent_url |>
    httr2::request() |>
    httr2::req_body_form(
      client_id = client_id,
      client_secret = client_secret,
      grant_type = "client_credentials"
    ) |>
    httr2::req_method("POST") |>
    httr2::req_perform() |>
    httr2::resp_body_json()
}

#' Client Information and Access Token for the Data Space Store API
#' 
#' To regulate server traffic, the Data Space Ecosystem uses user accounts to regulate
#' and ensure fair usage. These functions will get you a token and information about
#' your usage.
#' 
#' Before you can use most of the Data Space Ecosystem services, you need to create
#' an account and register as a client. This will let you retrieve an access token,
#' that can be used for authentication purposes in API requests. It is also used to
#' manage your usage and rate limiting (see also [dse_usage()] and
#' [dse_user_statistics()]).
#' 
#' Note that [Amazon Simple Storage Service (s3)](https://aws.amazon.com/s3/)
#' has separate authentication requirements. See [dse_s3_get_key()] for details.
#' 
#' `r .roxygen_account()`
#' 
#' ## Registering Client
#' In order to register a client, visit the
#' [Sentinel Dashboard](https://shapps.dataspace.copernicus.eu/dashboard/) and
#' go to "User settings". There you will have the option to create an OAuth client.
#' Follow the instructions, and make sure to safely copy the client id and secret.
#' The latter is only displayed once. More detailed instructions are provided by the
#' [API documentation](https://documentation.dataspace.copernicus.eu/APIs/SentinelHub/Overview/Authentication.html#registering-oauth-client)
#' 
#' ## Client Details as Environment Variables
#' When you share R code, you probably don't want to share your account details.
#' You can avoid using your `client_id` and `client_secret` in your script by
#' setting them as environment variable. You can do this yourself manually by
#' calling `dse_set_client_id()` and `dse_set_client_secret()` at the start of each session.
#' 
#' You can download OData simply through https with just you username and
#' password. You can also store those as environment variables for your convenience.
#' If you name those `CDSE_API_USERNAME` and `CDSE_API_PASSWORD` respectively,
#' they are picked up automatically with [dse_get_username()] and [dse_get_password()].
#' OData needs a public access token which is generated with
#' `dse_public_access_token()`, which needs your username and password.
#' 
#' You can also define them in your `.Rprofile` file with
#' `Sys.setenv(CDSE_API_CLIENTID = "<your id>")` and
#' `Sys.setenv(CDSE_API_CLIENTSECRET = "<your secret>")`. This way,
#' they are set each time you start a new R session.
#' 
#' The environment variables are used by default by `dse_access_token()`,
#' and `dse_public_access_token()` so you
#' don't have to specify the client details as arguments.
#' 
#' ## Obtain Token and Validity
#' After completing the previous two steps, your are now set to obtain an access
#' token with `dse_access_token()`.
#' 
#' Repeatedly requesting an access token may invoke rate limiting measures.
#' Therefore, this package uses caching to temporarily store the access token
#' during each R session. Calling `dse_access_token()` will therefore only
#' contact the server once for a token for each unique combination of `client_id`
#' and `client_server`. After that, the cached result will be reused during the session
#' 
#' There is a catch: the token provided by the server is only valid for a limited
#' time (usually 30 minutes). So, when the token has expired, you need to wipe the
#' cache. You can do so by calling `memoise::forget(dse_access_token)` or restarting
#' the R session.
#' @param client_id ID of the client registered under your account.
#' @param client_secret Secret provided for the client registered under your account.
#' @param username Your Copernicus Data Space username (usually your e-mail address).
#' @param password Your Copernicus Data Space password for your account.
#' @param value Assignment value for setting `username`, `password`, `client_id`
#' or `client_secret` as environment variable. Once set, it will persist for
#' the remainder of the R session.
#' @param ... Ignored
#' @seealso [dse_usage()]
#' @seealso [dse_user_statistics()]
#' @references <https://documentation.dataspace.copernicus.eu/APIs/SentinelHub/Overview/Authentication.html>
#' @returns In case of `dse_get_client_id()` and `dse_get_client_secret()`, you can get (or
#' set) client details as environment variables. This way, they will persist
#' throughout your R session.
#' 
#' The function `dse_has_client_info()` returns a `logical` value, indicating
#' whether client details (id and secret) are available as environmental variable.
#' Note that if this function returns `TRUE`, it doesn't guarantee that the details
#' are valid (just that they are available).
#' 
#' In case of `dse_access_token()` and `dse_public_access_token()`
#' a named `list` is returned, containing the
#' access token (named `"token"`) and some additional meta information.
#' @examples
#' if (interactive() && dse_has_client_info()) {
#'   token <- dse_access_token()
#' }
#' if (interactive() && dse_has_account()) {
#'   token_public <- dse_public_access_token()
#' }
#' @family authentication
#' @export
dse_access_token <- memoise::memoise(.dse_access_token) # Using memoise to avoid rate limiting errors

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_public_access_token <- memoise::memoise(.dse_public_access_token) # Using memoise to avoid rate limiting errors

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_get_client_id <- function(...) {
  Sys.getenv("CDSE_API_CLIENTID")
}

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_set_client_id <- function(value, ...) {
  Sys.setenv(CDSE_API_CLIENTID = as.character(value))
}

#' @rdname dse_access_token
#' @export
dse_set_username <- function(value, ...) {
  Sys.setenv(CDSE_API_USERNAME = as.character(value))
}

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_set_password <- function(value, ...) {
  Sys.setenv(CDSE_API_PASSWORD = as.character(value))
}

#' @rdname dse_access_token
#' @export
dse_get_client_secret <- function(...) {
  Sys.getenv("CDSE_API_CLIENTSECRET")
}

#' @rdname dse_access_token
#' @export
dse_get_username <- function(...) {
  Sys.getenv("CDSE_API_USERNAME")
}

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_get_password <- function(...) {
  Sys.getenv("CDSE_API_PASSWORD")
}

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_set_client_secret <- function(value, ...) {
  Sys.setenv(CDSE_API_CLIENTSECRET = as.character(value))
}

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_has_client_info <- function(...) {
  dse_get_client_id() != "" && dse_get_client_secret() != ""
}

#' @rdname dse_access_token
#' @family authentication
#' @export
dse_has_account <- function(...) {
  dse_get_password() != "" && dse_get_username() != ""
}

#' Setup Amazon Simple Storage Service for the Data Space Ecosystem
#' 
#' using [Amazon Simple Storage Service (s3)](https://aws.amazon.com/s3/) in the
#' Data Space Ecosystem requires a key and secret. These functions help you managing
#' these details and setting up an s3 client.
#' 
#' Working with s3 in the Data Space Ecosystem requires you to create an account,
#' then register an s3 key as described below. Note that the SentinelHub requires
#' a different authentication method. See [dse_access_token()] for more details on that.
#' 
#' `r .roxygen_account()`
#' 
#' ## Registering a s3 Key
#' Now that you have an account, you should visit the
#' [s3-credentials page](https://eodata-s3keysmanager.dataspace.copernicus.eu/panel/s3-credentials),
#' and log in with your account details. By clicking "add credential", you can
#' create a new key and secret. Store them in a safe place, as the secret is only
#' shown once. You can pass the key and secret as `s3_key` and `s3_secret` arguments
#' to functions requesting them. You can also store them as environment variables
#' such that they persist throughout the R session and don't have to be passed as
#' arguments (see below).
#' 
#' ## S3 Key and Secret as Environment Variables
#' When you share R code, you probably don't want to share your account details.
#' You can avoid using your `s3_key` and `s3_secret` in your script by
#' setting them as environment variable. You can do this yourself manually by
#' calling `dse_s3_set_key()` and `dse_s3_set_secret()` at the start of each session.
#' 
#' You can also define them in your `.Rprofile` file with
#' `Sys.setenv(CDSE_API_S3ID = "<your key>")` and
#' `Sys.setenv(CDSE_API_S3SECRET = "<your secret>")`. This way,
#' they are set each time you start a new R session.
#' @references <https://documentation.dataspace.copernicus.eu/APIs/S3.html>
#' @param ... Ignored
#' @param value Replacement value for the `s3_key` or `s3_secret`.
#' @param region [AWS Region](https://aws.amazon.com/about-aws/global-infrastructure/regions_az/)
#' used in instantiating the client
#' @param s3_key,s3_secret The s3 key and secret registered under your Data Space
#' Ecosystem account
#' @returns
#' [dse_s3_get_key()] and [dse_s3_get_secret()] will return the requested s3 details
#' if set as environment variable (see details).
#' 
#' [dse_has_s3_secret()] returns a logical value indicating whether s3 details
#' (key and secret) are set. It will not determine whether the details are valid.
#' @rdname dse_s3
#' @family authentication
#' @family s3
#' @export
dse_has_s3_secret <- function() {
  dse_s3_get_key() != "" && dse_s3_get_secret() != ""
}

#' @rdname dse_s3
#' @family authentication
#' @family s3
#' @export
dse_s3_get_key <- function(...) {
  Sys.getenv("CDSE_API_S3ID")
}

#' @rdname dse_s3
#' @family authentication
#' @family s3
#' @export
dse_s3_set_key <- function(value, ...) {
  Sys.setenv(CDSE_API_S3ID = as.character(value))
}

#' @rdname dse_s3
#' @family authentication
#' @family s3
#' @export
dse_s3_get_secret <- function(...) {
  Sys.getenv("CDSE_API_S3SECRET")
}

#' @rdname dse_s3
#' @family authentication
#' @family s3
#' @export
dse_s3_set_secret <- function(value, ...) {
  Sys.setenv(CDSE_API_S3SECRET = as.character(value))
}

#' Set Copernicus Data Space Ecosystem Access Token for GDAL Driver
#' 
#' This function sets system environment variables, such
#' that the GDAL library can access the Copernicus Data Space
#' Ecosystem https storage. Note that these settings can be used
#' by any package depending on the GDAL library. Most notably:
#' `stars`, `terra`, and `gdalraster`.
#' @inheritParams dse_usage
#' @returns Returns a `logical` value. `TRUE` if all variables
#' were successfully set. `FALSE` otherwise.
#' @examples
#' if (interactive() && dse_has_client_info() &&
#'     requireNamespace("stars")) {
#'   uri <-
#'     dse_stac_get_uri(
#'       "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
#'       "B01", type = "odata")
#'    
#'   dse_set_gdal_token()
#'
#'   ## As this URI is zipped, it need to be downloaded.
#'   ## But you can access it directly:
#'   jp2 <- stars::read_stars(uri)
#' }
#' @family authentication
#' @export
dse_set_gdal_token  <- function(token = dse_public_access_token()) {
  Sys.setenv(GDAL_HTTP_AUTH = "BEARER") &&
    Sys.setenv(GDAL_HTTP_BEARER = token$access_token)
}

#' Decode Access Token
#' 
#' This function decodes an access token and returns a named `list` with
#' information from the token.
#' @param token A token obtained with [dse_access_token()] or
#' [dse_public_access_token()].
#' @returns A named `list` with token info
#' @examples
#' if (interactive() && dse_has_client_info()) {
#'   dse_get_token_details()
#' }
#' @family authentication
#' @export
dse_get_token_details <- function(token = dse_access_token()) {
  if (requireNamespace("jose")) {
    jose::jwt_split(token$access_token)
  } else {
    rlang::abort(c(
      x = "Namespace 'jose' not found",
      i = "Install package 'jose' and try again"
    ))
  }
}

library(dplyr) |> suppressMessages()
library(stars) |> suppressMessages()

test_that("STAC collections are obtained", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    col <- dse_stac_collections("sentinel-2-l2a")
    col$id == "sentinel-2-l2a"
  })
})

test_that("STAC client info is obtained", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    cl <- dse_stac_client()
    cl$id == "cdse-stac"
  })
})

test_that("STAC queryables are obtained", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    qr <- dse_stac_queryables("sentinel-1-grd")
    qr$properties$id$type == "string"
  })
})

test_that("STAC collections can be searched", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    result <-
      dse_stac_search_request("sentinel-1-grd") |>
      filter(`sat:orbit_state` == "ascending") |>
      arrange("id") |>
      collect()
    all(result$`properties.sat:orbit_state` == "ascending")
  })
})

test_that("Files can be downloaded via STAC S3", {
  skip_if_offline()
  skip_if_not(dse_has_s3_secret())
  expect_no_error({
    fn <-
      dse_stac_download(
        asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
        asset = "B01",
        destination = tempdir()
      )
    tile <- read_stars(fn)
    NULL
  })
})

test_that("Files can be downloaded via STAC https", {
  skip_if_offline()
  skip_if_not(dse_has_account())
  expect_no_error({
    fn <-
      dse_stac_download(
        asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
        asset = "B01",
        s3_key = "", s3_secret = "",
        destination = tempdir()
      )
    tile <- read_stars(fn)
  })
})

test_that("req_perform can also handle ordinary requests", {
  skip_if_offline()
  skip_on_cran()
  expect_identical({
    "https://stac.dataspace.copernicus.eu/v1/_mgmt/ping" |>
      httr2::request() |>
      CopernicusDataspace::req_perform() |>
      httr2::resp_body_json()
  }, list(message = "PONG"))
})

test_that("Access token can be obtained", {
  skip_if_offline()
  skip_if_not(dse_has_client_info())
  expect_no_error({
    token <- dse_access_token()
  })
})

test_that("Public access token can be obtained", {
  skip_if_offline()
  skip_if_not(dse_has_account())
  expect_no_error({
    token <- dse_public_access_token()
  })
})

test_that("Roxygen function produces text", {
  testthat::expect_type({
    CopernicusDataspace:::.roxygen_account()
  }, "character")
})

test_that("Client usage is reported correctly", {
  skip_if_offline()
  skip_if_not(dse_has_client_info())
  expect_true({
    usg <- dse_usage()
    is.data.frame(usg) && all(c("consumed", "remaining") %in% names(usg))
  })
})

test_that("User statistics are reported", {
  skip_if_offline()
  skip_if_not(dse_has_client_info())
  expect_true({
    ustats <- dse_user_statistics()
    is.data.frame(ustats) && all(c("API", "CATALOG") %in% names(ustats))
  })
})

test_that("GDAL token can be set", {
  skip_if_not(dse_has_client_info())
  expect_true({
    dse_set_gdal_token()
  })
})

test_that("Token can be decrypted", {
  skip_if_offline()
  skip_if_not(dse_has_client_info())
  skip_if_not_installed("jose")
  expect_equal({
    token_info <-
      dse_access_token() |>
      dse_get_token_details()
    token_info$header$alg
  }, "RS256")
})

test_that("Secrets can be set", {
  skip_if_offline()
  skip_if_not(dse_has_account())
  skip_if_not(dse_has_client_info())
  skip_if_not(dse_has_s3_secret())
  expect_no_error({
    dse_set_username(dse_get_username())
    dse_set_password(dse_get_password())
    dse_set_client_id(dse_get_client_id())
    dse_set_client_secret(dse_get_client_secret())
    dse_s3_set_key(dse_s3_get_key())
    dse_s3_set_secret(dse_s3_get_secret())
  })
})

library(dplyr) |> suppressMessages()
library(lubridate) |> suppressMessages() |> suppressWarnings()

test_that("Complex stac request works with tidy verbs", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    d1 <- "2025-01-01 UTC"
    d2 <- "2025-01-02 UTC"
    n <- 5
    result <-
      dse_stac_search_request("sentinel-1-grd") |>
      ## The order of the filters can break the request.
      ## I think this is a problem caused by the parser on the
      ## server. Check with Copernicus
      filter(datetime >= d1 & datetime <= d2) |>
      filter(`sat:orbit_state` == "ascending") |>
      arrange(desc("datetime")) |>
      slice_head(n = n) |>
      select(.data$`properties.sat:orbit_state`, .data$properties.datetime) |>
      select(c("properties.constellation", "properties.view:incidence_angle")) |>
      collect()
    all(as_datetime(result$properties.datetime) >=
          as_datetime(d1)) &&
      all(as_datetime(result$properties.datetime) <=
            as_datetime(d2)) &&
      all(result$`properties.sat:orbit_state` == "ascending") &&
      nrow(result) == n
  })
})

test_that("Complex stac request works with tidy verbs", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    d1 <- "2025-01-01 UTC"
    d2 <- "2025-01-02 UTC"
    n <- 5
    result <-
      dse_odata_products_request() |>
      filter(`ContentDate/Start` >= d1 & `ContentDate/Start` <= d2) |>
      filter(Online == TRUE) |>
      arrange(desc("ContentDate/Start")) |>
      slice_head(n = n) |>
      select(.data$Id, .data$Name) |>
      select(c("ContentDate/Start", "Online")) |>
      collect()
    
    all(as_datetime(result$ContentDate.Start) >=
          as_datetime(d1)) &&
      all(as_datetime(result$ContentDate.Start) <=
            as_datetime(d2)) &&
      all(result$Online == "TRUE") &&
      nrow(result) == n
  })
})

test_that("Tidyverse operators work on sentinel_request", {
  skip_if_offline()
  skip_if_not(dse_has_client_info())
  expect_true({
    result <-
      dse_sh_search_request(
        collection = "sentinel-2-l2a",
        bbox       = c(5.261, 52.680, 5.319, 52.715),
        datetime   = c("2020-01-01 UTC", "2025-01-31 UTC")
      ) |>
      filter(`eo:cloud_cover` <= 10) |>
      slice_head(n = 5) |>
      ## Note that `select` is always merged with
      ## ["id", "type", "geometry", "bbox", "links", "assets", "properties.datetime"]
      ## as per API specs
      select(`properties.eo:cloud_cover`,
             `properties.proj:geometry.type`) |>
      collect()
    nrow(result) == 5 &&
      all(as.numeric(result$`properties.eo:cloud_cover`) <= 10)
  })
})

test_that("Specifying multiple slices, produces warning", {
  skip_if_offline()
  skip_if_not(dse_has_client_info())
  expect_warning({
    dse_sh_search_request(
      collection = "sentinel-2-l2a",
      bbox       = c(5.261, 52.680, 5.319, 52.715),
      datetime   = c("2020-01-01 UTC", "2025-01-31 UTC")
    ) |>
      slice_head(n=5) |>
      slice_head(n=3)
  }, "Previously")
})

#' @include init.R
NULL

#' Filter OData and STAC Requests Using Geometries
#' 
#' Filters OData and STAC rows that intersect with spatial feature `y`.
#' @param x Either an `odata_request` class object, generated with
#' [dse_odata_products_request()]/[dse_odata_bursts_request()]; or
#' a `stac_request` generated with [dse_stac_search_request()].
#' @param y A spatial geometry of either class `sf` (see [sf::st_as_sf()]) or
#' `sfc` (see [sf::st_as_sfc()]). It
#' will always be transformed to WGS 84 projection (EPSG:4326).
#' @param sparse Argument inherited from generic definition. Ignored in this context
#' @param ... Ignored
#' @returns Returns an object of the same class as `x`, with the geometry filter added
#' to it.
#' @examples
#' if (interactive()) {
#'   bbox <-
#'     sf::st_bbox(
#'       c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
#'       crs = 4326) |>
#'       sf::st_as_sfc()
#' 
#'   dse_odata_products_request() |>
#'     dplyr::filter(
#'       `ContentDate/Start` > "2025-01-01") |>
#'     sf::st_intersects(bbox) |>
#'     dplyr::collect()
#'
#'   dse_stac_search_request("sentinel-2-l1c") |>
#'     sf::st_intersects(bbox) |>
#'     dplyr::collect()
#' }
#' @rdname geometry
#' @name st_intersects
#' @family geometry
#' @export st_intersects.odata_request
st_intersects.odata_request <-
  function (x, y, sparse = FALSE, ...) {
    x$odata$geoms <- c(x$odata$geoms, list(y))
    x
  }

.translate_geoms <- function(geom) {
  if (!inherits(geom, "sfc"))
    geom <- sf::st_as_sfc(geom)
  geom |>
    sf::st_transform(4326) |>
    sf::st_geometry() |>
    sf::st_as_text()
}

#' @rdname geometry
#' @name st_intersects
#' @family geometry
#' @export st_intersects.stac_request
st_intersects.stac_request <-
  function (x, y, sparse = FALSE, ...) {
    cur_bbox <- x$body$data$bbox
    cur_geom <- x$body$data$intersects

    ## Other projections don't seem to work well, so transform:
    y <- sf::st_transform(y, 4326)

    w <- \()
      rlang::warn("Geometry and bbox are mutually exclusive. Removing defined geometry")
    
    if (inherits(y, "bbox")) {
      
      if (!(is.null(cur_bbox) || all(is.na(cur_bbox))))
        rlang::warn("Replacing previously defined bbox")
      
      x$body$data$bbox <- as.list(as.numeric(y))
      
      if (!(is.null(cur_geom) || all(is.na(cur_geom)))) {
        x$body$data$intersects <- NULL
        w()
      }
      
    } else {
      if (!(is.null(cur_geom) || all(is.na(cur_geom))))
        rlang::warn("Replacing previously defined geometry")

      ## Make sure to summarise all features to one Geometry(Collection)
      y <- dplyr::summarise(sf::st_as_sf(y)) |> sf::st_geometry()
      
      yjson <- .sfc_to_geojson(y)
      geom  <- yjson$features[[1]]$geometry
      
      x$body$data$intersects <- geom
      
      if (!(is.null(cur_bbox) || all(is.na(cur_bbox)))) {
        x$body$data$bbox <- NULL
        w()
      }

    }
    
    x
  }

.wrap_sh_features <- function(features) {
  links <- features$links |> .simplify()
  features <- features$features |> .simplify()
  attr(features, "links") <- links
  features
}

.simplify <- function(x, do_unnest = TRUE, what = "") {
  if (any(c("Product", "product") %in% names(x))) {
    nm <- names(x)
    nm[nm == "Product"] <- "product"
    names(x) <- nm
  }
  
  if (rlang::is_named(x)) {
    result <- tibble::enframe(x) |>
      tidyr::pivot_wider(names_from = "name", values_from = "value")
  } else {
    is_unnamed <- length(x) > 0 && all(!(lapply(x, rlang::is_named) |> unlist()))
    if (is_unnamed) {
      return(list(lapply(x, .simplify)))
    } else {
      result <-
        lapply(x, \(y) {
          tibble::enframe(y) |>
            tidyr::pivot_wider(names_from = "name", values_from = "value")
        }) |>
        dplyr::bind_rows()
    }
  }
  result <- result |>
    dplyr::mutate(
      dplyr::across(dplyr::everything(), ~ {
        if (lapply(.x, is.list) |> unlist() |> all()) {
          if (lapply(.x, rlang::is_named) |> unlist() |> all()) {
            lapply(.x, .simplify, what = dplyr::cur_column())
          } else {
            .x
          }
        } else {
          if (all(lengths(.x) == 1))
            unlist(.x) else
              .x
        }
      })
    )
  if (do_unnest) {
    tibble_columns <-
      lapply(result, \(y) lapply(y, tibble::is_tibble) |> unlist() |> all()) |> unlist()
    tibble_columns[names(tibble_columns) %in%
                     c("assets", "properties.cube:variables",
                       "geometry", "summaries")] <- FALSE
    if (any(tibble_columns)) {
      tidyr::unnest(result, names(tibble_columns)[tibble_columns], names_sep = ".")
    } else result
  } else result
}

.simplify2 <- function(x) {
  lapply(x, \(y) {
    if (length(y) != 1) list(y) else y
  }) |>
    tibble::as_tibble()
}

.add_token <- function(x, token) {
  x |>
    httr2::req_headers(
      authorization = paste("Bearer", token$access_token)
    )
}

.stac_items <- function(x) {
  attrib_names <- c("type", "links", "numberReturned")
  attribs <- x[attrib_names]
  x <- (x[setdiff(names(x), attrib_names)][[1]]) |>
    .simplify()
  .x <- NULL
  simplify_sub <- function(y, i) {
    if (i %in% names(y)) {
      dplyr::bind_cols(y |> dplyr::select(-.env$i),
                       y[[i]] |> .simplify() |>
                         dplyr::rename_with(~paste(i, .x, sep = ".")))
    } else y
  }

  x <-
    x |>
    simplify_sub("properties")
  
  attributes(x) <- c(attributes(x), x)
  x
  
}

## Use GDAL driver to convert sfc class object to geojson:
.sfc_to_geojson <- function(x) {
  tempfn <- tempfile(fileext = ".geojson")
  on.exit({ unlink(tempfn) })
  sf::st_write(x, tempfn, driver = "GeoJSON", quiet = TRUE)
  yjson <- jsonlite::read_json(tempfn)
}

#' Perform a Request to Get a Response
#' 
#' A wrapper around [httr2::req_perform()], which can also handle
#' `odata_request` class objects. Check [httr2::req_perform()]
#' for details.
#' @param req Either a [httr2::request()] class object or an `odata_request`
#' class object. The latter can be created with [dse_odata_products_request()] and
#' [dse_odata_bursts_request()].
#' @inheritParams httr2::req_perform
#' @returns Returns a `httr2::response` class object
#' @family tidyverse
#' @export
req_perform <- function(
    req,
    path = NULL,
    verbosity = NULL,
    mock = getOption("httr2_mock", NULL),
    error_call = rlang::current_env()) {
  
  if (inherits(req, "odata_request")) {
    
    req |>
      httr2::req_error(body = .odata_error) |>
      httr2::req_perform(path = path, verbosity = verbosity,
                         mock = mock, error_call = error_call)
    
  } else {
    
    httr2::req_perform(req = req, path = path, verbosity = verbosity,
                       mock = mock, error_call = error_call)
    
  }
}

#' @include login.R

#' @rdname dse_s3
#' @export
dse_s3 <- function(region    = "us-east-1", ...,
                   s3_key    = dse_s3_get_key(),
                   s3_secret = dse_s3_get_secret()) {
  paws::s3(
    credentials = list(
      creds = list(
        access_key_id     = s3_key,
        secret_access_key = s3_secret
      )
    ),
    endpoint = .odata_s3_endpoint,
    region   = region
  )
}

.download_s3 <- function(s3_path, destination, ps3) {
  prefix  <- gsub("^/eodata/", "", s3_path)
  objects <- ps3$list_objects("eodata", Prefix = prefix)
  keys    <- lapply(objects$Contents, `[[`, "Key") |> unlist()
  lapply(keys, \(k) {
    fn   <- gsub(paste0(dirname(prefix), "[/]"), "", k)
    dest <- file.path(destination, fn)
    if (!dir.exists(dirname(dest))) {
      dirresult <- dir.create(dirname(dest), recursive = TRUE)
      if (!dirresult) stop("Failed to create subdirectory for download file")
    }
    ps3$download_file("eodata", k, dest)
    return(fn)
  })
}

#' Download Asset Through Uniform Resource Identifier
#' 
#' When the Uniform Resource Identifier (URI, starting with "s3://") for
#' an asset is known, this function can be used to download it
#' @param uri A Uniform Resource Identifier (URI, starting with "s3://").
#' You can look for them in the STAC catalogue, either using a
#' [web browser](https://browser.stac.dataspace.copernicus.eu/) or
#' [dse_stac_search_request()] (see example).
#' @param destination Destination path to a directory where to store the downloaded file(s)
#' @param ... Ignored
#' @inheritParams dse_s3
#' @returns A vector of file names stored at `destination`
#' @examples
#' if (interactive() && dse_has_s3_secret()) {
#'   library(dplyr)
#'
#'   ## Retrieve a URI for a specific asset through the STAC
#'   ## catalogue:   
#'   my_uri <-
#'     dse_stac_search_request(
#'       ids = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148") |>
#'         select("assets.B01.href") |>
#'         arrange("id") |>
#'         collect() |>
#'         pull("assets") |>
#'         unlist()
#'   
#'   dse_s3_download(my_uri, tempdir())
#'
#' }
#' @family s3
#' @export
dse_s3_download <- function(
    uri, destination, ..., s3_key = dse_s3_get_key(), s3_secret = dse_s3_get_secret()) {
  if (!grepl("^[s|S]3://", uri, ignore.case = TRUE))
    rlang::abort(c(
      x = "Not a valid S3 URI",
      i = "Make sure the path starts with 's3://'"
    ))
  uri <- gsub("^[s|S]3\\:\\/", "", uri)
  .download_s3(uri, destination,
               dse_s3(s3_key = s3_key, s3_secret = s3_secret))
}

#' Convert Uniform Resource Identifier to Virtual System Identifier
#' 
#' Convert Uniform Resource Identifier (URI) to Virtual System
#' Identifier (VSI). The Copernicus Data Space Ecosystem returns
#' URIs for accessing assets. Packages that use the GDAL library
#' (e.g., `stars`, `terra` and `gdalraster`)
#' can use VSI to access raster data directly. Use this function
#' to convert such an URI to a VIS.
#' @param uri A Uniform Resource Identifier, pointing to an
#' S3 storage file. You can retrieve one with [dse_stac_get_uri()].
#' @param streaming A `logical` value that allows to toggle between
#' `"\\vsis3\\"` and `"\\vsis3_streaming\\"` (default). The latter is
#' faster for reading files from its resource, but does not allow
#' random access. The first supports random access, but is not as fast
#' at reading.
#' @returns A `character` string representing the VSI
#' @examples
#' if (interactive()) {
#'   dse_stac_get_uri(
#'     "S1A_IW_GRDH_1SDV_20241125T055820_20241125T055845_056707_06F55C_12F9_COG",
#'     "vh", "sentinel-1-grd") |>
#'     dse_s3_uri_to_vsi()
#' }
#' @family s3
#' @export
dse_s3_uri_to_vsi <- function(uri, streaming = TRUE) {
  vsi <- sprintf(
    "/vsis3%s/",
    ifelse(streaming, "_streaming", ""))
  gsub("s3://", vsi, uri)
}

#' Set-up S3 Configuration for GDAL Library
#' 
#' This function sets system environment variables, such
#' that the GDAL library can access the Copernicus Data Space
#' Ecosystem S3 storage. Note that these settings can be used
#' by any package depending on the GDAL library. Most notably:
#' `stars`, `terra`, and `gdalraster`.
#' @param ... Ignored
#' @param region [AWS Region](https://aws.amazon.com/about-aws/global-infrastructure/regions_az/)
#' used in instantiating the S3 client
#' @inheritParams dse_s3_download
#' @returns Returns a `logical` value. `TRUE` if all variables
#' were successfully set. `FALSE` otherwise.
#' @examples
#' if (interactive() && dse_has_s3_secret() &&
#'     requireNamespace("stars")) {
#'   library(dplyr)
#'
#'   ## Get a Virtual System Interface to a tiff file:
#'   vsi <-
#'     dse_stac_get_uri(
#'       "S1A_IW_GRDH_1SDV_20241125T055820_20241125T055845_056707_06F55C_12F9_COG",
#'       "vh", "sentinel-1-grd") |>
#'       dse_s3_uri_to_vsi()
#' 
#'   ## Make sure to set gdal options with required S3 settings
#'   dse_s3_set_gdal_options()
#'    
#'   ## You can now read the file directly from the online storage
#'   ## without having to download it completely:
#'   cog <- stars::read_stars(vsi)
#'    
#'   ## You can also easily plot a downsampled version
#'   plot(cog, downsample = 50)
#' }
#' @family s3
#' @export
dse_s3_set_gdal_options <- function(
    region = "us-east-1", ...,
    s3_key = dse_s3_get_key(), s3_secret = dse_s3_get_secret()) {
  Sys.setenv(AWS_REGION            = region) &&
    Sys.setenv(AWS_ACCESS_KEY_ID     = s3_key) &&
    Sys.setenv(AWS_SECRET_ACCESS_KEY = s3_secret) &&
    Sys.setenv(AWS_VIRTUAL_HOSTING   = "FALSE") &&
    Sys.setenv(AWS_S3_ENDPOINT       =
                 gsub("^https://", "", .odata_s3_endpoint))
}

test_that("URI to s3 works", {
  expect_true({
    startsWith(
      dse_s3_uri_to_vsi("s3://"),
      "/vsis3"
    )
  })
})

test_that("GDAL s3 authentication can be set up", {
  skip_if_not(dse_has_s3_secret())
  expect_true({
    dse_s3_set_gdal_options()
  })
})

#' @include helpers.R
NULL

.sh_api_url     <- "https://sh.dataspace.copernicus.eu/api/v1"
.accounting_url <- paste0(.sh_api_url, "/accounting")

#' Get Data Space Account Information
#' 
#' In order to guarantee good performance for all users, the Sentinel Hub applies
#' [rate limiting](https://docs.sentinel-hub.com/api/latest/api/overview/rate-limiting/).
#' This policy enforces monthly quotas to your usage. To check your quota and current
#' usage, you can call `dse_usage()` or `dse_user_statistics()`.
#' @param range Specify a time range for which to obtain user statistics.
#' The API expects a string starting with a capitalised time unit (`"DAYS"`, `"HOURS"`),
#' followed by a dash ("-") and an integer value specifying the length of the period.
#' Default is `"DAYS-31"`.
#' @param resolution Specifying a temporal resolution for the user statistics.
#' should be one of `"DAILY"` (default), `"MONTHLY"`, or `"HOURLY"`.
#' @param ... Ignored
#' @param token For authentication, many of the Data Space Ecosystem uses
#' an access token. Either provide your access token, or obtain one automatically
#' with [dse_access_token()] (default). Without a valid token you will likely get
#' an "access denied" error.
#' @returns A `data.frame` with requested information for the user associated with
#' the provided `token`.
#' @examples
#' if (interactive() && dse_has_client_info()) {
#'   dse_usage()
#'   dse_user_statistics()
#' }
#' @seealso [dse_access_token()]
#' @family authentication
#' @export
dse_usage <- function(..., token = dse_access_token()) {
  result <-
    .accounting_url |>
    paste("usage", sep = "/") |>
    httr2::request() |>
    .add_token(token) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  nm <- names(result)
  result |>
    .simplify() |>
    tidyr::pivot_longer(dplyr::everything(), names_sep = "[.]",
                        names_to = c("qualifier", "name")) |>
    tidyr::pivot_wider()
}

#' @rdname dse_usage
#' @family authentication
#' @export
dse_user_statistics <- function(
    range = "DAYS-31",
    resolution = "DAILY",
    token = dse_public_access_token()) {
  
  .res_args   <- c("HOURLY", "DAILY", "MONTHLY")
  .range_args <- c("SECONDS", "MINUTES", "HOURS", "DAYS")
  resolution  <- match.arg(toupper(resolution), .res_args)
  range       <- toupper(range)
  range_check <- strsplit(range, "-")[[1]]
  as.integer(range_check[[2]]) ## produces warnings/errors if second part isn't integer
  range_check <- match.arg(range_check[[1]], .range_args)
  
  result <-
    .accounting_url |>
    paste("statistics/requests/statistics", sep = "/") |>
    httr2::request() |>
    httr2::req_url_query(range = range, resolution = resolution) |>
    .add_token(token) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  result$data |>
    lapply(.simplify) |>
    tibble::enframe() |>
    tidyr::pivot_wider()
}

#' @include tidyverse.R
#' @include login.R
#' @include odata_products.R
NULL

#' @rdname dse_odata_bursts
#' @family odata
#' @export
dse_odata_bursts_request <- function(...) {
  result <-
    .odata_url |>
    paste("Bursts", sep = "/") |>
    httr2::request()
  class(result) <- union(c("odata_request", "odata_bursts_request"), class(result))
  result
}

#' Create a OData Request for a Data Space Ecosystem Bursts Data
#' 
#' Obtain metadata for burst data associated with specific products.
#' 
#' For more details about bursts check the
#' [burst API documentation](https://documentation.dataspace.copernicus.eu/APIs/Sentinel-1%20SLC%20Burst.html).
#' 
#' You can apply some tidyverse functions (see [tidy_verbs]) to `odata_request`
#' object returned by [dse_odata_bursts_request()]. These apply lazy evaluation.
#' Meaning that they are just added to the object and are only evaluated after
#' calling either [dplyr::compute()] or [dplyr::collect()] (see examples).
#' @param ... Ignored
#' @returns Returns a `data.frame` with burst information.
#' @examples
#' if (interactive() && dse_has_s3_secret()) {
#'   dse_odata_bursts(ParentProductId == "879d445c-2c67-5b30-8589-b1f478904269")
#'   
#'   burst_req <-
#'     dse_odata_bursts_request(ParentProductId == "879d445c-2c67-5b30-8589-b1f478904269")
#'   
#'   ## Note that these are large files and may take a while to download:
#'   dse_odata_download(
#'     burst_req,
#'     tempdir()
#'   )
#' }
#' @family odata
#' @export
dse_odata_bursts <- function(...) {
  dse_odata_bursts_request() |>
    dplyr::filter(...) |>
    dplyr::collect()
}

#' @include tidyverse.R
#' @include login.R
#' @include account.R
#' @include req_perform.R
#' @include s3.R
NULL

.odata_error <- function(resp) {
  if (resp$headers$`Content-Type` == "application/json") {
    body <- resp |> httr2::resp_body_json()
    result <- body$detail
  } else {
    result <- NULL
  }

  if (is.null(result)) {
    if (resp$status_code == 401) {
      c("Ensure that you have specified a token",
        "Ensure your token hasn't expired (call `memoise::forget(dse_access_token)`)")
    } else {
      "Details not returned"
    }
  } else {
    if (rlang::is_named(result)) result$message else as.character(result)
  }
}

.np_error <- function(node_path) {
  if (length(node_path) != 1)
    rlang::abort(c(
      x = "Argument `node_path` should only have one element",
      i = "Pass `node_path` elements in a loop"))
}

#' Create a OData Request for a Data Space Ecosystem Product
#' 
#' OData is an application programming interface (API) used to disseminate Copernicus
#' Data Space Ecosystem products. This function creates a request for this service,
#' which can be used to obtain a `data.frame` with product information. This request
#' supports several tidyverse methods for filtering and arranging the data.
#' 
#' You can apply some tidyverse functions (see [tidy_verbs]) to `odata_request`
#' object returned by [dse_odata_products_request()]. These apply lazy evaluation.
#' Meaning that they are just added to the object and are only evaluated after
#' calling either [dplyr::compute()] or [dplyr::collect()] (see examples).
#' @param ... Ignored in case of `dse_odata_products_request()`. Dots are passed
#' to embedded [dplyr::filter()] in case of `dse_odata_products()`
#' @param expand Additional information to be appended to the result.
#' Should be any of `"Attributes"`, `"Assets"`, or `"Locations"`. Note that,
#' these columns are not affected by [dplyr::select()] calls (before calling
#' [dplyr::collect()]).
#' @returns Returns an `odata_request` class object in case of
#' `dse_odata_products_request()`, which is an extension of [httr2::request()].
#' In case of `dse_odata_products()` a `data.frame` listing requested products is
#' returned.
#' @references <https://documentation.dataspace.copernicus.eu/APIs/OData.html>
#' @examples
#' if (interactive()) {
#'   bbox <-
#'     sf::st_bbox(
#'       c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
#'       crs = 4326) |>
#'       sf::st_as_sfc()
#' 
#'   dse_odata_products_request() |>
#'     dplyr::filter(
#'       `ContentDate/Start` > "2025-01-01") |>
#'     sf::st_intersects(bbox) |>
#'     dplyr::arrange(dplyr::desc(Id)) |>
#'     dplyr::slice_head(n = 100) |>
#'     dplyr::collect()
#' }
#' @family odata
#' @export
dse_odata_products_request <- function(..., expand) {
  exp_f <- if (missing(expand)) {
    I
  } else {
    match.arg(expand, c("Attributes", "Assets", "Locations"), TRUE)
    \(x) {
      httr2::req_url_query(x, `$expand` = expand, .multi = "explode")
    }
  }
  result <-
    .odata_url |>
    paste("Products", sep = "/") |>
    httr2::request() |>
    exp_f()
  class(result) <- union(c("odata_request", "odata_products_request"), class(result))
  result
}

#' @rdname dse_odata_products_request
#' @family odata
#' @export
dse_odata_products <- function(..., expand = NULL) {
  dse_odata_products_request(expand = expand) |>
    dplyr::filter(...) |>
    dplyr::collect()
}

#' List OData Product Nodes (i.e. Files and Directories)
#' 
#' If you know the product `Id`, you can use this function to retrieve
#' information about nodes (i.e. files and directories) within the product.
#' @param product A product identifier (`Id`)
#' @param node_path Path of nodes separated by forward slashes (`"/"`).
#' Path for which to list nodes. Default is `""`, which is the root of
#' the product
#' @param recursive A `logical` value. If set to `TRUE`, it will recursively
#' list all nested nodes. Default is `FALSE`.
#' @param ... Ignored
#' @returns A `data.frame` with information on the requested node(s)
#' @examples
#' if (interactive()) {
#'   nodes <- dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4")
#'   nodes <- dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4",
#'                                    recursive = TRUE)
#' }
#' @family odata
#' @export
dse_odata_product_nodes <- function(product, node_path = "", recursive = FALSE, ...) {
  .np_error(node_path)
  
  result <-
    .path_to_url(product, node_path) |>
    paste("Nodes", sep = "/") |>
    httr2::request() |>
    httr2::req_error(body = .odata_error) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  if (length(result$result) == 0) return(dplyr::tibble())
  result <- .simplify(result$result)
  if (recursive) {
    result <-
      dplyr::bind_rows(
        result,
        result |>
          dplyr::filter(.data$ChildrenNumber > 0) |>
          dplyr::pull("Nodes.uri") |>
          stringr::str_extract_all("(?<=/Nodes\\()(.*?)(?=\\))") |>
          lapply(paste, collapse = "/") |>
          unlist() |>
          lapply(\(x) dse_odata_product_nodes(product, x, recursive = TRUE))
      )
  }
  return (result)
}

.dse_odata_attributes <- function(...) {
  result <-
    .odata_url |>
    paste("Attributes", sep = "/") |>
    httr2::request() |>
    httr2::req_error(body = .odata_error) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  nm <- names(result)
  result |>
    .simplify2() |>
    tidyr::pivot_longer(
      dplyr::everything(),
      names_to = "Collection",
      values_to = "Attributes") |>
    dplyr::mutate(
      Attributes =
        lapply(.data$Attributes, \(x) dplyr::bind_rows(lapply(x, as.data.frame)))
    ) |>
    tidyr::unnest("Attributes") |>
    dplyr::rename(AttributeName = "Name", AttributeValueType = "ValueType")
}

#' List OData Attributes
#' 
#' Collect a list of OData attributes that can be used for
#' filtering products with [dse_odata_products_request()].
#' @param ... Ignored
#' @returns A `data.frame` listing all attributes for each collection.
#' @examples
#' if (interactive()) {
#'   dse_odata_attributes()
#' }
#' @family odata
#' @export
dse_odata_attributes <- memoise::memoise(.dse_odata_attributes)

#' Download Data Space Ecosystem Products Through OData API
#' 
#' Use [dse_odata_products()] or [dse_odata_products_request()],
#' [dse_odata_bursts()] or [dse_odata_bursts_request()] to find
#' a product or burst information. Use this function to download the product(s)
#' or burst information.
#' 
#' @param request A request containing products or burst data that you wish to download.
#' Use [dse_odata_products_request()] or [dse_odata_bursts_request()] to formulate
#' product or burst specifications.
#' @param destination A `character` string specifying the directory path,
#' where to store downloaded products
#' @param ... Ignored.
#' @inheritParams dse_s3_download
#' @returns A vector of downloaded file names stored at `destination`
#' @examples
#' if (interactive() && dse_has_s3_secret()) {
#'   dse_odata_download(
#'     dse_odata_products(Name == "S1C_AUX_PP2_V20241204T000000_G20251024T110034.SAFE"),
#'     destination = tempdir())
#'   dse_odata_download(
#'     dse_odata_products(
#'       Name ==
#'         "S1A_IW_OCN__2SDH_20250707T210608_20250707T210625_059983_07739B_893E.SAFE"),
#'     destination = tempdir())
#'   dse_odata_download(
#'     dse_odata_products(
#'       Id %in% c("c8ed8edb-9bef-4717-abfd-1400a57171a4",
#'                 "86288a07-560c-364f-b8ce-669d95f06fa0")),
#'     destination = tempdir())
#' }
#' @family odata
#' @export
dse_odata_download <- function(request, destination, ...,
                              s3_key = dse_s3_get_key(), s3_secret = dse_s3_get_secret()) {
  product_details <- request |> dplyr::collect()

  lapply(product_details$S3Path, .download_s3,
         destination = destination,
         dse_s3(s3_key = s3_key, s3_secret = s3_secret)) |>
    unlist()
}

.path_to_url <- function(product, node_path) {

  node_path <-
    strsplit(node_path, "/") |>
    unlist() |>
    sprintf(fmt = "Nodes(%s)") |>
    paste(collapse = "/")
  
  result <-
    .odata_url |>
    paste(sprintf("Products(%s)", product), node_path, sep = "/")
}

#' Alternative Route to Download OData Products
#' 
#' Downloading data using the OData API is probably fastest by using
#' [dse_odata_download()]. As an alternative, you can use this function
#' which uses the https protocol to download a product.
#' @param product Hexadecimal id of the product to be downloaded
#' @param node_path Path to a specific file in the product. When left blank (`""`)
#' The function will attempt to download the entire product as a zip archive.
#' @param destination Path to a directory where to store the downloaded file
#' @param progress `logical` value. If `TRUE` shows download progress.
#' @param ... Ignored
#' @inheritParams dse_usage
#' @returns Returns a `httr2_response` class object. It's body will hold the filename
#' of the downloaded file
#' @examples
#' if (interactive() && dse_has_account()) {
#' 
#'   dse_odata_download_path(
#'     product     = "2f497806-0101-5eea-83fa-c8f68bc56b0c",
#'     node_path   = 
#'       paste("DEM1_SAR_DTE_90_20101213T034716_20130408T035028_ADS_000000_5033.DEM",
#'             "Copernicus_DSM_30_S09_00_E026_00", "DEM",
#'             "Copernicus_DSM_30_S09_00_E026_00_DEM.dt1", sep = "/"),
#'     destination = tempdir()
#'   )
#'   
#'   dse_odata_download_path(
#'     product     = "ce4576eb-975b-40ff-8319-e04b00d8d444",
#'     destination = tempdir())
#'
#' }
#' @family odata
#' @export
dse_odata_download_path <- function(
    product, node_path = "", destination, progress = TRUE, ...,
    token = dse_public_access_token()) {
  .np_error(node_path)
  
  node_details <-
    dse_odata_product_nodes(product, node_path)
  
  if (length(node_details) == 0) {
    fn <- basename(node_path)
  } else {
    fn <- node_details$Name
    if (node_details$ChildrenNumber > 0) fn <- paste0(fn, ".zip")
  }
  
  prog <- if (progress) httr2::req_progress else I
  .path_to_url(product, node_path) |>
    stringr::str_replace("\\/$", "") |>
    paste("$value", sep = "/") |>
    httr2::request() |>
    prog() |>
    .add_token(token) |>
    ## Make sure that authentication is passed on to any redirect:
    httr2::req_options(unrestricted_auth = 1) |>
    httr2::req_error(body = .odata_error) |>
    httr2::req_perform(path = file.path(destination, fn))
}

#' Download a Quicklook for a Product
#' 
#' Downloads a 'quicklook' for a product. If the `rstudioapi` package is installed,
#' it will attempt to open the image in the Viewer panel.
#' @param product Identifier (Id) for the product for which to obtain a quicklook.
#' @param destination A destination path where to store the image.
#' @param ... Ignored
#' @returns Returns `NULL` invisibly
#' @examples
#' if (interactive()) {
#'   dse_odata_quicklook(
#'     "91822f33-b15c-5b60-aa39-6d9f6f5c773b",
#'     tempfile(fileext = ".jpg"))
#' }
#' @family odata
#' @export
dse_odata_quicklook <- function(product, destination, ...) {
  assets <-
    dse_odata_products(.data$Id == .env$product, expand = "Assets") |>
    dplyr::pull("Assets") |>
    dplyr::bind_rows()
  if ("Type" %in% names(assets))
    assets <-
      dplyr::filter(assets, dplyr::if_any("Type", \(x) x == "QUICKLOOK"))
  if (nrow(assets) == 0)
    rlang::abort(c(
      x = "This product does not have a 'quicklook'",
      i = "Try again with a different product"
    ))
  assets$DownloadLink |>
    httr2::request() |>
    httr2::req_perform(path = destination)
  if (requireNamespace("rstudioapi") && rstudioapi::isAvailable()) {
    rstudioapi::viewer(destination)
  } else {
    file.show(destination)
  }
  return(invisible())
}

library(stars) |> suppressMessages()

test_that("OData products can be listed", {
  skip_if_offline()
  skip_on_cran()
  expect_type({
    dse_odata_products(expand = c("Attributes", "Assets", "Locations"))$Id
  }, "character")
})

test_that("OData product nodes can be listed", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    nodes <- dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4",
                                     recursive = TRUE)
    nrow(nodes) > 1
  })
})

test_that("OData product details can be listed", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    id <- "c8ed8edb-9bef-4717-abfd-1400a57171a4"
    details <- dse_odata_products(Id == "c8ed8edb-9bef-4717-abfd-1400a57171a4")
    details$Id == id
  })
})

test_that("OData quicklook can be downloaded", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    tf <- tempfile(fileext = ".jpg")
    on.exit({ unlink(tf) })
    dse_odata_quicklook("91822f33-b15c-5b60-aa39-6d9f6f5c773b", tf)
    file.exists(tf) && file.size(tf) > 0
  })
})

test_that("OData attributes can be listed", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    attribs <- dse_odata_attributes()
    nrow(attribs) > 0 &&
      all(names(attribs) %in% c("Collection", "AttributeName", "AttributeValueType"))
  })
})

test_that("OData file can be downloaded through https", {
  skip_if_offline()
  skip_if_not(dse_has_account())
  skip_on_cran()
  expect_no_error({
    resp <-
      dse_odata_download_path(
        product     = "2f497806-0101-5eea-83fa-c8f68bc56b0c",
        node_path   =
          paste("DEM1_SAR_DTE_90_20101213T034716_20130408T035028_ADS_000000_5033.DEM",
                "Copernicus_DSM_30_S09_00_E026_00", "DEM",
                "Copernicus_DSM_30_S09_00_E026_00_DEM.dt1", sep = "/"),
        destination = tempdir(),
        progress = FALSE
      ) |>
      suppressMessages()
    on.exit({unlink(resp$body)})
    sp <- read_stars(resp$body)
    NULL
  })
})

test_that("OData zipped product can be downloaded through s3", {
  skip_if_offline()
  skip_if_not(dse_has_s3_secret())
  skip_on_cran()
  expect_no_error({
    fn <-
      dse_odata_download(
      dse_odata_products(Name == "S1C_AUX_PP2_V20241204T000000_G20251024T110034.SAFE"),
      destination = tempdir()) |>
      suppressMessages()
    on.exit({ unlink(fn) })
  })
})

test_that("OData zipped product can be downloaded through https", {
  skip_if_offline()
  skip_if_not(dse_has_account())
  skip_on_cran()
  expect_no_error({
    resp <-
      dse_odata_download_path(
        product     = "ce4576eb-975b-40ff-8319-e04b00d8d444",
        destination = tempdir()) |>
      suppressMessages()
    unlink(resp$body)
    NULL
  })
})

test_that("Bursts data can be listed", {
  skip_if_offline()
  skip_on_cran()
  expect_type({
    bursts <- dse_odata_bursts(ParentProductId == "879d445c-2c67-5b30-8589-b1f478904269")
    bursts$Id
  }, "character")
})

library(sf) |> suppressMessages()
library(dplyr) |> suppressMessages()

test_that("Returned STAC geometries intersect with request", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    bbox <-
      sf::st_bbox(
        c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
        crs = 4326)
    shape <- st_as_sfc(bbox)
    
    result1 <-
      dse_stac_search_request(collections = "sentinel-3-sl-2-aod-nrt") |>
      st_intersects(bbox) |>
      dplyr::collect()
    result2 <-
      dse_stac_search_request(collections = "sentinel-3-sl-2-aod-nrt") |>
      st_intersects(shape) |>
      dplyr::collect() |>
      suppressMessages()
    
    test_fun <- \(z) {
      lapply(z, \(y) {
        y <- unlist(y)
        names(y) <- c("xmin", "ymin", "xmax", "ymax")
        y <- st_bbox(y, crs = 4326) |> st_as_sfc()
        st_intersects(shape, y, sparse = FALSE) |>
          c() |>
          all() |>
          suppressMessages()
      }) |>
        unlist() |>
        all()
    }
    test_fun(result1$bbox) && test_fun(result2$bbox) 
  })
})

test_that("Returned OData geometries intersect with request", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    bbox <-
      sf::st_bbox(
        c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
        crs = 4326)
    shape <- st_as_sfc(bbox)
    
    ## In addition to geometry add a filter to avoid TimeOut error
    result1 <-
      dse_odata_products_request() |>
      dplyr::filter(`Collection/Name` == "SENTINEL-3") |>
      st_intersects(bbox) |>
      dplyr::collect()
    result2 <-
      dse_odata_products_request() |>
      dplyr::filter(`Collection/Name` == "SENTINEL-3") |>
      st_intersects(shape) |>
      dplyr::collect()
    
    test_fun <- \(z) {
      lapply(z, \(y) {
        coords <-
          do.call(rbind, lapply(y[[1]], unlist))
        coords[,1][coords[,1] < -180] <- -180
        geom <- sf::st_polygon(list(coords)) |>
          sf::st_sfc(crs = 4326)
        sf_use_s2(FALSE) |> suppressMessages()
        st_intersects(shape, geom, sparse = FALSE) |>
          c() |>
          all() |>
          suppressMessages()
      }) |>
        unlist() |>
        all()
    }
    # Test can only handle polygons, just skip in case of anything else
    (any(result1$GeoFootprint.type != "Polygon") || 
        any(result2$GeoFootprint.type != "Polygon")) || (
          test_fun(result1$GeoFootprint.coordinates) &&
            test_fun(result2$GeoFootprint.coordinates) 
        )
  })
})

library(sf) |> suppressMessages() |> suppressWarnings()

test_that("Invalid URI does not start S3 download", {
  expect_error({
    dse_s3_download("foobar", tempdir())
  }, "Not a valid S3 URI")
})

test_that("Non existing collections err", {
  skip_if_offline()
  expect_error({
    dse_stac_collections("foobar")
  }, "NotFoundError")
})

test_that("Non existing stac filter argument will produce warning", {
  skip_if_offline()
  expect_warning({
    dse_stac_search_request(foobar = NA)
  }, "Ignoring unknown filter arguments")
})

test_that("Files cannot be downloaded via STAC when authentication is missing", {
  skip_if_offline()
  expect_error({
    fn <-
      dse_stac_download(
        asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
        asset = "B01",
        destination = tempdir(),
        s3_key = "",
        s3_secret = "",
        token = NULL
      )
  }, "Need authentication")
})

test_that("Quicklook cannot be downloaded if it doesn't exist", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()
  expect_error({
    dse_odata_quicklook( "ce4576eb-975b-40ff-8319-e04b00d8d444", tempfile())
    NULL
  }, "does not have a 'quicklook'")
})

test_that("Invalid OData request produce meaningfull errors", {
  skip_if_offline()
  skip_on_cran()
  expect_error({
    dse_odata_products(foo == "bar")
  }, "Invalid field")
})

test_that("OData files cannot be downloaded without token", {
  skip_if_offline()
  skip_on_cran()
  expect_error({
    dse_odata_download_path(
      product     = "ce4576eb-975b-40ff-8319-e04b00d8d444",
      destination = tempdir(),
      token       = NULL)
    NULL
  }, "Ensure that you have specified a token")
})

test_that("`node_path` only works if it has one element", {
  skip_if_offline()
  skip_on_cran()
  expect_error({
    dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4",
                            node_path = c("", ""))
  }, "should only have one element")
})

test_that("st_intersects warns when adding multiple bboxes to stac request", {
  expect_warning({
    bbox <-
      st_bbox(
        c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
        crs = 4326)

    dse_stac_search_request() |>
      st_intersects(bbox) |>
      st_intersects(bbox)
  }, "Replacing previously defined bbox")
})

test_that("st_intersects warns when adding bbox after shape to stac request", {
  expect_warning({
    bbox <-
      st_bbox(
        c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
        crs = 4326)
    shape <- st_as_sfc(bbox)
    
    dse_stac_search_request() |>
      st_intersects(shape) |>
      st_intersects(bbox) |>
      suppressMessages()
  }, "mutually exclusive")
})

test_that("st_intersects warns when adding multiple shapes to stac request", {
  expect_warning({
    bbox <-
      st_bbox(
        c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
        crs = 4326)
    shape <- st_as_sfc(bbox)
    
    dse_stac_search_request() |>
      st_intersects(shape) |>
      st_intersects(shape) |>
      suppressMessages()
  }, "Replacing previously defined geometry")
})

test_that("st_intersects warns when adding shape after bbox to stac request", {
  expect_warning({
    bbox <-
      st_bbox(
        c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
        crs = 4326)
    shape <- st_as_sfc(bbox)
    
    dse_stac_search_request() |>
      st_intersects(bbox) |>
      st_intersects(shape) |>
      suppressMessages()
  }, "mutually exclusive")
})

test_that("Non existing assets produce error on STAC", {
  skip_if_offline()
  skip_on_cran()
  skip_if_not(dse_has_client_info())
  expect_error({
    dse_stac_download("foo", "bar", "foobar")
  }, "Asset not found")
})

test_that("Warning is thrown when collection id cannot be guessed", {
  expect_warning({
    dse_stac_guess_collection("foobar")
  }, "Couldn't guess")
})

test_that("slice_head.stac_request does not accept 'prop'", {
  expect_error({
    slice_head.stac_request(NULL, prop = 1)
  }, "'prop' argument is not implemented")
})

test_that("slice_head.odata_request does not accept 'prop'", {
  expect_error({
    slice_head.odata_request(NULL, prop = 1)
  }, "'prop' argument is not implemented")
})

test_that("Sentinel Hub errors are passed - A", {
  skip_if_not(dse_has_client_info())
  skip_if_offline()
  skip_on_cran()
  expect_error({
    dse_sh_queryables("/")
  }, "Ambiguous URI")
})

test_that("Sentinel Hub errors are passed - B", {
  skip_if_not(dse_has_client_info())
  skip_if_offline()
  skip_on_cran()
  expect_error({
    dse_sh_queryables("foobar")
  }, "Illegal collection")
})

test_that("Passing nonsense to SentinelHub will Yield 'bad request'", {
  skip_if_not(dse_has_client_info())
  skip_if_offline()
  skip_on_cran()
  expect_error({
    dse_sh_process(list(foo = "bar"), list(foo = "bar"), "", tempfile())
  }, "Bad Request")
})

test_that("Server error is passed on", {
  skip_if_offline()
  skip_on_cran()
  expect_error({
    dse_stac_search_request() |>
      dplyr::collect()
  }, "one collection")
})

## https://stac.dataspace.copernicus.eu/v1/api.html

#' @include tidyverse.R
#' @include helpers.R
#' @include account.R
#' @include req_perform.R
#' @include s3.R
NULL

.stac_base_url <- "https://stac.dataspace.copernicus.eu/v1"

.stac_get <- function(endpoint = "") {
  .stac_base_url |>
    paste(endpoint, sep = "/") |>
    httr2::request() |>
    httr2::req_error(body = .stac_error) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
}

.dse_stac_client <- function(...) {
  .stac_get("") |>
    .simplify2()
}

.stac_error <- function(resp) {
  if (httr2::resp_content_type(resp) == "application/json") {
    details <-
      resp |>
      httr2::resp_body_json()
    if ("detail" %in% names(details)) details <- details[["detail"]]
    names(details)[names(details) == "description"] <- "message"
    if ("code" %in% names(details))
      return (sprintf("%s: %s", details$code, details$message))
    loc <- lapply(details$detail, \(x) do.call(paste, c(x$loc, sep = "/"))) |> unlist()
    msg <- lapply(details$detail,`[[`, "msg") |> unlist()
    cbind(loc, msg) |> apply(1, paste, collapse = ": ", simplify = FALSE) |> unlist()
  } else {
    return("No error details returned by server")
  }
}

#' Obtain Information About the STAC Client
#' 
#' Returns information about the STAC client used in the Data Space Ecosystem
#' @param ... Ignored
#' @returns Returns a `data.frame` with the requested information
#' @examples
#' if (interactive()) {
#'   dse_stac_client()
#' }
#' @family stac
#' @export
dse_stac_client <- memoise::memoise(.dse_stac_client)

.dse_stac_collections <- function(collection, limit = 1000L, ...) {
  arg <- "collections"
  if (missing(collection)) {
    arg <- paste0(arg, "?limit=", limit)
  } else {
    arg <- paste(arg, collection, sep = "/")
  }
  result  <- .stac_get(arg)
  if (any(names(result) == "collections")) {
    attribs <- result[names(result) != "collections"]
    result  <- result$collections |> .simplify(do_unnest = FALSE)
    attr(result, "request") <- attribs
  } else {
    result <- .simplify(result, do_unnest = FALSE)
  }
  result
}

#' Get a Summary of all Data Space Ecosystem Collections
#' 
#' Use the STAC API to get a summary of all collections available
#' from the interface.
#' @param collection A specific collection for which to obtain summary information.
#' If missing (default), all collections are returned.
#' @param limit Maximum number of collections to be returned.
#' @param ... Ignored
#' @returns Returns a `data.frame` with the requested information
#' @examples
#' if (interactive()) {
#'   dse_stac_collections()
#'   dse_stac_collections("sentinel-2-l2a")
#' }
#' @family stac
#' @export
dse_stac_collections <- memoise::memoise(.dse_stac_collections)

.dse_stac_api_specs <- function(...) .stac_get("api")

.get_schema <- function(ref, api) {
  ref <- gsub("#", "api", ref)
  ref <- gsub("/", "$", ref)
  rlang::parse_expr(ref) |> rlang::eval_tidy()
}

.dse_fix_list <- function(target, source, api) {
  lapply(names(target), \(nm) {
    types <- target[[nm]]$anyOf |> .simplify()
    if ((is.null(source[[nm]]) || all(is.na(source[[nm]]))) && "null" %in% types$type) {
      if ("array" %in% types$type) {
        itms_nm  <- c("prefixItems", "items")
        itms_nm  <- itms_nm[itms_nm %in% names(types)]
        itms_idx <- which(types$type == "array")
        itms <- do.call(c, types[[itms_nm]])
        if (!any(c("$ref", "type") %in% names(itms))) {
          return(rep(NA, length(types)))
        } else {
          if ("string" %in% itms)
            return(NA) else {
              return(NA)
            }
        }
      } else {
        ## object e.g. intersects
        return(NA)
      }
    } else {
      target[[nm]] <- source[[nm]]
      if (grepl("date", nm)) {
        target[[nm]] <- lubridate::as_datetime(target[[nm]], tz = "")
        target[[nm]] <- lubridate::format_ISO8601(target[[nm]], usetz = TRUE)
      }
      if ("integer" %in% types$type) target[[nm]] <-   as.integer(target[[nm]])
      if ("number"  %in% types$type) target[[nm]] <-   as.numeric(target[[nm]])
      if ("string"  %in% types$type) target[[nm]] <- as.character(target[[nm]])
      if ("object"  %in% types$type) {
        schema <-
          .get_schema(types$additionalProperties[[1]]$propertyNames[["$ref"]], api)
        target[[nm]] <- .dse_fix_list(schema, target[[nm]], api)
      }
      if ("array" %in% types$type) target[[nm]] <- as.list(target[[nm]]) else {
        if (length(target[[nm]]) > 1)
          rlang::abort(c(x = sprintf("'%s' should have a length of 1 not %i",
                                     nm, length(target[[nm]]))))
      }
      return(target[[nm]])
    }
  })
}

.dse_stac_search_filter <- function(...) {
  api_props <- .dse_stac_api_specs()
  filt <- api_props$components$schemas$SearchPostRequest$properties
  args <- list(...)
  if (any(!names(args) %in% names(filt)))
    rlang::warn(sprintf("Ignoring unknown filter arguments: %s",
                        sprintf("'%s'", names(args)[!names(args) %in% names(filt)]) |>
                          paste(collapse = ", ")))
  result <- .dse_fix_list(filt, args, api_props)
  names(result) <- names(filt)
  result[["filter-lang"]] <- "cql2-json"
  result[["filter-crs"]] <- "http://www.opengis.net/def/crs/OGC/1.3/CRS84"
  class(result) <- union("stac_search", class(result))
  result
}

#' Create a Request for a STAC Search in the Data Space Ecosystem
#' 
#' In order to perform a search using the STAC API, you first need to
#' create a request using [dse_stac_search_request()]. This creates
#' a [httr2::request()] to which tidy verbs `?tidy_verbs` can be applied
#' (e.g., [dplyr::select()], [dplyr::filter()] and [dplyr::arrange()].
#' Results are retrieved by calling [dplyr::collect()] on the request.
#' 
#' If you prefer a graphical user interface, you can alternatively use
#' the [STAC web browser](https://browser.stac.dataspace.copernicus.eu/).
#' @param collections Restrict the search to the collections listed here.
#' If missing, the collection will be guessed from the `ids`,
#' using [`dse_stac_guess_collection()`].
#' @param ids Restrict the search to ids listed here.
#' @param ... Arguments appended to search filter request body.
#' @returns Returns a `data.frame` with search results.
#' @examples
#' if (interactive()) {
#'   library(dplyr)
#'   library(sf)
#'   
#'   bbox <-
#'     sf::st_bbox(
#'       c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
#'       crs = 4326)
#'
#'   dse_stac_search_request("sentinel-2-l1c") |>
#'     filter(`eo:cloud_cover` < 10) |>
#'     collect()
#'
#'   dse_stac_search_request("sentinel-1-grd") |>
#'     filter(`sat:orbit_state` == "ascending") |>
#'     arrange("id") |>
#'     st_intersects(bbox) |>
#'     collect()
#' }
#' @family stac
#' @export
dse_stac_search_request <- function(collections, ids, ...) {
  if (missing(collections)) {
    if (missing(ids)) {
      collections <- NA
    } else {
      collections <- lapply(ids, dse_stac_guess_collection) |>
        unlist()
    }
  }
  if (missing(ids)) ids <- NA
  filt <- .dse_stac_search_filter(collections = collections, ids = ids, ...)
  filt$intersects <- NULL # Mutually exclusive. Added by st_intersects()
  filt$bbox       <- NULL # Mutually exclusive. Added by st_intersects()
  result <-
    .stac_base_url |>
    paste("search", sep = "/") |>
    httr2::request() |>
    httr2::req_method("POST") |>
    httr2::req_body_json(filt) |>
    httr2::req_error(body = .stac_error)
  class(result) <- union("stac_request", class(result))
  result
}

#' Download Asset From STAC Catalogue
#' 
#' Use [dse_stac_search_request()] to identify assets that can be downloaded.
#' Use [dse_stac_download()] to download an asset by its STAC id and asset name.
#' @param asset_id STAC id, used for locating the asset download details.
#' @param asset Name of the asset to download
#' @param collection The identifier for a collection. The default argument
#' is the [dse_stac_guess_collection()] function which tries to guess the
#' collection id from the `asset_id`. A more rigid approach is to provide
#' the collection id as a `character` string.
#' @param destination Directory path where to store the downloaded file.
#' @param ... Ignored
#' @inheritParams dse_usage
#' @inheritParams dse_s3_download
#' @returns Returns the path to the downloaded file.
#' @examples
#' if (interactive() && (dse_has_s3_secret() || dse_has_client_info())) {
#'   dse_stac_download(
#'     asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
#'     asset = "B01",
#'     destination = tempdir()
#'   )
#' }
#' @family stac
#' @export
dse_stac_download <- function(
    asset_id, asset, collection = dse_stac_guess_collection, destination, ...,
    s3_key    = dse_s3_get_key(),
    s3_secret = dse_s3_get_secret(),
    token     = dse_public_access_token()) {

  type <- if (s3_key != "" && s3_secret != "") "s3" else
    if (!is.null(token)) "odata" else
      rlang::abort(c(
        x = "Need authentication details in order to download asset",
        i = ("Pass either S3 key and secret or OData access token")
      ))
  uri <- dse_stac_get_uri(asset_id, asset, collection, type)
  local_path <- attr(uri, "local_path")
  
  if (is.null(local_path)) fn <- NULL else
    fn <- file.path(destination, basename(local_path))
  
  if (s3_key != "" && s3_secret != "") {
    if (is.null(fn))
      fn <- file.path(destination, basename(uri))
    dse_s3_download(uri, destination,
                    s3_key = s3_key, s3_secret = s3_secret)
    return(fn)
  } else if (!is.null(token)) {
    if (is.null(fn))
      fn <- file.path(destination, basename(uri))
    uri |>
      httr2::request() |>
      .add_token(token) |>
      httr2::req_perform(path = fn)
    return(fn)
  }
}

#' Get a Uniform Resource Identifier (URI) for an Asset in a Product
#' 
#' Get a Uniform Resource Identifier (URI) for an asset in a product.
#' This can be used to download a file manually or connect to the
#' asset directly straight from the source.
#' @inheritParams dse_stac_download
#' @param type Which type of URI should be returned? Defaults
#' to `"s3"`. Use `"odata"` to get the alternative https URI.
#' @returns Returns the URI as a `character` string.
#' If available, the local path for an asset is returned as attribute.
#' @examples
#' if (interactive()) {
#'   dse_stac_get_uri(
#'     asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
#'     asset = "B01"
#'   )
#' }
#' @family stac
#' @export
dse_stac_get_uri <- function(
    asset_id, asset, collection = dse_stac_guess_collection,
    type = "s3", ...) {
  if (is.function(collection)) collection <- collection(asset_id)
  asset_info <-
    dse_stac_search_request(ids = asset_id, collections = collection) |>
    dplyr::select(!!paste0("assets.", asset)) |>
    dplyr::arrange("id") |>
    dplyr::collect()
  
  if (nrow(asset_info) == 0)
    rlang::abort(
      c(x = "Asset not found",
        i = "Ensure that you have specified the correct asset id and collection")
    )
  asset_info <- asset_info |>
    dplyr::pull("assets") |>
    dplyr::bind_rows()
  
  local_path <- asset_info[[paste0(asset, ".file:local_path")]]
  el <- ifelse(type == "s3", ".href", ".alternate.https.href")
  result <- asset_info[[paste0(asset, el)]]
  attr(result, "local_path") <- local_path
  result
}

.dse_stac_queryables <- function(collection, ...) {
  result <-
    .stac_base_url |>
    paste("collections", collection, "queryables", sep = "/") |>
    httr2::request() |>
    httr2::req_error(body = .stac_error) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
}

#' Get Queryables for a STAC Collection
#' 
#' When searching through a collection with [dse_stac_search_request()], it
#' can be helpful to know which elements can be used to filter the search
#' results (using [dplyr::filter()]). Calling [dse_stac_queryables()] tells
#' you which aspects are available for querying and expected formats.
#' @param collection Name of the collection for which to get the queryables.
#' @param ... Ignored
#' @returns Returns a named list with information about elements that can be used
#' to query the `collection`
#' @examples
#' if (interactive()) {
#'   dse_stac_queryables("sentinel-1-grd")
#' }
#' @family stac
#' @export
dse_stac_queryables <- memoise::memoise(.dse_stac_queryables)

.collection_codes <-
  data.frame(
    code = c(
      "^S1.\\_.._SLC\\_\\_",
      "^S1.\\_.._GRHD\\_",
      "^S1.\\_RTC\\_",
      "^S2.\\_MSIL1C\\_",
      "^S2.\\_MSIL2A\\_",
      "^S2.\\_GRI_L1C_",
      "^S3.\\_OL\\_1\\_EFR\\_",
      "^S3.\\_OL\\_2\\_LFR\\_",
      "^S3.\\_SL\\_1\\_RBT\\_",
      "^S3.\\_SR\\_2\\_LAN\\_",
      "^COP-DEM\\_",
      "^Copernicus\\_DSM\\_...\\_30",
      "^SRTM\\_"
    ),
    collection_id = c(
      "sentinel-1-slc",
      "sentinel-1-grd",
      "sentinel-1-rtc",
      "sentinel-2-l1c",
      "sentinel-2-l2a",
      "sentinel-2-gri-l1c",
      "sentinel-3-olci-l1-efr",
      "sentinel-3-olci-l2-lfr-ntc",
      "sentinel-3-slstr-l1-rbt",
      "sentinel-3-sral-l2-lan",
      "cop-dem",
      "cop-dem-glo-30-dged-cog",
      "srtm-dem"
    )
  )

#' Guess the Collection id from an Asset id
#' 
#' As the STAC catalogue contains a large number of records, your request
#' may receive a timeout error. To prevent this it is best to narrow down
#' your requests to specific collections. This function is a helper function
#' that tries to guess the collection id from an asset id. Note that this
#' method is not highly reliable, and it is always best to manually provide
#' a collection id to a request.
#' 
#' @param asset_id An asset identifier name, used to guess its parent
#' collection id.
#' @returns A `character` string with a guessed collection id. Or `NA` in
#' case it cannot make a guess.
#' @examples
#' dse_stac_guess_collection(
#'   "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148")
#' @family stac
#' @export
dse_stac_guess_collection <- function(asset_id) {
  idx <- lapply(asset_id, stringr::str_detect, pattern = .collection_codes$code) |>
    lapply(which)
  if (any(lengths(idx) == 0)) {
    rlang::warn(c(
      "Couldn't guess the collection id",
      i = "Please provide the collection id manually"
    ))
  }
  lapply(idx, \(i) {
    if (length(i) == 0) return(NA)
    .collection_codes$collection_id[[i]]
  }) |> unlist()
}

#' @include helpers.R
#' @include account.R
NULL

.sh_catalog_url <- paste0(.sh_api_url, "/catalog/1.0.0")
.sh_search_url <- paste0(.sh_catalog_url, "/search")

.sh_error <- function(resp) {
  if (grepl("text/html", resp$headers$`content-type`)) {
    resp |>
      httr2::resp_body_html() |>
      xml2::xml_find_all("//pre") |>
      xml2::xml_text()
  } else if (grepl("application/json", resp$headers$`content-type`)) {
    result <-
      resp |>
      httr2::resp_body_json()
    if (!is.null(result$description)) {
      result$description
    } else if (!is.null(result$error)) {
      paste(
        result$error$message,
        result$error$errors$parameter,
        result$error$errors$description,
        sep = ": ")
    } else {
      result |> lapply(as.character) |> unlist()
    }
  } else {
    "Details unknown"
  }
}

.dse_sh_collections <- function(...) {
  collections <- 
    .sh_catalog_url |>
    paste("collections", sep = "/") |>
    httr2::request() |>
    httr2::req_error(body = .sh_error) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  links <- collections$links |> .simplify()
  collections <- collections$collections |> .simplify()
  attr(collections, "links") <- links
  collections
}

.prepare_datetime <- function(datetime) {
  datetime <-
    lubridate::as_datetime(datetime) |>
    format("%Y-%m-%dT%H:%M:%SZ")
  datetime[is.na(datetime)] <- ".."
  datetime <- paste(datetime, collapse = "/")
  datetime
}

#' List Sentinel Hub Collections
#' 
#' List collections that are available from the Sentinel Hub.
#' 
#' @param ... Ignored
#' @returns Returns a `data.frame` with information about the
#' collections available from the Sentinel Hub
#' @examples
#' if (interactive()) {
#'   dse_sh_collections()
#' }
#' @include helpers.R
#' @family sentinelhub
#' @export
dse_sh_collections <- memoise::memoise(.dse_sh_collections)

#' List Sentinel Hub Features
#' 
#' List Sentinel Hub features for a specified period and region.
#' @param collection A collection for which to list the features.
#' See [dse_sh_collections()] for a list of Sentinel Hub collections.
#' @param bbox An object that can be converted into a `bbox` class
#' object (see [sf::st_bbox()]).
#' @param datetime A date-time object, or a vector of two date time
#' objects (in case of a range). Or an object that can be converted into
#' a datetime object.
#' @param limit The number of records to which the output is limited.
#' Should be between 1 and 100, and defaults to 10.
#' @inheritParams dse_usage
#' @returns Returns a `data.frame` listing features available on
#' SentinelHub for processing.
#' @examples
#' if (interactive() && dse_has_client_info()) {
#'   dse_sh_features(
#'     collection = "sentinel-2-l2a",
#'     bbox       = c(5.261, 52.680, 5.319, 52.715),
#'     datetime   = c("2025-01-01 UTC", "2025-01-07 UTC"))
#' }
#' @include helpers.R
#' @family sentinelhub
#' @export
dse_sh_features <- function(
    collection, bbox, datetime, limit = 10, ..., token = dse_access_token()) {
  bbox <- .prepare_bounds(bbox) |> sf::st_bbox() |> as.numeric()
  datetime <- .prepare_datetime(datetime)
  
  .sh_catalog_url |>
    paste("collections", collection, "items", sep = "/") |>
    httr2::request() |>
    .add_token(token) |>
    httr2::req_error(body = .sh_error) |>
    httr2::req_url_query(datetime = datetime,
                         bbox     = bbox,
                         limit    = limit,
                         .multi   = "comma") |>
    httr2::req_perform() |>
    httr2::resp_body_json() |>
    .wrap_sh_features()
  
}

.dse_sh_queryables <- function(collection, ..., token = dse_access_token()) {
  .sh_catalog_url |>
    paste("collections/%s/queryables", sep ="/") |>
    sprintf(collection) |>
    httr2::request() |>
    httr2::req_error(body = .sh_error) |>
    .add_token(token) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
}

#' List Queryable Fields on Sentinel Hub
#' 
#' Return queryable fields for a specific collection on Sentinel Hub.
#' This is useful information when composing a query with
#' [dse_sh_prepare_input()]. Use [dse_sh_collections()] to list
#' available collections.
#' @param collection Collection id for which to obtain queryable fields.
#' @param ... Ignored.
#' @inheritParams dse_usage
#' @returns Returns a named `list`, with information about queryable
#' fields for the specified `collection`.
#' @examples
#' if (interactive() && dse_has_client_info()) {
#'   dse_sh_queryables("sentinel-2-l1c")
#' }
#' @family sentinelhub
#' @export
dse_sh_queryables <- memoise::memoise(.dse_sh_queryables)


#' Process Satellite Data and Download Result
#' 
#' Users can request raw satellite data, simple band combinations such
#' as false colour composites, calculations of simple remote sensing
#' indices like NDVI, or more advanced processing such as calculation
#' of Leaf area index (LAI).
#' 
#' Use [dse_sh_use_requests_builder()] if you want to use the graphical
#' user interface at
#' [Sentinel Requests Builder](https://shapps.dataspace.copernicus.eu/requests-builder/).
#' to define a request.
#' @param input A named `list` specifying the input satellite data to
#' be processed with `evalscript` to an image. A correctly formatted
#' `list` can be created with [dse_sh_prepare_input()].
#' @param output A named `list` specifying the how to present the output
#' image, create with `evalscript`. A correctly formatted `list` can be
#' created with [dse_sh_prepare_output()].
#' @param evalscript A `character` string containing a piece of JavaScript,
#' that will be run on the Sentinel Hub server. It is used to translate
#' satellite data to pixel data in a geo-referenced image. For more information
#' on setting up such a script please consult
#' [the API documentation](https://docs.sentinel-hub.com/api/latest/evalscript/).
#' You can also use [dse_sh_get_custom_script()] to obtain ready-to-go
#' scripts from the SentinelHub repository.
#' @param destination A file name to store the downloaded image.
#' @param ... Ignored
#' @inheritParams dse_usage
#' @returns A `httr2_response` class object containing the
#' location of the downloaded file at its `destination`.
#' @references
#'  * <https://docs.sentinel-hub.com/api/latest/api/process/>
#'  * <https://shapps.dataspace.copernicus.eu/requests-builder/>
#'  * <https://custom-scripts.sentinel-hub.com/>
#'  * <https://github.com/sentinel-hub/custom-scripts>
#'  * <https://docs.sentinel-hub.com/api/latest/evalscript/>
#' @examples
#' if (interactive() && dse_has_client_info()) {
#' 
#'   bounds <- c(5.261, 52.680, 5.319, 52.715)
#'   
#'   ## prepare input data
#'   input <-
#'     dse_sh_prepare_input(
#'       bounds = bounds,
#'       time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
#'     )
#'
#'   ## prepare ouput format
#'   output <- dse_sh_prepare_output(bbox = bounds)
#'   
#'   ## retrieve processing script
#'   evalscript <- dse_sh_get_custom_script("/sentinel-2/l2a_optimized/")
#'   
#'   fl <- tempfile(fileext = ".tiff")
#'   ## send request and download result:
#'   dse_sh_process(input, output, evalscript, fl)
#' 
#'   if (requireNamespace("stars")) {
#'     library(stars)
#'     enkhuizen <- read_stars(fl) |> suppressWarnings()
#'     plot(enkhuizen, rgb = 1:3, axes = TRUE, main = "Enkhuizen")
#'   }
#' }
#' @family sentinelhub
#' @export
dse_sh_process <- function(
    input, output, evalscript, destination,
    ..., token = dse_access_token()) {
  
  .sh_api_url |>
    paste("process", sep = "/") |>
    httr2::request() |>
    httr2::req_method("POST") |>
    httr2::req_error(body = .sh_error) |>
    httr2::req_body_json(
      list(
        input      = input,
        output     = output,
        evalscript = evalscript
      ), auto_unbox = TRUE
    ) |>
    .add_token(token) |>
    httr2::req_perform(path = destination)
}

.dse_sh_custom_scripts <- function(...) {
  result <-
    "https://custom-scripts.sentinel-hub.com/custom-scripts/assets/js/search-data.json" |>
    httr2::request() |>
    httr2::req_perform() |>
    httr2::resp_body_json() |>
    lapply(tibble::as_tibble) |>
    dplyr::bind_rows() |>
    dplyr::select(-"url") |>
    dplyr::filter(endsWith(.data$relUrl, "/")) |>
    dplyr::select("title", "relUrl")
  nodes <- outer(result$relUrl, result$relUrl, startsWith)
  diag(nodes) <- FALSE
  nodes <- apply(nodes, 1, which) |> unlist() |> unique()
  result[-nodes,]
}

#' List Custom JavaScripts for Processing Sentinel Hub Data
#' 
#' Custom Eval Scripts, that can be used in Sentinel Hub requests,
#' for processing data. This functions lists scripts available
#' from <https://github.com/sentinel-hub/custom-scripts>. They
#' can be retrieved with [dse_sh_get_custom_script()].
#' 
#' Make sure that you have sufficient monthly quota left to
#' process images. You can check with [dse_usage()].
#' @references
#'  * <https://custom-scripts.sentinel-hub.com/>
#'  * <https://github.com/sentinel-hub/custom-scripts>
#' @param ... Ignored
#' @returns Returns a `data.frame` with custom scripts, containing
#' a column with a title and one with a relative URL.
#' @examples
#' if (interactive()) {
#'   dse_sh_custom_scripts()
#' }
#' @family sentinelhub
#' @export
dse_sh_custom_scripts <- memoise::memoise(.dse_sh_custom_scripts)

.dse_sh_get_custom_script <- function(rel_url) {
  "https://raw.githubusercontent.com/sentinel-hub/custom-scripts/refs/heads/main%sscript.js" |>
    sprintf(rel_url) |>
    httr2::request() |>
    httr2::req_perform() |>
    httr2::resp_body_string()
}

#' Retrieve Custom JavaScripts to be Used by Sentinel Hub
#' 
#' Sentinel Hub uses JavaScripts to process satellite images.
#' There is a repository with such custom scripts. They can be
#' listed with [dse_sh_custom_scripts()]. Use the relative URL
#' (relUrl) from that list to obtain the actual script with this
#' function.
#' @param rel_url A relative URL found with [dse_sh_custom_scripts()].
#' @returns A `character` string containing JavaScript code. This
#' script can be used with [dse_sh_process()]
#' @examples
#' if (interactive()) {
#'   dse_sh_get_custom_script("/sentinel-2/tonemapped_natural_color/")
#' }
#' @family sentinelhub
#' @export
dse_sh_get_custom_script <-
  memoise::memoise(.dse_sh_get_custom_script)

.prepare_bounds <- function(bounds) {
  if (is.numeric(bounds) && !inherits(bounds, "bbox")) {
    if (!rlang::is_named(bounds)) {
      names(bounds) <- c("xmin", "ymin", "xmax", "ymax")
      bounds <- sf::st_bbox(bounds, crs = 4326)
    }
  }
  if (inherits(bounds, "bbox")) {
    sf::st_as_sfc(bounds) |>
      sf::st_transform(4326) |>
      sf::st_bbox()
  } else if (inherits(bounds, "sf")) {
    bounds |> dplyr::summarise()
  } else {
    bounds
  }
}

#' Prepare Input and Output Fields for Sentinel Hub Request
#' 
#' [dse_sh_process()] requires a named `list` for `input` and
#' `output` settings. The functions documented here produce those lists
#' required by such a process request.
#' @param bounds A bounding box or geometry (classes `sf::bbox`,
#' `sf::sf`, `sf::sfc`) defining the boundaries of the output image.
#' @param time_range A `vector` of two date-time values, specifying
#' the time range for satellite data to include in the process.
#' @param collection_name A collection name. defaults to `"sentinel-2-l2a"`
#' to ensure you get Sentinel-2 L2A data.
#' @param id An identifier. Not documented by the API reference material.
#' @param max_cloud_coverage Maximum cloud cover to be included in the
#' process. Value between 0 and 100 (default) percent.
#' @param mosaicking_order Sets the order of overlapping tiles from which
#' the output result is mosaicked. Should be any of `"default"`,
#' `"mostRecent"`, `"leastRecent"`, or `"leastCC"`. See also
#' [the API documentation](https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/#mosaickingorder).
#' @param upsampling,downsampling Specify the interpolation technique
#' when the output resolution is smaller or larger respectively than
#' the available source data.
#' See also
#' [the API documentation](https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/#processing-options).
#' @param harmonize_values A `logical` value indicating whether units
#' are harmonised as indicated in
#' [the API documentation](https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/#harmonize-values).
#' @param width,height Size of the output image in pixels. These
#' are ignored if `bbox` is specified.
#' @param bbox You can optionally provide a bounding box (i.e., a
#' copy of `bounds`) to calculate width and height with fixed
#' aspect ratio. Width will be 512 be definition, the height is chosen
#' such that it matches with the bounding box
#' @param output_format File format for the output file. Should
#' be one of `"tiff"` (default), `"jpeg"`, `"png"`, or `"json"`.
#' @param ... Ignored
#' @returns A named `list` that can be used as `input` and `output`
#' argument to [dse_sh_process()].
#' @examples
#' dse_sh_prepare_input(
#'   bounds = c(5.261, 52.680, 5.319, 52.715),
#'   time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
#' )
#' 
#' library(sf)
#' shape <- st_bbox(c(xmin = 5.261, ymin = 52.680,
#'                    xmax = 5.319, ymax = 52.715), crs = 4326) |>
#'            st_as_sfc()
#' dse_sh_prepare_input(
#'   bounds = shape,
#'   time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
#' )
#' 
#' dse_sh_prepare_output(bbox = shape)
#' @references
#'  * <https://shapps.dataspace.copernicus.eu/requests-builder/>
#' @seealso [dse_sh_process()]
#' @rdname dse_sh_prepare_
#' @family sentinelhub
#' @export
dse_sh_prepare_input <-
  function(
    bounds,
    time_range,
    collection_name    = "sentinel-2-l2a",
    id                 = NA,
    max_cloud_coverage = 100,
    mosaicking_order   = "default",
    upsampling         = "default",
    downsampling       = "default",
    harmonize_values   = FALSE,
    ...) {
    
    upsampling   <- match.arg(
      upsampling, c("default", "nearest", "bilinear", "bicubic"),
      several.ok = TRUE) |> toupper()
    upsampling[upsampling == "DEFAULT"] <- NA
    downsampling <- match.arg(
      downsampling, c("default", "nearest", "bilinear", "bicubic"),
      several.ok = TRUE) |> toupper()
    downsampling[downsampling == "DEFAULT"] <- NA
    mosaicking_order <- match.arg(
      mosaicking_order, c("default", "mostRecent", "leastRecent", "leastCC"),
      several.ok = TRUE)
    mosaicking_order[mosaicking_order == "default"] <- NA
    
    result = list()
    bounds <- .prepare_bounds(bounds)
    
    if (inherits(bounds, "bbox")) {
      result$bounds <- list(bbox = unname(as.numeric(bounds)))
    } else {
      
      result$bounds <-
        list(
          geometry = .sfc_to_geojson(
            .prepare_bounds(bounds)
          )$features[[1]]$geometry$features[[1]]$geometry)
    }
    result$data <-
      mapply(function(cn, i, tr, mc, mo, us, ds, hv) {
        r <- list(type = cn)
        r$id <- if(is.na(i)) NULL else i
        if (!is.null(tr)) {
          r$dataFilter$timeRange <-
            lubridate::as_datetime(tr) |>
            lubridate::format_ISO8601(usetz = TRUE) |>
            rlang::set_names("from", "to") |>
            as.list()
        }
        r$dataFilter$maxCloudCoverage <- mc
        r$dataFilter$mosaickingOrder  <- if (is.na(mo)) NULL else mo
        r$processing$upsampling       <- if (is.na(us)) NULL else us
        r$processing$downsampling     <- if (is.na(ds)) NULL else ds
        r$processing$harmonizeValues  <- hv
        r
      },
      cn = collection_name,
      i  = id,
      tr = matrix(time_range, ncol = 2) |>
        apply(1, c, simplify = FALSE),
      mc = max_cloud_coverage,
      mo = mosaicking_order,
      us = upsampling,
      ds = downsampling,
      hv = harmonize_values,
      SIMPLIFY = FALSE) |>
      unname()

    result
  }

#' @rdname dse_sh_prepare_
#' @family sentinelhub
#' @export
dse_sh_prepare_output <-
  function(
    width = 512, height = 512,
    output_format     = "tiff",
    bbox,
    ...
  ) {
    output_format <-
      match.arg(output_format, c("tiff", "jpeg", "png", "json"))
    
    is_json <- output_format == "json"
    output_format[is_json] <- "application/json"
    output_format[!is_json] <- paste0("image/", output_format[!is_json])
    
    if (!missing(bbox)) {
      points <-
        .prepare_bounds(bbox) |>
        sf::st_bbox() |>
        sf::st_as_sfc() |>
        sf::st_coordinates() |>
        apply(1, sf::st_point, simplify = FALSE) |>
        sf::st_as_sfc(crs = 4326)
      dists <- sf::st_distance(sf::st_zm(points[1:3,]),
                               sf::st_zm(points[1:3,]))
      dists <- dists[cbind(1:2, 2:3)] |> as.numeric(unit = "m")
      ratio <- dists[2]/dists[1]
      width <- 512
      height <- ratio*width
    }
    
    result <-
      list(width = width, height = height)
    
    result$responses <-
      mapply(
        function(id, fm) {
          list(identifier = id, format = list(type = fm))
        },
        id = "default",
        fm = output_format,
        SIMPLIFY = FALSE
      ) |>
      unname()
    result
  }

#' Create a Request for the SentinelHub Catalogue
#' 
#' In order to perform a search using the STAC API, you first need to
#' create a request using [dse_sh_search_request()]. This creates
#' a [httr2::request()] to which tidy verbs `?tidy_verbs` can be applied
#' (e.g., [dplyr::select()] and [dplyr::filter()].
#' Results are retrieved by calling [dplyr::collect()] on the request.
#' @inheritParams dse_sh_features
#' @param ... Ignored
#' @inheritParams dse_usage
#' @returns Returns a `sentinel_request` class object, which inherits
#' from the `httr2::request` class. Call [dplyr::collect()] on it to
#' retrieve results.
#' @references
#'  * <https://docs.sentinel-hub.com/api/latest/api/catalog/>
#' @examples
#' if (interactive() && dse_has_client_info()) {
#'   library(dplyr)
#'   
#'   dse_sh_search_request(
#'     collection = "sentinel-2-l2a",
#'     bbox       = c(5.261, 52.680, 5.319, 52.715),
#'     datetime   = c("2025-01-01 UTC", "2025-01-31 UTC")
#'   ) |>
#'     filter(`eo:cloud_cover` <= 10) |>
#'     collect()
#' }
#' @family sentinelhub
#' @export
dse_sh_search_request <-
  function(collection, bbox, datetime, ..., token = dse_access_token()) {
    bbox <- .prepare_bounds(bbox) |> sf::st_bbox() |> as.numeric()
    datetime <- .prepare_datetime(datetime)
    result <-
      .sh_search_url |>
      httr2::request() |>
      httr2::req_method("POST") |>
      httr2::req_error(body = .sh_error) |>
      httr2::req_body_json(
        list(
          collections   = as.list(collection),
          bbox          = bbox,
          datetime      = datetime,
          filter        = NA,
          `filter-lang` = "cql2-json"
        )
      ) |>
      .add_token(token)
    class(result) <- union("sentinel_request", class(result))
    result
  }

#' Use Requests Builder to Send Processing Request to SentinelHub
#' 
#' Use
#' [Sentinel Requests Builder](https://shapps.dataspace.copernicus.eu/requests-builder/)
#' to compose a request. Copy the text from the 'Request Preview' panel
#' and submit with this function. Use [dse_sh_process()] when you want
#' to define a request in R, without using a web browser.
#' @param build A `character` string copied from the Request Preview
#' panel at
#' [Sentinel Requests Builder](https://shapps.dataspace.copernicus.eu/requests-builder/).
#' See `system.file("requests-builder.txt", package = "CopernicusDataspace")`
#' for an example of such a text. When you omit this argument, this function
#' will attempt to retrieve the text from the system's clipboard.
#' @inheritParams dse_sh_process
#' @param ... Ignored
#' @inheritParams dse_usage
#' @returns A `httr2_response` class object obtained after sending the request.
#' @references
#'  * <https://shapps.dataspace.copernicus.eu/requests-builder/>
#' @examples
#' ## Read text copied from 'Request Preview' panel on
#' ## <https://shapps.dataspace.copernicus.eu/requests-builder/:
#' requests_builder <-
#'   system.file("requests-builder.txt", package = "CopernicusDataspace") |>
#'     readLines(warn = FALSE) |>
#'     paste(collapse = "\n")
#' 
#' if (interactive() && dse_has_client_info()) {
#'   dest <- tempfile(fileext = ".tiff")
#'   dse_sh_use_requests_builder(requests_builder, destination = dest)
#' }
#' @family sentinelhub
#' @export
dse_sh_use_requests_builder <- function(
    build, destination, ..., token = dse_access_token()
) {
  if (missing(build)) {
    if (requireNamespace("clipr")) {
      build <- clipr::read_clip()
    } else {
      if (clipr::clipr_available()) {
        rlang::abort(
          c(x = "Trying to get 'build', however, system clipboard is not accessible",
            i = "Pass the request code explicitly as 'build' argument")
        )
      } else {
        rlang::abort(
          c(x = "Trying to get 'build' from clipboard, but missing required 'clipr' package",
            i = "Install package 'clipr' and try again")
        )
      }
    }
  }
  if (startsWith(build, "curl -X POST")) {
    build <-
      stringr::str_extract_all(
        build,
        "(?<=((-d \\')|(--form-string \\'request\\=)|(--form-string \\'evalscript\\=)))((.|\n|\r)*?)(?=\\')")
    result <- jsonlite::fromJSON(
      build[[1]][1],
      simplifyVector = FALSE)
    if (length(build[[1]]) > 1)
      result$evalscript <- build[[1]][2] 
  } else if (startsWith(build, "{")) {
    result <- jsonlite::fromJSON(build, simplifyVector = FALSE)
  } else {
    rlang::abort(c(
      x = "Cannot translate `build` to a SentinelHub request",
      i = "Make sure that you use either the 'curl' or 'body' option on the requests-builder"
    ))
  }
  dse_sh_process(
    input       = result$input,
    output      = result$output,
    evalscript  = result$evalscript,
    destination = destination,
    token       = token
  )
}

library(stars) |> suppressMessages() |> suppressWarnings()
library(lubridate) |> suppressMessages() |> suppressWarnings()
library(dplyr) |> suppressMessages() |> suppressWarnings()

test_that("Custum Eval Scripts can be listed", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    cs <- dse_sh_custom_scripts()
    nrow(cs) > 0 && typeof(cs$relUrl) == "character"
  })
})

test_that("Custom Eval Script can be retrieved", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    dse_sh_get_custom_script("/sentinel-2/true_color/") |>
      startsWith("//VERSION=3")
  })
})

test_that("Queryables for sentinel hub collection can be obtained", {
  skip_if_offline()
  skip_if_not(dse_has_client_info())
  skip_on_cran()
  expect_true({
    qt <- dse_sh_queryables("sentinel-2-l1c")
    is.list(qt) && length(qt) > 1
  })
})

test_that("SentinelHub Collections can be listed", {
  skip_if_offline()
  skip_on_cran()
  expect_true({
    cl <- dse_sh_collections()
    nrow(cl) > 0 && typeof(cl$id) == "character"
  })
})

test_that("Sentinel Hub request produces proper map", {
  skip_if_not_installed("lwgeom")
  skip_if_offline()
  skip_on_cran()
  skip_if_not(dse_has_client_info())
  skip_if({
    usage <- dse_usage()
    ratio <-
      usage[["remaining"]] |> as.numeric() |> sum() /
      usage[["configuration"]] |> as.numeric() |> sum()
    # Skip if half my monthly quota is already consumed
    ratio < 0.5
  })
  expect_true({
    bounds <- c(5.261, 52.680, 5.319, 52.715)
    
    ## prepare input data
    input <-
      dse_sh_prepare_input(
        bounds = bounds,
        time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
      )
    
    ## prepare ouput format
    output <- dse_sh_prepare_output(bbox = bounds)
    
    ## retrieve processing script
    evalscript <- dse_sh_get_custom_script("/sentinel-2/l2a_optimized/")
    
    fl <- tempfile(fileext = ".tiff")
    on.exit({unlink(fl)})
    ## send request and download result:
    dse_sh_process(input, output, evalscript, fl)
    
    enkhuizen <- read_stars(fl) |> suppressWarnings()
    dims <- dim(enkhuizen)
    all(abs(dims[1:2] - c(512, 509)) <= 1) &&
      dims[3] == 4 && any(!is.na(enkhuizen[[1]]))
  })
})

test_that("Sentinel Hub request produces proper map with Requests Builder", {
  skip_if_offline()
  skip_on_cran()
  skip_if_not(dse_has_client_info())
  skip_if({
    usage <- dse_usage()
    ratio <-
      usage[["remaining"]] |> as.numeric() |> sum() /
      usage[["configuration"]] |> as.numeric() |> sum()
    # Skip if half my monthly quota is already consumed
    ratio < 0.5
  })
  expect_true({

    requests_builder <-
      system.file("requests-builder.txt", package = "CopernicusDataspace") |>
      readLines(warn = FALSE) |>
      paste(collapse = "\n")
    
    fl <- tempfile(fileext = ".tiff")
    dse_sh_use_requests_builder(requests_builder, destination = fl)
    on.exit({unlink(fl)})

    enkhuizen <- read_stars(fl) |> suppressWarnings()
    dims <- dim(enkhuizen)
    all(abs(dims[1:2] - c(512, 509)) <= 1) &&
      dims[3] == 3 && any(!is.na(enkhuizen[[1]]))
  })
})

test_that("SentinelHub features are correctly listed", {
  skip_if_offline()
  skip_on_cran()
  skip_if_not(dse_has_client_info())
  expect_true({
    rng <- as_datetime(c("2025-01-01 UTC", "2025-01-07 UTC"))
    features <-
      dse_sh_features(
        collection = "sentinel-2-l2a",
        bbox       = c(5.261, 52.680, 5.319, 52.715),
        datetime   = rng)
    
    dt <-
      features$properties.datetime |>
      as_datetime()
    all(between(dt, rng[[1]], rng[[2]])) &&
      all(features$collection == "sentinel-2-l2a")
  })
})

test_that("SentinelHub search produces correct results", {
  skip_if_offline()
  skip_on_cran()
  skip_if_not(dse_has_client_info())
  expect_true({
    dt <- as_datetime(c("2025-01-01 UTC", "2025-01-31 UTC"))
    result <-
      dse_sh_search_request(
        collection = "sentinel-2-l2a",
        bbox       = c(5.261, 52.680, 5.319, 52.715),
        datetime   = dt
      ) |>
      filter(`eo:cloud_cover` <= 10) |>
      collect()
    
    all(between(
      as_datetime(result$properties.datetime),
      dt[[1]], dt[[2]])) &&
      all(result$`properties.eo:cloud_cover` <= 10)
  })
})

#' @importFrom memoise memoise
#' @importFrom methods setOldClass
#' @importFrom rlang .data .env
NULL

setOldClass("odata_request")
setOldClass("stac_request")
setOldClass("sentinel_request")

.onLoad = function(libname, pkgname) {
  register_all_s3_methods() # dynamically registers non-imported pkgs (tidyverse)
}

register_all_s3_methods = function() {
  register_s3_method("dplyr", "arrange",          "odata_request")
  register_s3_method("dplyr", "arrange",           "stac_request")
  register_s3_method("dplyr", "filter",           "odata_request")
  register_s3_method("dplyr", "filter",        "sentinel_request")
  register_s3_method("dplyr", "filter",            "stac_request")
  register_s3_method("dplyr", "collect",          "odata_request")
  register_s3_method("dplyr", "collect",       "sentinel_request")
  register_s3_method("dplyr", "collect",           "stac_request")
  register_s3_method("dplyr", "compute",          "odata_request")
  register_s3_method("dplyr", "select",           "odata_request")
  register_s3_method("dplyr", "select",        "sentinel_request")
  register_s3_method("dplyr", "select",            "stac_request")
  register_s3_method("dplyr", "slice_head",       "odata_request")
  register_s3_method("dplyr", "slice_head",    "sentinel_request")
  register_s3_method("dplyr", "slice_head",        "stac_request")
  register_s3_method("sf",    "st_intersects",    "odata_request")
  register_s3_method("sf",    "st_intersects",     "stac_request")
}

# from: https://github.com/tidyverse/hms/blob/master/R/zzz.R
register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)
  
  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }
  
  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }
  
  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}

test_that("Source code should not have things on TODO list", {
  skip_on_ci()
  skip_on_cran()
  expect_false({
    skip_if(length(unclass(packageVersion("CopernicusDataspace"))[[1]]) > 3,
            "Skipping during development")
    r_path <- normalizePath(test_path()) |> dirname() |> dirname()
    files_to_check <- list.files(r_path, pattern = ".[r|R]$|NEWS|DESCRIPTION|README\\.md",
                                 recursive = TRUE, full.names = TRUE)
    files_to_check <- files_to_check[!endsWith(files_to_check, "test-dev.r")]
    any(
      unlist(
        lapply(files_to_check, function(file) {
          content <- suppressWarnings(readLines(file))
          result  <- grepl("TODO", content) & !grepl("grepl\\(\"TODO\"", content) & !grepl("on TODO list", content)
          if (any(result)) {
            warning(sprintf("File `%s` has items on TODO list at lines `%s`",
                            file, paste(which(result), collapse = "`, `")))
          }
          any(result)
        })
      )
    )
  })
})

---
title: "Using the STAC Catalogue"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using the STAC Catalogue}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
eval_chunks <-
  CopernicusDataspace::dse_has_client_info() &&
  CopernicusDataspace::dse_has_account() &&
  CopernicusDataspace::dse_has_s3_secret()
```

In the Copernicus Data Space Ecosystem (CDSE),
[STAC stands for SpatioTemporal Asset Catalog](https://stacspec.org/en).
It is a standardized, open-source metadata specification used to structure, search,
and discover Earth Observation (EO) data.

Instead of requiring users to download massive, raw satellite images, the CDSE STAC
[RESTful API](https://en.wikipedia.org/wiki/REST) allows developers to query precise
metadata (e.g., location, time, cloud cover, and specific bands) to locate exact data
assets.

Key Functions of STAC:

  * *Data Access*: It provides a direct path (such as S3 storage links) to cloud-hosted
    imagery, enabling tools to process a subset of data without full downloads.
  * *System Interoperability*: It replaces satellite-specific extensions with a
    unified data model, allowing the same code or software to seamlessly handle diverse
    datasets like Sentinel-1 and Sentinel-2.

The package also offers features for the complementary primary catalogue via
OData. For more details on that read `vignette("OData")`.

In general to understand which STAC client the server is offering, you can
call `dse_stac_client()`.

## Data Exploration

A good starting point of exploring data with STAC is being aware which collections
of data are available in the first place. You can list them as follows:

```{r stac-collections, eval=eval_chunks}
library(CopernicusDataspace)
dse_stac_collections()
```

The returned `data.frame` contains descriptive information about each of the collections.
It can help you focus your search. Once you have identified a collection, you can
check which filter/search are available for further narrowing your exploration tour:

```{r queryables, eval=eval_chunks}
dse_stac_queryables("sentinel-1-grd") |> summary()
```

The example above shows 11 properties that can be used to focus the search.
You can start an actual search by creating a STAC search request with:
`dse_stac_search_request()`.  It creates a special
class of `httr2` request object. In essence, it is a request to the API
server, which you can modify with tidyverse operators. This sounds more
complicated than it is.

Once you have created the request, you can add tidyverse operators (like
`filter()`, `arrange()` and `slice_head()`), to modify this request.
You can join those modifications with the pipe operator (`|>` or `%>%`).
You can also query products that intersect with specific spatial features (`sf`)
using `st_intersects()`.

```{r stac-search, eval=eval_chunks}
library(dplyr)
library(sf)

bbox <-
  sf::st_bbox(
    c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
    crs = 4326)

dse_stac_search_request("sentinel-1-grd") |>
  filter(`sat:orbit_state` == "ascending") |>
  arrange("id") |>
  st_intersects(bbox) |>
  collect()
```

## Downloading Data

When downloading data, you could retrieve the Uniform Resource Identifier (URI) with
`dse_stac_get_uri()`. To get an URI, you need at least the asset identifier and
the specific asset. Both can be obtained with a search as shown above. The example
below shows you how:

```{r stac-uri, eval=eval_chunks}
dse_stac_get_uri(
  asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
  asset = "B01",
  collection = "sentinel-2-l1c"
)
```

This approach also needs the collection from which the asset is made available.
If you don't provide it, it will be guessed with `dse_stac_guess_collection()`.
This function is not 100% reliable, so it's best practice to provide the collection
manually. Instead of working with the URI yourself it is easier to call
`dse_stac_download()`. It will automatically takes care of required authentication
for downloading the file (if properly provided). Check `vignette("Authentication")`
for more information about the authentication process.

```{r stac-download, eval=FALSE}
dse_stac_download(
  asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
  asset = "B01",
  collection = "sentinel-2-l1c",
  destination = tempdir()
)
```

---
title: "Reading Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reading Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 8,
  fig.height = 7,
  fig.ext = "png"
)
eval_chunks <- Sys.getenv("COPERNICUS_DATA_VIGNETTE") != ""
```

```{r setup, message=FALSE}
library(CopernicusDataspace)
library(dplyr)
library(jsonlite)
library(sf)
library(stars)
```

The Copernicus Data Space Ecosystem (CDSE) provides a wide range of data formats,
each with their own quirks. Therefore, this package focuses only on wrapping the
CDSE Application Programming Interface (API). Allowing to search and download
the files. Because of the diversity of data and data formats, there is no consistent
way of supporting this for CDSE from this package.

Instead, some frequently found formats are presented here, and a user case is demonstrated.
With the purpose of making this package more applicable in practical settings.

## Ground Control Points
In some cases data is stored as raster data, in the orientation
in which the satellite made the observation. Such data may not have
a valid Coordinate Reference System (CRS), but instead have
Ground Control Points (GCP), to geo-reference the data. For any
spatial analyses, you first need to warp that data to a meaningful
projection. But any data masks are preferably applied to the raw
data before such warps (which require interpolation).

Let's demonstrate this step-by-step with a user-case. We have a product Id,
in a specific collection, and we have a bounding box: the area of interest.
We first obtain the Uniform Resource Identifier (URI) for the S3 service of this
product (`dse_stac_get_uri()`). We then convert this URI to a
Virtual System Interface (VSI) using `dse_s3_uri_to_vsi()`. We do this because
we can directly connect with a product using the VSI in `stars::read_stars()`.
Note that we do need to run `dse_s3_set_gdal_options()` first, this will
make the S3 secrets available to the GDAL driver used by the `stars` package.

This will create a proxy raster object. Meaning that no data is transferred unless
necessary. This allows us to lazily specify operations for the raster object,
before it is even downloaded. All these steps are shown below.

```{r data-download, eval=eval_chunks}
if (dse_has_s3_secret()) {
  ## Define the area of interest:
  bbox <- st_bbox(c(xmin =  4.66,
                    ymin = 52.842,
                    xmax =  7.15,
                    ymax = 53.62), crs = 4326)
  
  ## Assume that we have already located the asset below within our
  ## area of interest:
  id <- "S1A_IW_GRDH_1SDV_20241125T055820_20241125T055845_056707_06F55C_12F9_COG"
  
  ## And it's from this collection:
  cllctn <- "sentinel-1-grd"
  
  ## Get the URI for the manifest.safe file. This is the preferred entry
  ## point into the data. It contains all bands and all meta-information
  ## (like accurate Ground Control Points):
  uri_safe <- dse_stac_get_uri(id, "safe_manifest", cllctn)
  
  ## Translate the URI into a virtual system indicator (VSI), such that
  ## the stars package can interact with it:
  vsi_safe <- dse_s3_uri_to_vsi(uri_safe)
  
  ## Set up the S3 credentials such that the stars package (and its
  ## underpinning GDAL library) will recognise them.
  dse_s3_set_gdal_options()
  
  ## Create a stars proxy object from the VSI. No data is downloaded
  ## yet. But you can set up lazy data manipulations with tidyverse.
  proxy <- read_stars(vsi_safe, proxy = TRUE)
}
```

As we are working with the `stars` package in this demonstration, we have to
acknowledge that this package does not preserve the GCPs when manipulating the
raster. Therefore, we explicitly read the information from the original product,
such that we can later rely on it.

As the raw data is not geo-referenced, we can also use the GCP information to subset
the raw data to the area of interest. This in order to reduce data transfer and
computational effort.

```{r meta-info, eval=eval_chunks}
## Use `gdal_info` to retrieve the raster information
json_info <-
  gdal_utils("info", vsi_safe, options = c("-json"), quiet = TRUE) |>
  fromJSON()

## Get the ground control points for the specific raster.
## You will need it for subsetting the proxy.
gcp <-
  st_as_sf(json_info$gcps$gcpList, coords= c("x", "y"), crs = 
             st_crs(json_info$gcps$coordinateSystem$wkt))

## A plot showing where the raster's GCPs are, and where the bbox lies
par(mar = c(1, 1, 1, 1))
plot(gcp["pixel"] |> st_geometry(), main = "Ground Control Points", col = NA)
plot(gcp["pixel"], add = TRUE)
plot(bbox |> st_as_sfc() |> st_geometry(), border = "red", add = TRUE)

## Determine which raster indices are within the bbox
## of our area of interest:
gcp_crop <-
  st_crop(gcp,
          st_transform(bbox |>
                         st_as_sfc() |>
                         st_buffer(units::as_units(1, "km")),
                       st_crs(json_info$gcps$coordinateSystem$wkt)))

dim_x <- st_get_dimension_values(proxy, "x") |> sort()
dim_x <- which(dim_x >= min(gcp_crop$pixel) & dim_x <= max(gcp_crop$pixel))
dim_y <- st_get_dimension_values(proxy, "y") |> sort()
dim_y <- which(dim_y >= min(gcp_crop$line) & dim_y <= max(gcp_crop$line))
```

```{r, echo=FALSE}
knitr::include_graphics("../man/figures/read-data-gcp.png")
```

The product contains two bands (VH and VV).
Both bands share the same geometry and use the same GCPs. Therefore,
we can use the VV band to mask the VH band, and only use VH values when
VV measurements are at least 30 dB. Calling `stars::write_stars()` will
force the lazy operations to be executed and only the require data is downloaded
and stored at the specified location.

```{r data-mask, eval=eval_chunks}
masked_proxy <-
  proxy[,dim_x, dim_y,] |>
  st_apply(c("x", "y"), \(x) {
    ## From json_info$bands we know that band 1=VH and 2=VV
    ## We want VH but only when VV >= 30:
    ifelse(x[2] >= 30, x[1], NA)
  })

## Store the masked file:
masked_file <- tempfile(fileext = ".tif")
write_stars(masked_proxy, masked_file, chunk_size = c(1024, 1024),
            NA_value = -9999)
```

The VH band is now masked. But when it is plotted, you will
notice (if you are familiar with the area) that the image
is mirrored. That is because it is shown in the raster index
coordinates.

```{r data-read-masked, eval=eval_chunks}
masked <- read_stars(masked_file)
plot(masked, main = "masked")
```

```{r, echo=FALSE}
knitr::include_graphics("../man/figures/read-data-masked.png")
```

In the process of masking, sub-setting and downloading the data,
the GCPs got lost. So we first need to reintroduce them, before we
can continue.

```{r data-add-gcp, eval=eval_chunks}
## The masked file has lost original GCP information
## So create a reference file where we explicitly add the GCP info
ref_file <- tempfile(fileext = ".tif")

## Update GCP info for selected raster indices:
gcp <-
  json_info$gcps$gcpList |>
  mutate(pixel = pixel - min(dim_x) - 1,
         line = line - min(dim_y) - 1) |>
  filter(line >= 0 & pixel >= 0)
opts <-
  cbind("-gcp", as.matrix(gcp[, c("pixel", "line", "x", "y", "x")])) |>
  apply(1, c) |> c()

## Create reference file with GCP info
gdal_utils(
  "translate",
  source = masked_file,
  destination = ref_file,
  options = c(opts,
              "-a_srs", json_info$gcps$coordinateSystem$wkt,
              "-a_nodata", "-9999") ## missing values
)
```

Now that the GCPs have been reintroduced, we can warp the image to
a relevant CRS. In our case we create a 10m x 10m grid in UTM31N projection.

```{r data-warp, eval=eval_chunks}
## Warp the masked file to a fixed raster with meaningful CRS
## (In this case UTM31 with 10x10m cells)
warped_file <- tempfile(fileext = ".tif")

gdal_utils(
  "warp",
  source = ref_file,
  destination = warped_file,
  options = c(
    "-t_srs", "EPSG:32631",
    "-tps",
    "-tr", c(10, 10), "-tap", "-r", "bilinear",
    "-srcnodata", "-9999", ## missing values
    "-dstnodata", "-9999", ## missing values
    "-overwrite"))
warped <- read_stars(warped_file)

plot(warped, reset = FALSE, axes = TRUE, main = "warped")
```

```{r, echo=FALSE}
knitr::include_graphics("../man/figures/read-data-warped.png")
```

## Alternative Raster Packages

As this package is only a wrapper around the API and files retrieved with the package
can be handled by any suitable package. Here we choose to work with `stars` as
it is relatively easy to use and go hand-in-hand with tidyverse operators.

However, files could just as well be processed with the
[`terra`](https://rspatial.github.io/terra/) package, which
is generally faster than [`stars`](https://r-spatial.github.io/stars/).
You could also work with [`gdalraster`](https://firelab.github.io/gdalraster/)
which provides low level access to GDAL library features for raster files.

---
title: "Using SentinelHub"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using SentinelHub}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
eval_chunks <-
  CopernicusDataspace::dse_has_client_info() &&
  CopernicusDataspace::dse_has_account() &&
  CopernicusDataspace::dse_has_s3_secret()
```

Sentinel Hub functions as the core engine for accessing, processing, and
visualizing massive amounts of satellite data. It acts as a cloud-based service
that allows users to interact with the full archive of Sentinel and other Earth
Observation (EO) data without needing to download large files to their local
machines.

The overall process involves identifying relevant data sets. The send the
description of the input data together with a special script to process data
to the server. The server processes the data and you can download the result.

## Data Exploration

A good point to start using SentinelHub is by listing the collections that
are available for this service.

```{r sh-collections, message=FALSE, eval=eval_chunks}
library(CopernicusDataspace)
library(stars) ## required below for reading and plotting data
library(dplyr) ## used for data manipulation below
dse_sh_collections()
```

You can also search products using SentinelHub by creating a `dse_sh_search_request()`.
You can add a narrow down the search with a tidyverse `filter()`. You need to retrieve
the search results by calling `collect()`:

```{r sh-search, eval=eval_chunks}
bounds <- c(5.261, 52.680, 5.319, 52.715)
if (dse_has_client_info()) {
  dse_sh_search_request(
    collection = "sentinel-2-l2a",
    bbox       = bounds,
    datetime   = c("2025-01-01 UTC", "2025-01-31 UTC")
  ) |>
    filter(`eo:cloud_cover` <= 10) |>
    collect()
}
```

## Data Downloading

SentinelHub is a processing service, meaning that you should not to attempt
downloading raw data. Instead, you need to specify which subset of data you
wish to process, how you wish to process it, and how you like to have your results.

In order to specify your input data you can use the helper function
`dse_sh_prepare_input()`. The function `dse_sh_prepare_output()` is to help
you formulate your output specifications.

For the processing of the data, SentinelHub uses JavaScript. It is referred to
as [EvalScript](https://docs.sentinel-hub.com/api/latest/evalscript/).
The [technical manual](https://docs.sentinel-hub.com/api/latest/evalscript/v3/)
describes what the script should look like.

You can use `dse_sh_get_custom_script()` to retrieve existing custom scripts from
the default [repository](https://custom-scripts.sentinel-hub.com/).

The steps presented above are demonstrated below:

```{r prepare-sentinelhub, eval=eval_chunks}
## prepare input data:
input <-
  dse_sh_prepare_input(
    bounds = bounds,
    time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
  )

## prepare ouput format:
output <- dse_sh_prepare_output(bbox = bounds)

## retrieve processing script:
evalscript <- dse_sh_get_custom_script("/sentinel-2/l2a_optimized/")

## destination file:
fl <- tempfile(fileext = ".tiff")
```

Once you have all these elements ready, you can submit your request to
SentinelHub and retrieve the result, using `dse_sh_process()`. In order
to submit this request, you need an access token, which in turn requires
a `client_id` and `client_secret`. For more information about this
check out `vignette("Authentication")`.

```{r download-sentinelhub, eval=FALSE}
if (dse_has_client_info()) {
  ## send request and download result:
  dse_sh_process(input, output, evalscript, fl)
}
```

If you like to use a graphical user interface to prepare your SentinelHub
request, you can use the
[Requests-Builder](https://shapps.dataspace.copernicus.eu/requests-builder/)
in your web browser. You can create a request through the provided menu and
it is shown in a text field. You can copy this text and submit it to SentinelHub
with `dse_sh_use_requests_builder()`.

## Eval Scripts

As mentioned above, there are many readily developed
[Eval Scripts](https://docs.sentinel-hub.com/api/latest/evalscript/) available.
You can list the available scripts on the repository using:

```{r eval-scripts, eval=eval_chunks}
dse_sh_custom_scripts()
```

As each script has a unique way of processing the raw data, you can quickly create
different visualisations of the satellite information. Lets illustrate this
by reusing the same `input` and `output` specifications as defined above.

```{r water-bodies, eval=eval_chunks}

if (dse_has_client_info()) {
  evalscript <-
    dse_sh_get_custom_script("/sentinel-2/simple_water_bodies_mapping-swbm/")
  
  ## destination file:
  fl_waterbody <- tempfile(fileext = ".tiff")
  
  ## send request and download result:
  dse_sh_process(input, output, evalscript, fl_waterbody)

  ## read and plot result:
  waterbodies <- read_stars(fl_waterbody) |> suppressWarnings()
  plot(waterbodies, rgb = 1:3, main = "Water bodies")
}
```

This maps shows water bodies in blue. In this particular case, it does
a decent job, however some greenhouses show up as water
bodies (i.e., false-positives). It must be noted that here the script
was just used as is, and might need some tweaking of parameters for
better performance.

Below an example of how the same raw data can be visualised differently,
using another dedicated EvalScript.

```{r land-use, eval=eval_chunks}
if (dse_has_client_info()) {
  evalscript <-
    dse_sh_get_custom_script("/sentinel-2/land_use_with_linear_discriminant_analysis/")
  
  ## destination file:
  fl_landuse <- tempfile(fileext = ".tiff")
  
  ## send request and download result:
  dse_sh_process(input, output, evalscript, fl_landuse)

  ## read and plot result:
  landuse <- read_stars(fl_landuse) |> suppressWarnings()
  plot(landuse, rgb = 1:3, main = "Land use")
}
```

---
title: "Using the OData API"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using the OData API}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
eval_chunks <-
  CopernicusDataspace::dse_has_client_info() &&
  CopernicusDataspace::dse_has_account() &&
  CopernicusDataspace::dse_has_s3_secret()
```

In the Copernicus Data Space Ecosystem (CDSE), OData (Open Data Protocol) is the
primary [RESTful API](https://en.wikipedia.org/wiki/REST) standard used for
searching, discovering, and downloading satellite data products.

Key functions of OData in CDSE are:

 * *Metadata Querying*: You can search for products using complex filters such
   as collection name (e.g., Sentinel-2), geographical area (AOI), acquisition
   dates, and cloud cover percentage.
 * *Data Download*: Once a specific product ID is retrieved via a search, the
   OData API can be used to trigger the actual download of the data files.

This package provides convenient wrappers to access those features, and are
described in this vignette.

The package also offers features for the complementary alternative catalogue via
STAC. For more details on that read `vignette("STAC")`.

## Searching Products

For searching products, you could simply use `dse_odata_products()`,
where you could directly add filters as arguments. But for more sophisticated
queries, you can use `dse_odata_products_request()`. It creates a special
class of `httr2` request object. In essence, it is a request to the API
server, which you can modify with tidyverse operators. This sounds more
complicated than it is.

Once you have created the request, you can add tidyverse operators (like
`filter()`, `arrange()` and `slice_head()`), to modify this request.
You can join those modifications with the pipe operator (`|>` or `%>%`).
You can also query products that intersect with specific spatial features (`sf`)
using `st_intersects()`. 

Each collection will have specific fields which you can use to filter
your product search. To discover which fields you use, call
`dse_odata_attributes()`.

All of this is shown in the example below, where we filter products
that were created after January first of 2025. Where the image overlaps
with the bounding box we specify. We create a descending arrangement of
identifiers, and select only the first 5 results. By calling `collect()`
we ensure that the request is performed and the result is returned in a
human readable form (`data.frame`).

```{r odata-search, message=FALSE, eval=eval_chunks}
library(CopernicusDataspace)
library(dplyr) # Tidiverse package for data manipulation
library(sf)    # package for handling simple (spatial) features

## Define a bounding box for our search query:
bbox <-
  st_bbox(
    c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
    crs = 4326) |>
    st_as_sfc()

## Create an API request:
dse_odata_products_request() |>
  ## Filter content on ContentDate/Start`:
  filter(
    `ContentDate/Start` > "2025-01-01") |>
  ## Only data that intersects with our bounding box:
  st_intersects(bbox) |>
  ## Arrange by descending Id:
  arrange(desc(Id)) |>
  ## Only the first 5 hits:
  slice_head(n = 5) |>
  ## Collect the results:
  collect()
```


## Burst Data

[Bursts](https://documentation.dataspace.copernicus.eu/APIs/Sentinel-1%20SLC%20Burst.html)
is a concept associated with the Sentinel-1 C-SAR instrument.
They have dedicated functions for searching for specific burst data:
`dse_odata_bursts()` and `dse_odata_bursts_request()`. These are the
equivalents of `dse_odata_products()` and `dse_odata_products_request()`.
They work with the same mechanisms and downloading them works the same as
for normal OData products, as shown below. Below a basic example of looking
for and downloading of burst data:

```{r odata-bursts, eval=FALSE}
burst_req <-
  dse_odata_bursts_request(ParentProductId == "879d445c-2c67-5b30-8589-b1f478904269")

## Note that these are large files and may take a while to download:
dse_odata_download(
  burst_req,
  tempdir()
)
```

## Exploring a Product

Once you have identified an interesting product, there are ways to further
explore it. You need the product identifier (ID) to do so. With
`dse_odata_product_nodes()` you can see which files are associated with the
product:

```{r odata-nodes, eval=eval_chunks}
dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4",
                        recursive = TRUE)
```

If available you can also get a quick peek at your product by calling
`dse_odata_quicklook()`:

```{r odata-quicklook, eval=FALSE}
dse_odata_quicklook(
    "91822f33-b15c-5b60-aa39-6d9f6f5c773b",
    tempfile(fileext = ".jpg"))
```

## Downloading Data

As shown in the 'exploring' section above, products can contain 'nodes',
which could either represent a file or a directory. If you want to download
all files in such a directory you can call `dse_odata_download_path()`.
It will download all files in the specified node path. This is demonstrated
below:

```{r odata-download-path, eval=FALSE}
if (dse_has_account()) {
  dse_odata_download_path(
    product     = "2f497806-0101-5eea-83fa-c8f68bc56b0c",
    node_path   =
      paste("DEM1_SAR_DTE_90_20101213T034716_20130408T035028_ADS_000000_5033.DEM",
            "Copernicus_DSM_30_S09_00_E026_00", "DEM",
            "Copernicus_DSM_30_S09_00_E026_00_DEM.dt1", sep = "/"),
    destination = tempdir()
    )
}
```

The example above uses an alternative https route and requires a
private access token, which in turn requires your Copernicus Data Space
user id (usually your e-mail address) and password (see
`vignette("Authentication")`).

The formal OData route uses [S3 buckets](https://en.wikipedia.org/wiki/Amazon_S3)
and requires S3 authentication (see `vignette("Authentication")`).
You can perform such downloads with `dse_odata_downloads()` as shown below

```{r download-odata, eval=FALSE}
if (dse_has_s3_secret()) {
  dse_odata_download(
      dse_odata_products(
        Name ==
          "S1A_IW_OCN__2SDH_20250707T210608_20250707T210625_059983_07739B_893E.SAFE"),
      destination = tempdir())
}
```

---
title: "Authentication"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Authentication}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Unfortunately, the authentication process is rather complex in the Copernicus Data Space
Ecosystem (CDSE). Different services require different authentication schemes. This
vignette attempts to address all relevant authentication methods, and which is needed
where. And which steps you need to complete to obtain the required secrets.

## Authentication methods

In essence there are three authentication methods used in this package for the CDSE:

 * Access Token; it needs:
     * Client ID
     * Client Secret
 * Public Access Token; it needs:
     * User identifier
     * User password
 * S3 key + secret

Both access tokens are long pieces of encrypted codes that you can request with
specific secrets that are associated with your CDSE account. Both access tokens
are only valid for a short period of time (usually about 10 minutes), so you
may need to obtain a fresh token once in a while.

## What and Where?

The list below shows which of the three authentication methods is required by which
functions in this package:

 * Access Token; required by:
     * `dse_get_token_details()` For decoding the token
     * Account info
       * `dse_usage()`
     * SentinelHub functions:
       * `dse_sh_features()`
       * `dse_sh_process()`
       * `dse_sh_search_request()`
       * `dse_sh_use_requests_builder()`
       * `dse_sh_features()`
 * Public Access Token; required by:
     * `dse_set_gdal_token()` For making the token visible to the GDAL driver
     * `dse_get_token_details()` For decoding the token
     * `dse_odata_download_path()`
     * `dse_stac_download()` in case of a https download
     * Account info
       * `dse_user_statistics()`
 * S3 key + secret; required by:
     * `dse_stac_download()` in case of an S3 download
     * `dse_odata_download()`
     * `dse_s3_download()`

## Setting up Authentication Methods

### A CDSE Account

All authentication methods share a commonality: They all require a CDSE user account.
You can create one by visiting the
[login page](https://identity.dataspace.copernicus.eu/auth/realms/CDSE/account/#/personal-info)
and click "register". Fill out the form and follow the instructions.

### A Public Access Token

After completing this registration you will have a user id (usually your e-mail) and
a password (created by yourself). You can use those to obtain a *public access token*,
which you can use for the functions listed above. You can obtain such a token by
calling `dse_public_access_token()`, with your user Id and password.

### A Non-Public Access Token

Getting a (non-public) *access token*, you need to complete additional steps.
In order to retrieve such a key, you need to register a client.
To register such a client, visit the
[Sentinel Dashboard](https://shapps.dataspace.copernicus.eu/dashboard/) and go
to "User settings". There you will have the option to create an OAuth client.
Follow the instructions, and make sure to safely copy the client id and secret.
The latter is only displayed once. More detailed instructions are provided by the
[API documentation](https://documentation.dataspace.copernicus.eu/APIs/SentinelHub/Overview/Authentication.html#registering-oauth-client)

You can now retrieve an access token (required by the functions listed above)
by calling `dse_access_token()` with the client Id and client secret that you have
created with the steps above.

### S3 Key and Secret

To create an S3 key and secret, you should visit the
[s3-credentials page](https://eodata-s3keysmanager.dataspace.copernicus.eu/panel/s3-credentials),
and log in with your CDSE account details. By clicking "add credential",
you can create a new key and secret. Store them in a safe place, as the secret is
only shown once. You can pass the key and secret as `s3_key` and `s3_secret` arguments
to functions requesting them (see above).

## Masking Secrets

As you might want to share your R scripts that use the CopernicusDataspace package,
but don't want to share personal secret information, there is a way around including
them in your script.

You can store them as
[system environment variables](https://en.wikipedia.org/wiki/Environment_variable),
which are automatically detected by this package.

You can set them with `Sys.setenv()` at the start of each R session. You can
also edit your `.renviron`, such that they are loaded automatically each new R
session. The easiest way to do this is by calling `usethis::edit_r_environ()` and
add the following lines.

```
CDSE_API_USERNAME="<your_CDSE_username>"
CDSE_API_PASSWORD="<your_CDSE_password>"
CDSE_API_CLIENTID="<your_client_id>"
CDSE_API_CLIENTSECRET="<your_client_secret>"
CDSE_API_S3ID="<your_s3_key>"
CDSE_API_S3SECRET="<your_s3_secret>"
```

Obviously, you need to replace the text between `<>` with your personal details.
In a new R session, these environment variables are available, and
you no longer have to pass secret information as arguments to the functions.
They are obtained from the environment.

## Memoise

It is important to note that the functions `dse_access_token()` and
`dse_public_access_token()` are both wrapped in `memoise::memoise()`.
This means that the function result is stored in memory and when the function
is called again, previous results are reused.

One advantage is that it speeds up operations, when you have multiple sequential
calls that need access tokens. It will also prevent rate limiting actions, where
the server might block you when you excessively request tokens.

A disadvantage is that you token might expire, and an old token is reused due to
the memoise strategy. If this happens, you can wipe the memoised token by calling
`memoise::forget(dse_access_token)` or `memoise::forget(dse_public_access_token)`,
respectively. After that you will get a fresh token when calling the token functions.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odata_bursts.R
\name{dse_odata_bursts_request}
\alias{dse_odata_bursts_request}
\alias{dse_odata_bursts}
\title{Create a OData Request for a Data Space Ecosystem Bursts Data}
\usage{
dse_odata_bursts_request(...)

dse_odata_bursts(...)
}
\arguments{
\item{...}{Ignored}
}
\value{
Returns a \code{data.frame} with burst information.
}
\description{
Obtain metadata for burst data associated with specific products.
}
\details{
For more details about bursts check the
\href{https://documentation.dataspace.copernicus.eu/APIs/Sentinel-1\%20SLC\%20Burst.html}{burst API documentation}.

You can apply some tidyverse functions (see \link{tidy_verbs}) to \code{odata_request}
object returned by \code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}}. These apply lazy evaluation.
Meaning that they are just added to the object and are only evaluated after
calling either \code{\link[dplyr:compute]{dplyr::compute()}} or \code{\link[dplyr:collect]{dplyr::collect()}} (see examples).
}
\examples{
if (interactive() && dse_has_s3_secret()) {
  dse_odata_bursts(ParentProductId == "879d445c-2c67-5b30-8589-b1f478904269")
  
  burst_req <-
    dse_odata_bursts_request(ParentProductId == "879d445c-2c67-5b30-8589-b1f478904269")
  
  ## Note that these are large files and may take a while to download:
  dse_odata_download(
    burst_req,
    tempdir()
  )
}
}
\seealso{
Other odata:
\code{\link[=dse_odata_attributes]{dse_odata_attributes()}},
\code{\link[=dse_odata_download]{dse_odata_download()}},
\code{\link[=dse_odata_download_path]{dse_odata_download_path()}},
\code{\link[=dse_odata_product_nodes]{dse_odata_product_nodes()}},
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}},
\code{\link[=dse_odata_quicklook]{dse_odata_quicklook()}}
}
\concept{odata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stac.R
\name{dse_stac_collections}
\alias{dse_stac_collections}
\title{Get a Summary of all Data Space Ecosystem Collections}
\usage{
dse_stac_collections(collection, limit = 1000L, ...)
}
\arguments{
\item{collection}{A specific collection for which to obtain summary information.
If missing (default), all collections are returned.}

\item{limit}{Maximum number of collections to be returned.}

\item{...}{Ignored}
}
\value{
Returns a \code{data.frame} with the requested information
}
\description{
Use the STAC API to get a summary of all collections available
from the interface.
}
\examples{
if (interactive()) {
  dse_stac_collections()
  dse_stac_collections("sentinel-2-l2a")
}
}
\seealso{
Other stac:
\code{\link[=dse_stac_client]{dse_stac_client()}},
\code{\link[=dse_stac_download]{dse_stac_download()}},
\code{\link[=dse_stac_get_uri]{dse_stac_get_uri()}},
\code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}},
\code{\link[=dse_stac_queryables]{dse_stac_queryables()}},
\code{\link[=dse_stac_search_request]{dse_stac_search_request()}}
}
\concept{stac}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/login.R
\name{dse_set_gdal_token}
\alias{dse_set_gdal_token}
\title{Set Copernicus Data Space Ecosystem Access Token for GDAL Driver}
\usage{
dse_set_gdal_token(token = dse_public_access_token())
}
\arguments{
\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}
}
\value{
Returns a \code{logical} value. \code{TRUE} if all variables
were successfully set. \code{FALSE} otherwise.
}
\description{
This function sets system environment variables, such
that the GDAL library can access the Copernicus Data Space
Ecosystem https storage. Note that these settings can be used
by any package depending on the GDAL library. Most notably:
\code{stars}, \code{terra}, and \code{gdalraster}.
}
\examples{
if (interactive() && dse_has_client_info() &&
    requireNamespace("stars")) {
  uri <-
    dse_stac_get_uri(
      "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
      "B01", type = "odata")
   
  dse_set_gdal_token()

  ## As this URI is zipped, it need to be downloaded.
  ## But you can access it directly:
  jp2 <- stars::read_stars(uri)
}
}
\seealso{
Other authentication:
\code{\link[=dse_access_token]{dse_access_token()}},
\code{\link[=dse_get_token_details]{dse_get_token_details()}},
\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}},
\code{\link[=dse_usage]{dse_usage()}}
}
\concept{authentication}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req_perform.R
\name{req_perform}
\alias{req_perform}
\title{Perform a Request to Get a Response}
\usage{
req_perform(
  req,
  path = NULL,
  verbosity = NULL,
  mock = getOption("httr2_mock", NULL),
  error_call = rlang::current_env()
)
}
\arguments{
\item{req}{Either a \code{\link[httr2:request]{httr2::request()}} class object or an \code{odata_request}
class object. The latter can be created with \code{\link[=dse_odata_products_request]{dse_odata_products_request()}} and
\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}}.}

\item{path}{Optionally, path to save body of the response. This is useful
for large responses since it avoids storing the response in memory.}

\item{verbosity}{How much information to print? This is a wrapper
around \code{\link[httr2:req_verbose]{req_verbose()}} that uses an integer to control verbosity:
\itemize{
\item \code{0}: no output
\item \code{1}: show headers
\item \code{2}: show headers and bodies
\item \code{3}: show headers, bodies, and curl status messages.
}

Use \code{\link[httr2:with_verbosity]{with_verbosity()}} to control the verbosity of requests that
you can't affect directly.}

\item{mock}{A mocking function. If supplied, this function is called
with the request. It should return either \code{NULL} (if it doesn't want to
handle the request) or a \link[httr2]{response} (if it does). See
\code{\link[httr2:with_mocked_responses]{with_mocked_responses()}}/\code{local_mocked_responses()} for more details.}

\item{error_call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\value{
Returns a \code{httr2::response} class object
}
\description{
A wrapper around \code{\link[httr2:req_perform]{httr2::req_perform()}}, which can also handle
\code{odata_request} class objects. Check \code{\link[httr2:req_perform]{httr2::req_perform()}}
for details.
}
\seealso{
Other tidyverse:
\code{\link{tidy_verbs}}
}
\concept{tidyverse}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odata_products.R
\name{dse_odata_download_path}
\alias{dse_odata_download_path}
\title{Alternative Route to Download OData Products}
\usage{
dse_odata_download_path(
  product,
  node_path = "",
  destination,
  progress = TRUE,
  ...,
  token = dse_public_access_token()
)
}
\arguments{
\item{product}{Hexadecimal id of the product to be downloaded}

\item{node_path}{Path to a specific file in the product. When left blank (\code{""})
The function will attempt to download the entire product as a zip archive.}

\item{destination}{Path to a directory where to store the downloaded file}

\item{progress}{\code{logical} value. If \code{TRUE} shows download progress.}

\item{...}{Ignored}

\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}
}
\value{
Returns a \code{httr2_response} class object. It's body will hold the filename
of the downloaded file
}
\description{
Downloading data using the OData API is probably fastest by using
\code{\link[=dse_odata_download]{dse_odata_download()}}. As an alternative, you can use this function
which uses the https protocol to download a product.
}
\examples{
if (interactive() && dse_has_account()) {

  dse_odata_download_path(
    product     = "2f497806-0101-5eea-83fa-c8f68bc56b0c",
    node_path   = 
      paste("DEM1_SAR_DTE_90_20101213T034716_20130408T035028_ADS_000000_5033.DEM",
            "Copernicus_DSM_30_S09_00_E026_00", "DEM",
            "Copernicus_DSM_30_S09_00_E026_00_DEM.dt1", sep = "/"),
    destination = tempdir()
  )
  
  dse_odata_download_path(
    product     = "ce4576eb-975b-40ff-8319-e04b00d8d444",
    destination = tempdir())

}
}
\seealso{
Other odata:
\code{\link[=dse_odata_attributes]{dse_odata_attributes()}},
\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}},
\code{\link[=dse_odata_download]{dse_odata_download()}},
\code{\link[=dse_odata_product_nodes]{dse_odata_product_nodes()}},
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}},
\code{\link[=dse_odata_quicklook]{dse_odata_quicklook()}}
}
\concept{odata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_process}
\alias{dse_sh_process}
\title{Process Satellite Data and Download Result}
\usage{
dse_sh_process(
  input,
  output,
  evalscript,
  destination,
  ...,
  token = dse_access_token()
)
}
\arguments{
\item{input}{A named \code{list} specifying the input satellite data to
be processed with \code{evalscript} to an image. A correctly formatted
\code{list} can be created with \code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}}.}

\item{output}{A named \code{list} specifying the how to present the output
image, create with \code{evalscript}. A correctly formatted \code{list} can be
created with \code{\link[=dse_sh_prepare_output]{dse_sh_prepare_output()}}.}

\item{evalscript}{A \code{character} string containing a piece of JavaScript,
that will be run on the Sentinel Hub server. It is used to translate
satellite data to pixel data in a geo-referenced image. For more information
on setting up such a script please consult
\href{https://docs.sentinel-hub.com/api/latest/evalscript/}{the API documentation}.
You can also use \code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}} to obtain ready-to-go
scripts from the SentinelHub repository.}

\item{destination}{A file name to store the downloaded image.}

\item{...}{Ignored}

\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}
}
\value{
A \code{httr2_response} class object containing the
location of the downloaded file at its \code{destination}.
}
\description{
Users can request raw satellite data, simple band combinations such
as false colour composites, calculations of simple remote sensing
indices like NDVI, or more advanced processing such as calculation
of Leaf area index (LAI).
}
\details{
Use \code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}} if you want to use the graphical
user interface at
\href{https://shapps.dataspace.copernicus.eu/requests-builder/}{Sentinel Requests Builder}.
to define a request.
}
\examples{
if (interactive() && dse_has_client_info()) {

  bounds <- c(5.261, 52.680, 5.319, 52.715)
  
  ## prepare input data
  input <-
    dse_sh_prepare_input(
      bounds = bounds,
      time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
    )

  ## prepare ouput format
  output <- dse_sh_prepare_output(bbox = bounds)
  
  ## retrieve processing script
  evalscript <- dse_sh_get_custom_script("/sentinel-2/l2a_optimized/")
  
  fl <- tempfile(fileext = ".tiff")
  ## send request and download result:
  dse_sh_process(input, output, evalscript, fl)

  if (requireNamespace("stars")) {
    library(stars)
    enkhuizen <- read_stars(fl) |> suppressWarnings()
    plot(enkhuizen, rgb = 1:3, axes = TRUE, main = "Enkhuizen")
  }
}
}
\references{
\itemize{
\item \url{https://docs.sentinel-hub.com/api/latest/api/process/}
\item \url{https://shapps.dataspace.copernicus.eu/requests-builder/}
\item \url{https://custom-scripts.sentinel-hub.com/}
\item \url{https://github.com/sentinel-hub/custom-scripts}
\item \url{https://docs.sentinel-hub.com/api/latest/evalscript/}
}
}
\seealso{
Other sentinelhub:
\code{\link[=dse_sh_collections]{dse_sh_collections()}},
\code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}},
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}},
\code{\link[=dse_sh_queryables]{dse_sh_queryables()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}},
\code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stac.R
\name{dse_stac_download}
\alias{dse_stac_download}
\title{Download Asset From STAC Catalogue}
\usage{
dse_stac_download(
  asset_id,
  asset,
  collection = dse_stac_guess_collection,
  destination,
  ...,
  s3_key = dse_s3_get_key(),
  s3_secret = dse_s3_get_secret(),
  token = dse_public_access_token()
)
}
\arguments{
\item{asset_id}{STAC id, used for locating the asset download details.}

\item{asset}{Name of the asset to download}

\item{collection}{The identifier for a collection. The default argument
is the \code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}} function which tries to guess the
collection id from the \code{asset_id}. A more rigid approach is to provide
the collection id as a \code{character} string.}

\item{destination}{Directory path where to store the downloaded file.}

\item{...}{Ignored}

\item{s3_key, s3_secret}{The s3 key and secret registered under your Data Space
Ecosystem account}

\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}
}
\value{
Returns the path to the downloaded file.
}
\description{
Use \code{\link[=dse_stac_search_request]{dse_stac_search_request()}} to identify assets that can be downloaded.
Use \code{\link[=dse_stac_download]{dse_stac_download()}} to download an asset by its STAC id and asset name.
}
\examples{
if (interactive() && (dse_has_s3_secret() || dse_has_client_info())) {
  dse_stac_download(
    asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
    asset = "B01",
    destination = tempdir()
  )
}
}
\seealso{
Other stac:
\code{\link[=dse_stac_client]{dse_stac_client()}},
\code{\link[=dse_stac_collections]{dse_stac_collections()}},
\code{\link[=dse_stac_get_uri]{dse_stac_get_uri()}},
\code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}},
\code{\link[=dse_stac_queryables]{dse_stac_queryables()}},
\code{\link[=dse_stac_search_request]{dse_stac_search_request()}}
}
\concept{stac}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odata_products.R
\name{dse_odata_product_nodes}
\alias{dse_odata_product_nodes}
\title{List OData Product Nodes (i.e. Files and Directories)}
\usage{
dse_odata_product_nodes(product, node_path = "", recursive = FALSE, ...)
}
\arguments{
\item{product}{A product identifier (\code{Id})}

\item{node_path}{Path of nodes separated by forward slashes (\code{"/"}).
Path for which to list nodes. Default is \code{""}, which is the root of
the product}

\item{recursive}{A \code{logical} value. If set to \code{TRUE}, it will recursively
list all nested nodes. Default is \code{FALSE}.}

\item{...}{Ignored}
}
\value{
A \code{data.frame} with information on the requested node(s)
}
\description{
If you know the product \code{Id}, you can use this function to retrieve
information about nodes (i.e. files and directories) within the product.
}
\examples{
if (interactive()) {
  nodes <- dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4")
  nodes <- dse_odata_product_nodes("c8ed8edb-9bef-4717-abfd-1400a57171a4",
                                   recursive = TRUE)
}
}
\seealso{
Other odata:
\code{\link[=dse_odata_attributes]{dse_odata_attributes()}},
\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}},
\code{\link[=dse_odata_download]{dse_odata_download()}},
\code{\link[=dse_odata_download_path]{dse_odata_download_path()}},
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}},
\code{\link[=dse_odata_quicklook]{dse_odata_quicklook()}}
}
\concept{odata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geometry.R
\name{st_intersects}
\alias{st_intersects}
\alias{st_intersects.odata_request}
\alias{st_intersects.stac_request}
\title{Filter OData and STAC Requests Using Geometries}
\usage{
st_intersects.odata_request(x, y, sparse = FALSE, ...)

st_intersects.stac_request(x, y, sparse = FALSE, ...)
}
\arguments{
\item{x}{Either an \code{odata_request} class object, generated with
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}}/\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}}; or
a \code{stac_request} generated with \code{\link[=dse_stac_search_request]{dse_stac_search_request()}}.}

\item{y}{A spatial geometry of either class \code{sf} (see \code{\link[sf:st_as_sf]{sf::st_as_sf()}}) or
\code{sfc} (see \code{\link[sf:st_as_sfc]{sf::st_as_sfc()}}). It
will always be transformed to WGS 84 projection (EPSG:4326).}

\item{sparse}{Argument inherited from generic definition. Ignored in this context}

\item{...}{Ignored}
}
\value{
Returns an object of the same class as \code{x}, with the geometry filter added
to it.
}
\description{
Filters OData and STAC rows that intersect with spatial feature \code{y}.
}
\examples{
if (interactive()) {
  bbox <-
    sf::st_bbox(
      c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
      crs = 4326) |>
      sf::st_as_sfc()

  dse_odata_products_request() |>
    dplyr::filter(
      `ContentDate/Start` > "2025-01-01") |>
    sf::st_intersects(bbox) |>
    dplyr::collect()

  dse_stac_search_request("sentinel-2-l1c") |>
    sf::st_intersects(bbox) |>
    dplyr::collect()
}
}
\concept{geometry}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CopernicusDataspace-package.R
\docType{package}
\name{CopernicusDataspace-package}
\alias{CopernicusDataspace}
\alias{CopernicusDataspace-package}
\title{CopernicusDataspace: Search Download and Handle Data from the Copernicus Data Space Ecosystem}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

The Copernicus Data Space Ecosystem, is an open ecosystem that provides free instant access to a wide range of data and services from the Copernicus Sentinel missions and more on our planet’s land, oceans and atmosphere. This package provides entry points to several APIs allowing users to access the data directly in R.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/pepijn-devries/CopernicusDataspace}
  \item \url{https://pepijn-devries.github.io/CopernicusDataspace/}
  \item Report bugs at \url{https://github.com/pepijn-devries/CopernicusDataspace/issues}
}

}
\author{
\strong{Maintainer}: Pepijn de Vries \email{pepijn.devries@outlook.com} (\href{https://orcid.org/0000-0002-7961-6646}{ORCID})

Authors:
\itemize{
  \item Pepijn de Vries \email{pepijn.devries@outlook.com} (\href{https://orcid.org/0000-0002-7961-6646}{ORCID})
}

Other contributors:
\itemize{
  \item Alicia Hamer \email{alicia.hamer@wur.nl} (\href{https://orcid.org/0000-0001-6809-622X}{ORCID}) [research team member]
  \item LVVN (Ministerie van Landbouw, Visserij, Voedselzekerheid en Natuur) (\href{https://ror.org/03b1hdw57}{ROR}) [funder]
  \item WMR (Wageningen Marine Research) (\href{https://ror.org/013dzwk66}{ROR}) [contractor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{dse_s3_uri_to_vsi}
\alias{dse_s3_uri_to_vsi}
\title{Convert Uniform Resource Identifier to Virtual System Identifier}
\usage{
dse_s3_uri_to_vsi(uri, streaming = TRUE)
}
\arguments{
\item{uri}{A Uniform Resource Identifier, pointing to an
S3 storage file. You can retrieve one with \code{\link[=dse_stac_get_uri]{dse_stac_get_uri()}}.}

\item{streaming}{A \code{logical} value that allows to toggle between
\code{"\\\\vsis3\\\\"} and \code{"\\\\vsis3_streaming\\\\"} (default). The latter is
faster for reading files from its resource, but does not allow
random access. The first supports random access, but is not as fast
at reading.}
}
\value{
A \code{character} string representing the VSI
}
\description{
Convert Uniform Resource Identifier (URI) to Virtual System
Identifier (VSI). The Copernicus Data Space Ecosystem returns
URIs for accessing assets. Packages that use the GDAL library
(e.g., \code{stars}, \code{terra} and \code{gdalraster})
can use VSI to access raster data directly. Use this function
to convert such an URI to a VIS.
}
\examples{
if (interactive()) {
  dse_stac_get_uri(
    "S1A_IW_GRDH_1SDV_20241125T055820_20241125T055845_056707_06F55C_12F9_COG",
    "vh", "sentinel-1-grd") |>
    dse_s3_uri_to_vsi()
}
}
\seealso{
Other s3:
\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}},
\code{\link[=dse_s3_download]{dse_s3_download()}},
\code{\link[=dse_s3_set_gdal_options]{dse_s3_set_gdal_options()}}
}
\concept{s3}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_custom_scripts}
\alias{dse_sh_custom_scripts}
\title{List Custom JavaScripts for Processing Sentinel Hub Data}
\usage{
dse_sh_custom_scripts(...)
}
\arguments{
\item{...}{Ignored}
}
\value{
Returns a \code{data.frame} with custom scripts, containing
a column with a title and one with a relative URL.
}
\description{
Custom Eval Scripts, that can be used in Sentinel Hub requests,
for processing data. This functions lists scripts available
from \url{https://github.com/sentinel-hub/custom-scripts}. They
can be retrieved with \code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}}.
}
\details{
Make sure that you have sufficient monthly quota left to
process images. You can check with \code{\link[=dse_usage]{dse_usage()}}.
}
\examples{
if (interactive()) {
  dse_sh_custom_scripts()
}
}
\references{
\itemize{
\item \url{https://custom-scripts.sentinel-hub.com/}
\item \url{https://github.com/sentinel-hub/custom-scripts}
}
}
\seealso{
Other sentinelhub:
\code{\link[=dse_sh_collections]{dse_sh_collections()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}},
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}},
\code{\link[=dse_sh_process]{dse_sh_process()}},
\code{\link[=dse_sh_queryables]{dse_sh_queryables()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}},
\code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odata_products.R
\name{dse_odata_attributes}
\alias{dse_odata_attributes}
\title{List OData Attributes}
\usage{
dse_odata_attributes(...)
}
\arguments{
\item{...}{Ignored}
}
\value{
A \code{data.frame} listing all attributes for each collection.
}
\description{
Collect a list of OData attributes that can be used for
filtering products with \code{\link[=dse_odata_products_request]{dse_odata_products_request()}}.
}
\examples{
if (interactive()) {
  dse_odata_attributes()
}
}
\seealso{
Other odata:
\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}},
\code{\link[=dse_odata_download]{dse_odata_download()}},
\code{\link[=dse_odata_download_path]{dse_odata_download_path()}},
\code{\link[=dse_odata_product_nodes]{dse_odata_product_nodes()}},
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}},
\code{\link[=dse_odata_quicklook]{dse_odata_quicklook()}}
}
\concept{odata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyverse.R
\name{tidy_verbs}
\alias{tidy_verbs}
\alias{filter}
\alias{filter.odata_request}
\alias{filter.sentinel_request}
\alias{filter.stac_request}
\alias{compute}
\alias{compute.odata_request}
\alias{collect}
\alias{collect.odata_request}
\alias{collect.sentinel_request}
\alias{collect.stac_request}
\alias{arrange}
\alias{arrange.odata_request}
\alias{arrange.stac_request}
\alias{slice_head}
\alias{slice_head.odata_request}
\alias{slice_head.stac_request}
\alias{slice_head.sentinel_request}
\alias{select}
\alias{select.odata_request}
\alias{select.stac_request}
\alias{select.sentinel_request}
\title{Tidy Verbs for OData, SentinelHub and STAC API Requests}
\usage{
filter.odata_request(.data, ..., .by = NULL, .preserve = FALSE)

filter.sentinel_request(.data, ..., .by = NULL, .preserve = FALSE)

filter.stac_request(.data, ..., .by = NULL, .preserve = FALSE)

compute.odata_request(x, skip = 0L, ...)

collect.odata_request(x, skip = 0L, ...)

collect.sentinel_request(x, skip = 0L, ...)

collect.stac_request(x, ...)

arrange.odata_request(.data, ..., .by_group = FALSE)

arrange.stac_request(.data, ..., .by_group = FALSE)

slice_head.odata_request(.data, ..., n, prop, by = NULL)

slice_head.stac_request(.data, ..., n, prop, by = NULL)

slice_head.sentinel_request(.data, ..., n, prop, by = NULL)

select.odata_request(.data, ...)

select.stac_request(.data, ...)

select.sentinel_request(.data, ...)
}
\arguments{
\item{.data, x}{An object of any of the following classes \code{odata_request},
\code{sentinel_request} or \code{stac_request}.
These are produced by
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}}, \code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}} and \code{\link[=dse_stac_search_request]{dse_stac_search_request()}}}

\item{...}{Data masking expressions, or arguments passed to embedded functions}

\item{skip}{Number of rows to skip when collecting results. The APIs
return a limited number rows. Specify the number of rows to skip
in order to get results beyond the predefined limit.}

\item{n}{Maximum number of rows to return.}

\item{by, .by, .by_group, .preserve, prop}{Arguments inherited from generic
\code{dplyr} functions. Ignored in the current context as either grouping
is not allowed for an OData API request or is otherwise not supported.}
}
\value{
All functions (except \code{collect()}) return a modified
\code{stac_request}/\code{sentinel_request}/\code{odata_request}
object, containing the lazy tidy operations. \code{collect()} will return a
\code{data.frame()} yielding the result of the request.
}
\description{
Implementation of tidy generics for features supported any of OData,
SentinelHub or STAC API requests.
They can be called on objects any of the classes: \code{odata_request},
\code{sentinel_request} or \code{stac_request}.
The first is produced
by \code{\link[=dse_odata_products_request]{dse_odata_products_request()}} and \code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}};
the second by \code{\link[=dse_sh_search_request]{dse_sh_search_request()}}; and the last
by \code{\link[=dse_stac_search_request]{dse_stac_search_request()}}.
}
\details{
These special request class objects use lazy evaluation.
This means that functions are only evaluated after calling \code{\link[dplyr:collect]{dplyr::collect()}}
on a request.

Note that you should not call the functions exported in this package directly.
Instead, call the generics as declared in the \code{dplyr} package. This is
illustrated by the examples.
\subsection{Slice Head}{

In order to manage server traffic, the OData API never returns more than
20 rows. If you want to obtain results beyond the first 20 rows, you need
to specify the \code{skip} argument.

The Sentinel and STAC API limits its results to the first 10 rows. You can expand that limit
with \code{\link[dplyr:slice_head]{dplyr::slice_head()}}. For STAC the number of rows is capped at 10,000 records.
For SentinelHub this number is capped at 100.
}

\subsection{Deviations}{

Due to limitations posed by the OData API, some tidyverse verbs deviate
from its tidy standards. Most notably:
\itemize{
\item \code{\link[dplyr:select]{dplyr::select()}}: Cannot change the order of columns. It will only
affect which columns are selected. Also, tidy selection helpers like
\code{\link[dplyr:any_of]{dplyr::any_of()}} and \code{\link[dplyr:all_of]{dplyr::all_of()}} are NOT supported
\item \code{\link[dplyr:arrange]{dplyr::arrange()}}: OData only allows to sort up to 32 columns.
Adding more columns will produce a warning.
\item Grouping is not supported
\item Only tidy methods listed in the usage section are supported for the
special request class objects. If you want to apply the full spectrum of
tidyverse methods, call \code{dplyr::collect()} on the request class
object first. That will return a normal \code{data.frame}, which can be
manipulated further.
}
}
}
\examples{
library(dplyr)
if (interactive()) {
  dse_odata_products_request() |>
    filter(contains(Name, "WRR")) |>
    select("Id", "Name") |>
    arrange(Id, desc(Name)) |>
    slice_head(n = 5) |>
    collect()

  dse_stac_search_request("sentinel-1-grd") |>
    filter(`sat:orbit_state` == "ascending") |>
    arrange("id") |>
    collect()
  
  if (dse_has_client_info()) {
    dse_sh_search_request(
      collection = "sentinel-2-l2a",
      bbox       = c(5.261, 52.680, 5.319, 52.715),
      datetime   = c("2025-01-01 UTC", "2025-01-31 UTC")
    ) |>
      filter(`eo:cloud_cover` <= 10) |>
      collect()
  }
}
}
\seealso{
Other tidyverse:
\code{\link[=req_perform]{req_perform()}}
}
\concept{tidyverse}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_prepare_input}
\alias{dse_sh_prepare_input}
\alias{dse_sh_prepare_output}
\title{Prepare Input and Output Fields for Sentinel Hub Request}
\usage{
dse_sh_prepare_input(
  bounds,
  time_range,
  collection_name = "sentinel-2-l2a",
  id = NA,
  max_cloud_coverage = 100,
  mosaicking_order = "default",
  upsampling = "default",
  downsampling = "default",
  harmonize_values = FALSE,
  ...
)

dse_sh_prepare_output(
  width = 512,
  height = 512,
  output_format = "tiff",
  bbox,
  ...
)
}
\arguments{
\item{bounds}{A bounding box or geometry (classes \code{sf::bbox},
\code{sf::sf}, \code{sf::sfc}) defining the boundaries of the output image.}

\item{time_range}{A \code{vector} of two date-time values, specifying
the time range for satellite data to include in the process.}

\item{collection_name}{A collection name. defaults to \code{"sentinel-2-l2a"}
to ensure you get Sentinel-2 L2A data.}

\item{id}{An identifier. Not documented by the API reference material.}

\item{max_cloud_coverage}{Maximum cloud cover to be included in the
process. Value between 0 and 100 (default) percent.}

\item{mosaicking_order}{Sets the order of overlapping tiles from which
the output result is mosaicked. Should be any of \code{"default"},
\code{"mostRecent"}, \code{"leastRecent"}, or \code{"leastCC"}. See also
\href{https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/#mosaickingorder}{the API documentation}.}

\item{upsampling, downsampling}{Specify the interpolation technique
when the output resolution is smaller or larger respectively than
the available source data.
See also
\href{https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/#processing-options}{the API documentation}.}

\item{harmonize_values}{A \code{logical} value indicating whether units
are harmonised as indicated in
\href{https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/#harmonize-values}{the API documentation}.}

\item{...}{Ignored}

\item{width, height}{Size of the output image in pixels. These
are ignored if \code{bbox} is specified.}

\item{output_format}{File format for the output file. Should
be one of \code{"tiff"} (default), \code{"jpeg"}, \code{"png"}, or \code{"json"}.}

\item{bbox}{You can optionally provide a bounding box (i.e., a
copy of \code{bounds}) to calculate width and height with fixed
aspect ratio. Width will be 512 be definition, the height is chosen
such that it matches with the bounding box}
}
\value{
A named \code{list} that can be used as \code{input} and \code{output}
argument to \code{\link[=dse_sh_process]{dse_sh_process()}}.
}
\description{
\code{\link[=dse_sh_process]{dse_sh_process()}} requires a named \code{list} for \code{input} and
\code{output} settings. The functions documented here produce those lists
required by such a process request.
}
\examples{
dse_sh_prepare_input(
  bounds = c(5.261, 52.680, 5.319, 52.715),
  time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
)

library(sf)
shape <- st_bbox(c(xmin = 5.261, ymin = 52.680,
                   xmax = 5.319, ymax = 52.715), crs = 4326) |>
           st_as_sfc()
dse_sh_prepare_input(
  bounds = shape,
  time_range = c("2025-06-01 UTC", "2025-07-01 UTC")
)

dse_sh_prepare_output(bbox = shape)
}
\references{
\itemize{
\item \url{https://shapps.dataspace.copernicus.eu/requests-builder/}
}
}
\seealso{
\code{\link[=dse_sh_process]{dse_sh_process()}}

Other sentinelhub:
\code{\link[=dse_sh_collections]{dse_sh_collections()}},
\code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}},
\code{\link[=dse_sh_process]{dse_sh_process()}},
\code{\link[=dse_sh_queryables]{dse_sh_queryables()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}},
\code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stac.R
\name{dse_stac_guess_collection}
\alias{dse_stac_guess_collection}
\title{Guess the Collection id from an Asset id}
\usage{
dse_stac_guess_collection(asset_id)
}
\arguments{
\item{asset_id}{An asset identifier name, used to guess its parent
collection id.}
}
\value{
A \code{character} string with a guessed collection id. Or \code{NA} in
case it cannot make a guess.
}
\description{
As the STAC catalogue contains a large number of records, your request
may receive a timeout error. To prevent this it is best to narrow down
your requests to specific collections. This function is a helper function
that tries to guess the collection id from an asset id. Note that this
method is not highly reliable, and it is always best to manually provide
a collection id to a request.
}
\examples{
dse_stac_guess_collection(
  "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148")
}
\seealso{
Other stac:
\code{\link[=dse_stac_client]{dse_stac_client()}},
\code{\link[=dse_stac_collections]{dse_stac_collections()}},
\code{\link[=dse_stac_download]{dse_stac_download()}},
\code{\link[=dse_stac_get_uri]{dse_stac_get_uri()}},
\code{\link[=dse_stac_queryables]{dse_stac_queryables()}},
\code{\link[=dse_stac_search_request]{dse_stac_search_request()}}
}
\concept{stac}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_get_custom_script}
\alias{dse_sh_get_custom_script}
\title{Retrieve Custom JavaScripts to be Used by Sentinel Hub}
\usage{
dse_sh_get_custom_script(rel_url)
}
\arguments{
\item{rel_url}{A relative URL found with \code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}}.}
}
\value{
A \code{character} string containing JavaScript code. This
script can be used with \code{\link[=dse_sh_process]{dse_sh_process()}}
}
\description{
Sentinel Hub uses JavaScripts to process satellite images.
There is a repository with such custom scripts. They can be
listed with \code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}}. Use the relative URL
(relUrl) from that list to obtain the actual script with this
function.
}
\examples{
if (interactive()) {
  dse_sh_get_custom_script("/sentinel-2/tonemapped_natural_color/")
}
}
\seealso{
Other sentinelhub:
\code{\link[=dse_sh_collections]{dse_sh_collections()}},
\code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}},
\code{\link[=dse_sh_process]{dse_sh_process()}},
\code{\link[=dse_sh_queryables]{dse_sh_queryables()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}},
\code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_search_request}
\alias{dse_sh_search_request}
\title{Create a Request for the SentinelHub Catalogue}
\usage{
dse_sh_search_request(
  collection,
  bbox,
  datetime,
  ...,
  token = dse_access_token()
)
}
\arguments{
\item{collection}{A collection for which to list the features.
See \code{\link[=dse_sh_collections]{dse_sh_collections()}} for a list of Sentinel Hub collections.}

\item{bbox}{An object that can be converted into a \code{bbox} class
object (see \code{\link[sf:st_bbox]{sf::st_bbox()}}).}

\item{datetime}{A date-time object, or a vector of two date time
objects (in case of a range). Or an object that can be converted into
a datetime object.}

\item{...}{Ignored}

\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}
}
\value{
Returns a \code{sentinel_request} class object, which inherits
from the \code{httr2::request} class. Call \code{\link[dplyr:collect]{dplyr::collect()}} on it to
retrieve results.
}
\description{
In order to perform a search using the STAC API, you first need to
create a request using \code{\link[=dse_sh_search_request]{dse_sh_search_request()}}. This creates
a \code{\link[httr2:request]{httr2::request()}} to which tidy verbs \code{?tidy_verbs} can be applied
(e.g., \code{\link[dplyr:select]{dplyr::select()}} and \code{\link[dplyr:filter]{dplyr::filter()}}.
Results are retrieved by calling \code{\link[dplyr:collect]{dplyr::collect()}} on the request.
}
\examples{
if (interactive() && dse_has_client_info()) {
  library(dplyr)
  
  dse_sh_search_request(
    collection = "sentinel-2-l2a",
    bbox       = c(5.261, 52.680, 5.319, 52.715),
    datetime   = c("2025-01-01 UTC", "2025-01-31 UTC")
  ) |>
    filter(`eo:cloud_cover` <= 10) |>
    collect()
}
}
\references{
\itemize{
\item \url{https://docs.sentinel-hub.com/api/latest/api/catalog/}
}
}
\seealso{
Other sentinelhub:
\code{\link[=dse_sh_collections]{dse_sh_collections()}},
\code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}},
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}},
\code{\link[=dse_sh_process]{dse_sh_process()}},
\code{\link[=dse_sh_queryables]{dse_sh_queryables()}},
\code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odata_products.R
\name{dse_odata_download}
\alias{dse_odata_download}
\title{Download Data Space Ecosystem Products Through OData API}
\usage{
dse_odata_download(
  request,
  destination,
  ...,
  s3_key = dse_s3_get_key(),
  s3_secret = dse_s3_get_secret()
)
}
\arguments{
\item{request}{A request containing products or burst data that you wish to download.
Use \code{\link[=dse_odata_products_request]{dse_odata_products_request()}} or \code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}} to formulate
product or burst specifications.}

\item{destination}{A \code{character} string specifying the directory path,
where to store downloaded products}

\item{...}{Ignored.}

\item{s3_key, s3_secret}{The s3 key and secret registered under your Data Space
Ecosystem account}
}
\value{
A vector of downloaded file names stored at \code{destination}
}
\description{
Use \code{\link[=dse_odata_products]{dse_odata_products()}} or \code{\link[=dse_odata_products_request]{dse_odata_products_request()}},
\code{\link[=dse_odata_bursts]{dse_odata_bursts()}} or \code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}} to find
a product or burst information. Use this function to download the product(s)
or burst information.
}
\examples{
if (interactive() && dse_has_s3_secret()) {
  dse_odata_download(
    dse_odata_products(Name == "S1C_AUX_PP2_V20241204T000000_G20251024T110034.SAFE"),
    destination = tempdir())
  dse_odata_download(
    dse_odata_products(
      Name ==
        "S1A_IW_OCN__2SDH_20250707T210608_20250707T210625_059983_07739B_893E.SAFE"),
    destination = tempdir())
  dse_odata_download(
    dse_odata_products(
      Id \%in\% c("c8ed8edb-9bef-4717-abfd-1400a57171a4",
                "86288a07-560c-364f-b8ce-669d95f06fa0")),
    destination = tempdir())
}
}
\seealso{
Other odata:
\code{\link[=dse_odata_attributes]{dse_odata_attributes()}},
\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}},
\code{\link[=dse_odata_download_path]{dse_odata_download_path()}},
\code{\link[=dse_odata_product_nodes]{dse_odata_product_nodes()}},
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}},
\code{\link[=dse_odata_quicklook]{dse_odata_quicklook()}}
}
\concept{odata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stac.R
\name{dse_stac_get_uri}
\alias{dse_stac_get_uri}
\title{Get a Uniform Resource Identifier (URI) for an Asset in a Product}
\usage{
dse_stac_get_uri(
  asset_id,
  asset,
  collection = dse_stac_guess_collection,
  type = "s3",
  ...
)
}
\arguments{
\item{asset_id}{STAC id, used for locating the asset download details.}

\item{asset}{Name of the asset to download}

\item{collection}{The identifier for a collection. The default argument
is the \code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}} function which tries to guess the
collection id from the \code{asset_id}. A more rigid approach is to provide
the collection id as a \code{character} string.}

\item{type}{Which type of URI should be returned? Defaults
to \code{"s3"}. Use \code{"odata"} to get the alternative https URI.}

\item{...}{Ignored}
}
\value{
Returns the URI as a \code{character} string.
If available, the local path for an asset is returned as attribute.
}
\description{
Get a Uniform Resource Identifier (URI) for an asset in a product.
This can be used to download a file manually or connect to the
asset directly straight from the source.
}
\examples{
if (interactive()) {
  dse_stac_get_uri(
    asset_id = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148",
    asset = "B01"
  )
}
}
\seealso{
Other stac:
\code{\link[=dse_stac_client]{dse_stac_client()}},
\code{\link[=dse_stac_collections]{dse_stac_collections()}},
\code{\link[=dse_stac_download]{dse_stac_download()}},
\code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}},
\code{\link[=dse_stac_queryables]{dse_stac_queryables()}},
\code{\link[=dse_stac_search_request]{dse_stac_search_request()}}
}
\concept{stac}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{dse_s3_download}
\alias{dse_s3_download}
\title{Download Asset Through Uniform Resource Identifier}
\usage{
dse_s3_download(
  uri,
  destination,
  ...,
  s3_key = dse_s3_get_key(),
  s3_secret = dse_s3_get_secret()
)
}
\arguments{
\item{uri}{A Uniform Resource Identifier (URI, starting with "s3://").
You can look for them in the STAC catalogue, either using a
\href{https://browser.stac.dataspace.copernicus.eu/}{web browser} or
\code{\link[=dse_stac_search_request]{dse_stac_search_request()}} (see example).}

\item{destination}{Destination path to a directory where to store the downloaded file(s)}

\item{...}{Ignored}

\item{s3_key, s3_secret}{The s3 key and secret registered under your Data Space
Ecosystem account}
}
\value{
A vector of file names stored at \code{destination}
}
\description{
When the Uniform Resource Identifier (URI, starting with "s3://") for
an asset is known, this function can be used to download it
}
\examples{
if (interactive() && dse_has_s3_secret()) {
  library(dplyr)

  ## Retrieve a URI for a specific asset through the STAC
  ## catalogue:   
  my_uri <-
    dse_stac_search_request(
      ids = "S2A_MSIL1C_20260109T132741_N0511_R024_T39XVL_20260109T142148") |>
        select("assets.B01.href") |>
        arrange("id") |>
        collect() |>
        pull("assets") |>
        unlist()
  
  dse_s3_download(my_uri, tempdir())

}
}
\seealso{
Other s3:
\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}},
\code{\link[=dse_s3_set_gdal_options]{dse_s3_set_gdal_options()}},
\code{\link[=dse_s3_uri_to_vsi]{dse_s3_uri_to_vsi()}}
}
\concept{s3}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stac.R
\name{dse_stac_queryables}
\alias{dse_stac_queryables}
\title{Get Queryables for a STAC Collection}
\usage{
dse_stac_queryables(collection, ...)
}
\arguments{
\item{collection}{Name of the collection for which to get the queryables.}

\item{...}{Ignored}
}
\value{
Returns a named list with information about elements that can be used
to query the \code{collection}
}
\description{
When searching through a collection with \code{\link[=dse_stac_search_request]{dse_stac_search_request()}}, it
can be helpful to know which elements can be used to filter the search
results (using \code{\link[dplyr:filter]{dplyr::filter()}}). Calling \code{\link[=dse_stac_queryables]{dse_stac_queryables()}} tells
you which aspects are available for querying and expected formats.
}
\examples{
if (interactive()) {
  dse_stac_queryables("sentinel-1-grd")
}
}
\seealso{
Other stac:
\code{\link[=dse_stac_client]{dse_stac_client()}},
\code{\link[=dse_stac_collections]{dse_stac_collections()}},
\code{\link[=dse_stac_download]{dse_stac_download()}},
\code{\link[=dse_stac_get_uri]{dse_stac_get_uri()}},
\code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}},
\code{\link[=dse_stac_search_request]{dse_stac_search_request()}}
}
\concept{stac}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/login.R
\name{dse_access_token}
\alias{dse_access_token}
\alias{dse_public_access_token}
\alias{dse_get_client_id}
\alias{dse_set_client_id}
\alias{dse_set_username}
\alias{dse_set_password}
\alias{dse_get_client_secret}
\alias{dse_get_username}
\alias{dse_get_password}
\alias{dse_set_client_secret}
\alias{dse_has_client_info}
\alias{dse_has_account}
\title{Client Information and Access Token for the Data Space Store API}
\usage{
dse_access_token(
  client_id = dse_get_client_id(),
  client_secret = dse_get_client_secret(),
  ...
)

dse_public_access_token(
  username = dse_get_username(),
  password = dse_get_password()
)

dse_get_client_id(...)

dse_set_client_id(value, ...)

dse_set_username(value, ...)

dse_set_password(value, ...)

dse_get_client_secret(...)

dse_get_username(...)

dse_get_password(...)

dse_set_client_secret(value, ...)

dse_has_client_info(...)

dse_has_account(...)
}
\arguments{
\item{client_id}{ID of the client registered under your account.}

\item{client_secret}{Secret provided for the client registered under your account.}

\item{...}{Ignored}

\item{username}{Your Copernicus Data Space username (usually your e-mail address).}

\item{password}{Your Copernicus Data Space password for your account.}

\item{value}{Assignment value for setting \code{username}, \code{password}, \code{client_id}
or \code{client_secret} as environment variable. Once set, it will persist for
the remainder of the R session.}
}
\value{
In case of \code{dse_get_client_id()} and \code{dse_get_client_secret()}, you can get (or
set) client details as environment variables. This way, they will persist
throughout your R session.

The function \code{dse_has_client_info()} returns a \code{logical} value, indicating
whether client details (id and secret) are available as environmental variable.
Note that if this function returns \code{TRUE}, it doesn't guarantee that the details
are valid (just that they are available).

In case of \code{dse_access_token()} and \code{dse_public_access_token()}
a named \code{list} is returned, containing the
access token (named \code{"token"}) and some additional meta information.
}
\description{
To regulate server traffic, the Data Space Ecosystem uses user accounts to regulate
and ensure fair usage. These functions will get you a token and information about
your usage.
}
\details{
Before you can use most of the Data Space Ecosystem services, you need to create
an account and register as a client. This will let you retrieve an access token,
that can be used for authentication purposes in API requests. It is also used to
manage your usage and rate limiting (see also \code{\link[=dse_usage]{dse_usage()}} and
\code{\link[=dse_user_statistics]{dse_user_statistics()}}).

Note that \href{https://aws.amazon.com/s3/}{Amazon Simple Storage Service (s3)}
has separate authentication requirements. See \code{\link[=dse_s3_get_key]{dse_s3_get_key()}} for details.
\subsection{Creating an Account}{

First step is creating an account. You can create one by visiting the
\href{https://identity.dataspace.copernicus.eu/auth/realms/CDSE/account/#/personal-info}{login page}
and click "register". Fill out the form and follow the instructions.
}

\subsection{Registering Client}{

In order to register a client, visit the
\href{https://shapps.dataspace.copernicus.eu/dashboard/}{Sentinel Dashboard} and
go to "User settings". There you will have the option to create an OAuth client.
Follow the instructions, and make sure to safely copy the client id and secret.
The latter is only displayed once. More detailed instructions are provided by the
\href{https://documentation.dataspace.copernicus.eu/APIs/SentinelHub/Overview/Authentication.html#registering-oauth-client}{API documentation}
}

\subsection{Client Details as Environment Variables}{

When you share R code, you probably don't want to share your account details.
You can avoid using your \code{client_id} and \code{client_secret} in your script by
setting them as environment variable. You can do this yourself manually by
calling \code{dse_set_client_id()} and \code{dse_set_client_secret()} at the start of each session.

You can download OData simply through https with just you username and
password. You can also store those as environment variables for your convenience.
If you name those \code{CDSE_API_USERNAME} and \code{CDSE_API_PASSWORD} respectively,
they are picked up automatically with \code{\link[=dse_get_username]{dse_get_username()}} and \code{\link[=dse_get_password]{dse_get_password()}}.
OData needs a public access token which is generated with
\code{dse_public_access_token()}, which needs your username and password.

You can also define them in your \code{.Rprofile} file with
\code{Sys.setenv(CDSE_API_CLIENTID = "<your id>")} and
\code{Sys.setenv(CDSE_API_CLIENTSECRET = "<your secret>")}. This way,
they are set each time you start a new R session.

The environment variables are used by default by \code{dse_access_token()},
and \code{dse_public_access_token()} so you
don't have to specify the client details as arguments.
}

\subsection{Obtain Token and Validity}{

After completing the previous two steps, your are now set to obtain an access
token with \code{dse_access_token()}.

Repeatedly requesting an access token may invoke rate limiting measures.
Therefore, this package uses caching to temporarily store the access token
during each R session. Calling \code{dse_access_token()} will therefore only
contact the server once for a token for each unique combination of \code{client_id}
and \code{client_server}. After that, the cached result will be reused during the session

There is a catch: the token provided by the server is only valid for a limited
time (usually 30 minutes). So, when the token has expired, you need to wipe the
cache. You can do so by calling \code{memoise::forget(dse_access_token)} or restarting
the R session.
}
}
\examples{
if (interactive() && dse_has_client_info()) {
  token <- dse_access_token()
}
if (interactive() && dse_has_account()) {
  token_public <- dse_public_access_token()
}
}
\references{
\url{https://documentation.dataspace.copernicus.eu/APIs/SentinelHub/Overview/Authentication.html}
}
\seealso{
\code{\link[=dse_usage]{dse_usage()}}

\code{\link[=dse_user_statistics]{dse_user_statistics()}}

Other authentication:
\code{\link[=dse_get_token_details]{dse_get_token_details()}},
\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}},
\code{\link[=dse_set_gdal_token]{dse_set_gdal_token()}},
\code{\link[=dse_usage]{dse_usage()}}
}
\concept{authentication}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/login.R, R/s3.R
\name{dse_has_s3_secret}
\alias{dse_has_s3_secret}
\alias{dse_s3_get_key}
\alias{dse_s3_set_key}
\alias{dse_s3_get_secret}
\alias{dse_s3_set_secret}
\alias{dse_s3}
\title{Setup Amazon Simple Storage Service for the Data Space Ecosystem}
\usage{
dse_has_s3_secret()

dse_s3_get_key(...)

dse_s3_set_key(value, ...)

dse_s3_get_secret(...)

dse_s3_set_secret(value, ...)

dse_s3(
  region = "us-east-1",
  ...,
  s3_key = dse_s3_get_key(),
  s3_secret = dse_s3_get_secret()
)
}
\arguments{
\item{...}{Ignored}

\item{value}{Replacement value for the \code{s3_key} or \code{s3_secret}.}

\item{region}{\href{https://aws.amazon.com/about-aws/global-infrastructure/regions_az/}{AWS Region}
used in instantiating the client}

\item{s3_key, s3_secret}{The s3 key and secret registered under your Data Space
Ecosystem account}
}
\value{
\code{\link[=dse_s3_get_key]{dse_s3_get_key()}} and \code{\link[=dse_s3_get_secret]{dse_s3_get_secret()}} will return the requested s3 details
if set as environment variable (see details).

\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}} returns a logical value indicating whether s3 details
(key and secret) are set. It will not determine whether the details are valid.
}
\description{
using \href{https://aws.amazon.com/s3/}{Amazon Simple Storage Service (s3)} in the
Data Space Ecosystem requires a key and secret. These functions help you managing
these details and setting up an s3 client.
}
\details{
Working with s3 in the Data Space Ecosystem requires you to create an account,
then register an s3 key as described below. Note that the SentinelHub requires
a different authentication method. See \code{\link[=dse_access_token]{dse_access_token()}} for more details on that.
\subsection{Creating an Account}{

First step is creating an account. You can create one by visiting the
\href{https://identity.dataspace.copernicus.eu/auth/realms/CDSE/account/#/personal-info}{login page}
and click "register". Fill out the form and follow the instructions.
}

\subsection{Registering a s3 Key}{

Now that you have an account, you should visit the
\href{https://eodata-s3keysmanager.dataspace.copernicus.eu/panel/s3-credentials}{s3-credentials page},
and log in with your account details. By clicking "add credential", you can
create a new key and secret. Store them in a safe place, as the secret is only
shown once. You can pass the key and secret as \code{s3_key} and \code{s3_secret} arguments
to functions requesting them. You can also store them as environment variables
such that they persist throughout the R session and don't have to be passed as
arguments (see below).
}

\subsection{S3 Key and Secret as Environment Variables}{

When you share R code, you probably don't want to share your account details.
You can avoid using your \code{s3_key} and \code{s3_secret} in your script by
setting them as environment variable. You can do this yourself manually by
calling \code{dse_s3_set_key()} and \code{dse_s3_set_secret()} at the start of each session.

You can also define them in your \code{.Rprofile} file with
\code{Sys.setenv(CDSE_API_S3ID = "<your key>")} and
\code{Sys.setenv(CDSE_API_S3SECRET = "<your secret>")}. This way,
they are set each time you start a new R session.
}
}
\references{
\url{https://documentation.dataspace.copernicus.eu/APIs/S3.html}
}
\seealso{
Other authentication:
\code{\link[=dse_access_token]{dse_access_token()}},
\code{\link[=dse_get_token_details]{dse_get_token_details()}},
\code{\link[=dse_set_gdal_token]{dse_set_gdal_token()}},
\code{\link[=dse_usage]{dse_usage()}}

Other s3:
\code{\link[=dse_s3_download]{dse_s3_download()}},
\code{\link[=dse_s3_set_gdal_options]{dse_s3_set_gdal_options()}},
\code{\link[=dse_s3_uri_to_vsi]{dse_s3_uri_to_vsi()}}
}
\concept{authentication}
\concept{s3}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_collections}
\alias{dse_sh_collections}
\title{List Sentinel Hub Collections}
\usage{
dse_sh_collections(...)
}
\arguments{
\item{...}{Ignored}
}
\value{
Returns a \code{data.frame} with information about the
collections available from the Sentinel Hub
}
\description{
List collections that are available from the Sentinel Hub.
}
\examples{
if (interactive()) {
  dse_sh_collections()
}
}
\seealso{
Other sentinelhub:
\code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}},
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}},
\code{\link[=dse_sh_process]{dse_sh_process()}},
\code{\link[=dse_sh_queryables]{dse_sh_queryables()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}},
\code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stac.R
\name{dse_stac_client}
\alias{dse_stac_client}
\title{Obtain Information About the STAC Client}
\usage{
dse_stac_client(...)
}
\arguments{
\item{...}{Ignored}
}
\value{
Returns a \code{data.frame} with the requested information
}
\description{
Returns information about the STAC client used in the Data Space Ecosystem
}
\examples{
if (interactive()) {
  dse_stac_client()
}
}
\seealso{
Other stac:
\code{\link[=dse_stac_collections]{dse_stac_collections()}},
\code{\link[=dse_stac_download]{dse_stac_download()}},
\code{\link[=dse_stac_get_uri]{dse_stac_get_uri()}},
\code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}},
\code{\link[=dse_stac_queryables]{dse_stac_queryables()}},
\code{\link[=dse_stac_search_request]{dse_stac_search_request()}}
}
\concept{stac}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/login.R
\name{dse_get_token_details}
\alias{dse_get_token_details}
\title{Decode Access Token}
\usage{
dse_get_token_details(token = dse_access_token())
}
\arguments{
\item{token}{A token obtained with \code{\link[=dse_access_token]{dse_access_token()}} or
\code{\link[=dse_public_access_token]{dse_public_access_token()}}.}
}
\value{
A named \code{list} with token info
}
\description{
This function decodes an access token and returns a named \code{list} with
information from the token.
}
\examples{
if (interactive() && dse_has_client_info()) {
  dse_get_token_details()
}
}
\seealso{
Other authentication:
\code{\link[=dse_access_token]{dse_access_token()}},
\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}},
\code{\link[=dse_set_gdal_token]{dse_set_gdal_token()}},
\code{\link[=dse_usage]{dse_usage()}}
}
\concept{authentication}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stac.R
\name{dse_stac_search_request}
\alias{dse_stac_search_request}
\title{Create a Request for a STAC Search in the Data Space Ecosystem}
\usage{
dse_stac_search_request(collections, ids, ...)
}
\arguments{
\item{collections}{Restrict the search to the collections listed here.
If missing, the collection will be guessed from the \code{ids},
using \code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}}.}

\item{ids}{Restrict the search to ids listed here.}

\item{...}{Arguments appended to search filter request body.}
}
\value{
Returns a \code{data.frame} with search results.
}
\description{
In order to perform a search using the STAC API, you first need to
create a request using \code{\link[=dse_stac_search_request]{dse_stac_search_request()}}. This creates
a \code{\link[httr2:request]{httr2::request()}} to which tidy verbs \code{?tidy_verbs} can be applied
(e.g., \code{\link[dplyr:select]{dplyr::select()}}, \code{\link[dplyr:filter]{dplyr::filter()}} and \code{\link[dplyr:arrange]{dplyr::arrange()}}.
Results are retrieved by calling \code{\link[dplyr:collect]{dplyr::collect()}} on the request.
}
\details{
If you prefer a graphical user interface, you can alternatively use
the \href{https://browser.stac.dataspace.copernicus.eu/}{STAC web browser}.
}
\examples{
if (interactive()) {
  library(dplyr)
  library(sf)
  
  bbox <-
    sf::st_bbox(
      c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
      crs = 4326)

  dse_stac_search_request("sentinel-2-l1c") |>
    filter(`eo:cloud_cover` < 10) |>
    collect()

  dse_stac_search_request("sentinel-1-grd") |>
    filter(`sat:orbit_state` == "ascending") |>
    arrange("id") |>
    st_intersects(bbox) |>
    collect()
}
}
\seealso{
Other stac:
\code{\link[=dse_stac_client]{dse_stac_client()}},
\code{\link[=dse_stac_collections]{dse_stac_collections()}},
\code{\link[=dse_stac_download]{dse_stac_download()}},
\code{\link[=dse_stac_get_uri]{dse_stac_get_uri()}},
\code{\link[=dse_stac_guess_collection]{dse_stac_guess_collection()}},
\code{\link[=dse_stac_queryables]{dse_stac_queryables()}}
}
\concept{stac}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_queryables}
\alias{dse_sh_queryables}
\title{List Queryable Fields on Sentinel Hub}
\usage{
dse_sh_queryables(collection, ..., token = dse_access_token())
}
\arguments{
\item{collection}{Collection id for which to obtain queryable fields.}

\item{...}{Ignored.}

\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}
}
\value{
Returns a named \code{list}, with information about queryable
fields for the specified \code{collection}.
}
\description{
Return queryable fields for a specific collection on Sentinel Hub.
This is useful information when composing a query with
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}}. Use \code{\link[=dse_sh_collections]{dse_sh_collections()}} to list
available collections.
}
\examples{
if (interactive() && dse_has_client_info()) {
  dse_sh_queryables("sentinel-2-l1c")
}
}
\seealso{
Other sentinelhub:
\code{\link[=dse_sh_collections]{dse_sh_collections()}},
\code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}},
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}},
\code{\link[=dse_sh_process]{dse_sh_process()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}},
\code{\link[=dse_sh_use_requests_builder]{dse_sh_use_requests_builder()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odata_products.R
\name{dse_odata_quicklook}
\alias{dse_odata_quicklook}
\title{Download a Quicklook for a Product}
\usage{
dse_odata_quicklook(product, destination, ...)
}
\arguments{
\item{product}{Identifier (Id) for the product for which to obtain a quicklook.}

\item{destination}{A destination path where to store the image.}

\item{...}{Ignored}
}
\value{
Returns \code{NULL} invisibly
}
\description{
Downloads a 'quicklook' for a product. If the \code{rstudioapi} package is installed,
it will attempt to open the image in the Viewer panel.
}
\examples{
if (interactive()) {
  dse_odata_quicklook(
    "91822f33-b15c-5b60-aa39-6d9f6f5c773b",
    tempfile(fileext = ".jpg"))
}
}
\seealso{
Other odata:
\code{\link[=dse_odata_attributes]{dse_odata_attributes()}},
\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}},
\code{\link[=dse_odata_download]{dse_odata_download()}},
\code{\link[=dse_odata_download_path]{dse_odata_download_path()}},
\code{\link[=dse_odata_product_nodes]{dse_odata_product_nodes()}},
\code{\link[=dse_odata_products_request]{dse_odata_products_request()}}
}
\concept{odata}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sentinelhub.R
\name{dse_sh_use_requests_builder}
\alias{dse_sh_use_requests_builder}
\title{Use Requests Builder to Send Processing Request to SentinelHub}
\usage{
dse_sh_use_requests_builder(
  build,
  destination,
  ...,
  token = dse_access_token()
)
}
\arguments{
\item{build}{A \code{character} string copied from the Request Preview
panel at
\href{https://shapps.dataspace.copernicus.eu/requests-builder/}{Sentinel Requests Builder}.
See \code{system.file("requests-builder.txt", package = "CopernicusDataspace")}
for an example of such a text. When you omit this argument, this function
will attempt to retrieve the text from the system's clipboard.}

\item{destination}{A file name to store the downloaded image.}

\item{...}{Ignored}

\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}
}
\value{
A \code{httr2_response} class object obtained after sending the request.
}
\description{
Use
\href{https://shapps.dataspace.copernicus.eu/requests-builder/}{Sentinel Requests Builder}
to compose a request. Copy the text from the 'Request Preview' panel
and submit with this function. Use \code{\link[=dse_sh_process]{dse_sh_process()}} when you want
to define a request in R, without using a web browser.
}
\examples{
## Read text copied from 'Request Preview' panel on
## <https://shapps.dataspace.copernicus.eu/requests-builder/:
requests_builder <-
  system.file("requests-builder.txt", package = "CopernicusDataspace") |>
    readLines(warn = FALSE) |>
    paste(collapse = "\n")

if (interactive() && dse_has_client_info()) {
  dest <- tempfile(fileext = ".tiff")
  dse_sh_use_requests_builder(requests_builder, destination = dest)
}
}
\references{
\itemize{
\item \url{https://shapps.dataspace.copernicus.eu/requests-builder/}
}
}
\seealso{
Other sentinelhub:
\code{\link[=dse_sh_collections]{dse_sh_collections()}},
\code{\link[=dse_sh_custom_scripts]{dse_sh_custom_scripts()}},
\code{\link[=dse_sh_features]{dse_sh_features()}},
\code{\link[=dse_sh_get_custom_script]{dse_sh_get_custom_script()}},
\code{\link[=dse_sh_prepare_input]{dse_sh_prepare_input()}},
\code{\link[=dse_sh_process]{dse_sh_process()}},
\code{\link[=dse_sh_queryables]{dse_sh_queryables()}},
\code{\link[=dse_sh_search_request]{dse_sh_search_request()}}
}
\concept{sentinelhub}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{dse_s3_set_gdal_options}
\alias{dse_s3_set_gdal_options}
\title{Set-up S3 Configuration for GDAL Library}
\usage{
dse_s3_set_gdal_options(
  region = "us-east-1",
  ...,
  s3_key = dse_s3_get_key(),
  s3_secret = dse_s3_get_secret()
)
}
\arguments{
\item{region}{\href{https://aws.amazon.com/about-aws/global-infrastructure/regions_az/}{AWS Region}
used in instantiating the S3 client}

\item{...}{Ignored}

\item{s3_key, s3_secret}{The s3 key and secret registered under your Data Space
Ecosystem account}
}
\value{
Returns a \code{logical} value. \code{TRUE} if all variables
were successfully set. \code{FALSE} otherwise.
}
\description{
This function sets system environment variables, such
that the GDAL library can access the Copernicus Data Space
Ecosystem S3 storage. Note that these settings can be used
by any package depending on the GDAL library. Most notably:
\code{stars}, \code{terra}, and \code{gdalraster}.
}
\examples{
if (interactive() && dse_has_s3_secret() &&
    requireNamespace("stars")) {
  library(dplyr)

  ## Get a Virtual System Interface to a tiff file:
  vsi <-
    dse_stac_get_uri(
      "S1A_IW_GRDH_1SDV_20241125T055820_20241125T055845_056707_06F55C_12F9_COG",
      "vh", "sentinel-1-grd") |>
      dse_s3_uri_to_vsi()

  ## Make sure to set gdal options with required S3 settings
  dse_s3_set_gdal_options()
   
  ## You can now read the file directly from the online storage
  ## without having to download it completely:
  cog <- stars::read_stars(vsi)
   
  ## You can also easily plot a downsampled version
  plot(cog, downsample = 50)
}
}
\seealso{
Other s3:
\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}},
\code{\link[=dse_s3_download]{dse_s3_download()}},
\code{\link[=dse_s3_uri_to_vsi]{dse_s3_uri_to_vsi()}}
}
\concept{s3}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/account.R
\name{dse_usage}
\alias{dse_usage}
\alias{dse_user_statistics}
\title{Get Data Space Account Information}
\usage{
dse_usage(..., token = dse_access_token())

dse_user_statistics(
  range = "DAYS-31",
  resolution = "DAILY",
  token = dse_public_access_token()
)
}
\arguments{
\item{...}{Ignored}

\item{token}{For authentication, many of the Data Space Ecosystem uses
an access token. Either provide your access token, or obtain one automatically
with \code{\link[=dse_access_token]{dse_access_token()}} (default). Without a valid token you will likely get
an "access denied" error.}

\item{range}{Specify a time range for which to obtain user statistics.
The API expects a string starting with a capitalised time unit (\code{"DAYS"}, \code{"HOURS"}),
followed by a dash ("-") and an integer value specifying the length of the period.
Default is \code{"DAYS-31"}.}

\item{resolution}{Specifying a temporal resolution for the user statistics.
should be one of \code{"DAILY"} (default), \code{"MONTHLY"}, or \code{"HOURLY"}.}
}
\value{
A \code{data.frame} with requested information for the user associated with
the provided \code{token}.
}
\description{
In order to guarantee good performance for all users, the Sentinel Hub applies
\href{https://docs.sentinel-hub.com/api/latest/api/overview/rate-limiting/}{rate limiting}.
This policy enforces monthly quotas to your usage. To check your quota and current
usage, you can call \code{dse_usage()} or \code{dse_user_statistics()}.
}
\examples{
if (interactive() && dse_has_client_info()) {
  dse_usage()
  dse_user_statistics()
}
}
\seealso{
\code{\link[=dse_access_token]{dse_access_token()}}

Other authentication:
\code{\link[=dse_access_token]{dse_access_token()}},
\code{\link[=dse_get_token_details]{dse_get_token_details()}},
\code{\link[=dse_has_s3_secret]{dse_has_s3_secret()}},
\code{\link[=dse_set_gdal_token]{dse_set_gdal_token()}}
}
\concept{authentication}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odata_products.R
\name{dse_odata_products_request}
\alias{dse_odata_products_request}
\alias{dse_odata_products}
\title{Create a OData Request for a Data Space Ecosystem Product}
\usage{
dse_odata_products_request(..., expand)

dse_odata_products(..., expand = NULL)
}
\arguments{
\item{...}{Ignored in case of \code{dse_odata_products_request()}. Dots are passed
to embedded \code{\link[dplyr:filter]{dplyr::filter()}} in case of \code{dse_odata_products()}}

\item{expand}{Additional information to be appended to the result.
Should be any of \code{"Attributes"}, \code{"Assets"}, or \code{"Locations"}. Note that,
these columns are not affected by \code{\link[dplyr:select]{dplyr::select()}} calls (before calling
\code{\link[dplyr:collect]{dplyr::collect()}}).}
}
\value{
Returns an \code{odata_request} class object in case of
\code{dse_odata_products_request()}, which is an extension of \code{\link[httr2:request]{httr2::request()}}.
In case of \code{dse_odata_products()} a \code{data.frame} listing requested products is
returned.
}
\description{
OData is an application programming interface (API) used to disseminate Copernicus
Data Space Ecosystem products. This function creates a request for this service,
which can be used to obtain a \code{data.frame} with product information. This request
supports several tidyverse methods for filtering and arranging the data.
}
\details{
You can apply some tidyverse functions (see \link{tidy_verbs}) to \code{odata_request}
object returned by \code{\link[=dse_odata_products_request]{dse_odata_products_request()}}. These apply lazy evaluation.
Meaning that they are just added to the object and are only evaluated after
calling either \code{\link[dplyr:compute]{dplyr::compute()}} or \code{\link[dplyr:collect]{dplyr::collect()}} (see examples).
}
\examples{
if (interactive()) {
  bbox <-
    sf::st_bbox(
      c(xmin = 5.261, ymin = 52.680, xmax = 5.319, ymax = 52.715),
      crs = 4326) |>
      sf::st_as_sfc()

  dse_odata_products_request() |>
    dplyr::filter(
      `ContentDate/Start` > "2025-01-01") |>
    sf::st_intersects(bbox) |>
    dplyr::arrange(dplyr::desc(Id)) |>
    dplyr::slice_head(n = 100) |>
    dplyr::collect()
}
}
\references{
\url{https://documentation.dataspace.copernicus.eu/APIs/OData.html}
}
\seealso{
Other odata:
\code{\link[=dse_odata_attributes]{dse_odata_attributes()}},
\code{\link[=dse_odata_bursts_request]{dse_odata_bursts_request()}},
\code{\link[=dse_odata_download]{dse_odata_download()}},
\code{\link[=dse_odata_download_path]{dse_odata_download_path()}},
\code{\link[=dse_odata_product_nodes]{dse_odata_product_nodes()}},
\code{\link[=dse_odata_quicklook]{dse_odata_quicklook()}}
}
\concept{odata}
